
package com.myarch.antutil;

import java.io.Serializable;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DefaultLogger;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Mapping from Ant logging to commons logging: info -> info verbose->debug
 * debug->trace
 * 
 * Note that in Ant debug is the lowest, two levels down from info. In JCL, it's
 * only one level down, and trace is the lowest.
 * 
 * 
 * @author Alexander Ananiev
 */

public class AntCommonsLogger implements Log, Serializable {
    public final static String LOGBACK_DPBUDDY_ROOT_LOGGER_NAME="dpbuddy";

    public final static String LOG_HTTP_PROP="dpbuddy.log.http";

    // Save prev. value of org.apache.commons.logging.Log here
    private static String previousDefault;

    // We store current project on the thread
    private static ThreadLocal<Project> threadContext=new ThreadLocal<Project>();

    private static String rootLogger=null;

    public static Log init( String rootLoggerParam ) {
        return init( null, rootLoggerParam );
    }

    public static Log init( Project project, String rootLoggerParam ) {

        rootLogger=rootLoggerParam;

        makeDefault();
        String initialLogger="";
        if (rootLogger != null)
            initialLogger=rootLogger;
        if (project != null)
            setProject( project );

        return LogFactory.getLog( initialLogger );
    }

    public static void setProject( Project project ) {
        threadContext.set( project );
    }

    private final static String COMMONS_LOG_DEFAULT_LOG_IMPL_PROP="org.apache.commons.logging.Log";

    static void makeDefault() {

        if (isSetAsDefault()) {
            return;
        }

        // This would redirect all previously defined loggers to us
        LogFactory.releaseAll();
        previousDefault=(String) LogFactory.getFactory()
                        .getAttribute( COMMONS_LOG_DEFAULT_LOG_IMPL_PROP );
        // Log factory uses context class loader which defaults to system - need
        // to change it so it can find us
        Thread.currentThread().setContextClassLoader( AntCommonsLogger.class.getClassLoader() );
        LogFactory.getFactory().setAttribute( COMMONS_LOG_DEFAULT_LOG_IMPL_PROP,
                        AntCommonsLogger.class.getName() );
    }

    private static boolean isSetAsDefault() {
        String currentFactory=(String) LogFactory.getFactory()
                        .getAttribute( COMMONS_LOG_DEFAULT_LOG_IMPL_PROP );
        return ( AntCommonsLogger.class.getName().equals( currentFactory ) );
    }

    private static LoggerConfiguration loggerConfiguration=LoggerConfiguration.ANT_CONFIG;

    public static void setLoggerConfiguration( LoggerConfiguration loggerConfig ) {
        loggerConfiguration=loggerConfig;
    }

    public static void deregister() {
        // If we don't clean cache, commons logging will continue redirecting
        // log to us
        LogFactory.releaseAll();
        LogFactory.getFactory().setAttribute( "org.apache.commons.logging.Log", previousDefault );
    }

    public static Project getProject() {
        Project project=threadContext.get();
        return project;
    }

    private static boolean warningFlag=true;

    @SuppressWarnings("unused")
    private static void printWarning() {
        if (warningFlag)
            System.out.println( "WARNING-AntCommonsLogger was not properly initialized. You must "
                + "initialize AntLogger by calling AntLogger.init before using it." );
        warningFlag=false;

    }

    private static final long serialVersionUID=1L;

    protected String name=null;

    /**
     * Construct a named instance of this Logger. Note that the name is ignored.
     *
     * @param name
     *            Name of the logger to be constructed
     */
    public AntCommonsLogger(String name) {
        this.name=name;
    }

    private Project createProject() throws BuildException {
        Project project;
        project=new Project();
        project.init();
        DefaultLogger antLogger=new DefaultLogger();
        antLogger.setOutputPrintStream( System.out );
        antLogger.setErrorPrintStream( System.err );
        antLogger.setMessageOutputLevel( Project.MSG_INFO );
        project.addBuildListener( antLogger );
        return project;
    }

    private boolean isApacheHttpLoggingEnabled( Project project ) {
        // currently we use log.http prop to enable logging from all commons
        // components
        return AntUtils.isPropertyTrue( project, LOG_HTTP_PROP );
    }

    private final static String APACHE_HTTP_LOGGER_NAME="org.apache.http";

    private void log( int level, String msg, Throwable ex ) {

        Project project=getProject();
        if (project == null) {
            project=createProject();
        }

        // special treatment for apache http logging -- otherwise it dumps
        // enormous amount of logs in verbose mode
        if (name != null && name.startsWith( APACHE_HTTP_LOGGER_NAME )
            && !isApacheHttpLoggingEnabled( project )) {
            return;
        }

        Task task=project.getThreadTask( Thread.currentThread() );

        if (level == Project.MSG_WARN) {
            msg=loggerConfiguration.getWarningPrefix() + msg;
        }

        if (ex == null) {
            performLogging( project, task, msg, level );
        }
        else {
            project.log( msg, ex, level );
        }
    }

    private void performLogging( Project project, Task task, String msg, int level ) {
        // task can be null when for example a build file is opened in eclipse
        if (task != null && loggerConfiguration.isPrefixWithTaskName()) {
            project.log( task, msg, level );
        }
        else {
            project.log( msg, level );
        }

        logUsingLogback( task, msg, level );

    }

    private void logUsingLogback( Task task, String msg, int level ) {
        String loggerName=LOGBACK_DPBUDDY_ROOT_LOGGER_NAME;
        if (task != null) {
            loggerName+="." + AntUtils.removeTaskPrefix( task.getTaskName() );
        }

        Logger logbackLogger=LoggerFactory.getLogger( loggerName );

        if (level == Project.MSG_WARN && logbackLogger.isWarnEnabled()) {
            logbackLogger.warn( msg );
        }
        else if (level == Project.MSG_INFO && logbackLogger.isInfoEnabled()) {
            logbackLogger.info( msg );
        }
        else if (level == Project.MSG_VERBOSE && logbackLogger.isDebugEnabled()) {
            logbackLogger.debug( msg );
        }
        else if (level == Project.MSG_DEBUG && logbackLogger.isTraceEnabled()) {
            logbackLogger.trace( msg );
        }
    }

    /**
     * Logs a message with <code>Project.MSG_VERBOSE</code>.
     *
     * @param message
     *            to log
     * @see org.apache.commons.logging.Log#debug(Object)
     */
    public void debug( Object message ) {
        log( determineDebugLogLevel(), String.valueOf( message ), null );
    }

    /**
     * Logs a message with <code>Project.MSG_VERBOSE</code>.
     *
     * @param message
     *            to log
     * @param exception
     *            log this cause
     * @see org.apache.commons.logging.Log#debug(Object, Throwable)
     */
    public void debug( Object message, Throwable exception ) {
        log( determineDebugLogLevel(), String.valueOf( message ), exception );
    }

    /**
     * Logs a message with <code>Project.MSG_ERR</code>.
     *
     * @param message
     *            to log
     * @see org.apache.commons.logging.Log#error(Object)
     */
    public void error( Object message ) {
        log( Project.MSG_ERR, String.valueOf( message ), null );
    }

    /**
     * Logs a message with <code>Project.MSG_ERR</code>.
     *
     * @param message
     *            to log
     * @param exception
     *            log this cause
     * @see org.apache.commons.logging.Log#error(Object, Throwable)
     */
    public void error( Object message, Throwable exception ) {
        log( Project.MSG_ERR, String.valueOf( message ), exception );
    }

    /**
     * Logs a message with <code>Project.MSG_ERR</code>.
     *
     * @param message
     *            to log
     * @see org.apache.commons.logging.Log#fatal(Object)
     */
    public void fatal( Object message ) {
        log( Project.MSG_ERR, String.valueOf( message ), null );
    }

    /**
     * Logs a message with <code>Project.MSG_ERR</code>.
     *
     * @param message
     *            to log
     * @param exception
     *            log this cause
     * @see org.apache.commons.logging.Log#fatal(Object, Throwable)
     */
    public void fatal( Object message, Throwable exception ) {
        log( Project.MSG_ERR, String.valueOf( message ), exception );
    }

    /**
     * Logs a message with <code>java.util.logging.Level.INFO</code>.
     *
     * @param message
     *            to log
     * @see org.apache.commons.logging.Log#info(Object)
     */
    public void info( Object message ) {
        log( determineInfoLogLevel(), String.valueOf( message ), null );
    }

    /**
     * Logs a message with <code>Project.MSG_INFO</code>.
     *
     * @param message
     *            to log
     * @param exception
     *            log this cause
     * @see org.apache.commons.logging.Log#info(Object, Throwable)
     */
    public void info( Object message, Throwable exception ) {
        log( determineInfoLogLevel(), String.valueOf( message ), exception );
    }

    private int determineInfoLogLevel() {
        int antLogLevel=Project.MSG_INFO;
        if (loggerConfiguration.isLogInfoAtWarningLevel())
            antLogLevel=Project.MSG_WARN;

        return antLogLevel;
    }

    private int determineDebugLogLevel() {
        int antLogLevel=Project.MSG_VERBOSE;
        if (loggerConfiguration.isLogInfoAtWarningLevel())
            antLogLevel=Project.MSG_INFO;

        return antLogLevel;
    }

    /**
     * Logs a message with <code>Project.MSG_DEBUG</code>.
     *
     * @param message
     *            to log
     * @see org.apache.commons.logging.Log#trace(Object)
     */
    public void trace( Object message ) {
        log( Project.MSG_DEBUG, String.valueOf( message ), null );
    }

    /**
     * Logs a message with <code>Project.MSG_DEBUG</code>.
     *
     * @param message
     *            to log
     * @param exception
     *            log this cause
     * @see org.apache.commons.logging.Log#trace(Object, Throwable)
     */
    public void trace( Object message, Throwable exception ) {
        log( Project.MSG_DEBUG, String.valueOf( message ), exception );
    }

    /**
     * Logs a message with <code>Project.MSG_WARN</code>.
     *
     * @param message
     *            to log
     * @see org.apache.commons.logging.Log#warn(Object)
     */
    public void warn( Object message ) {
        log( Project.MSG_WARN, String.valueOf( message ), null );
    }

    /**
     * Logs a message with <code>jProject.MSG_WARN</code>.
     *
     * @param message
     *            to log
     * @param exception
     *            log this cause
     * @see org.apache.commons.logging.Log#warn(Object, Throwable)
     */
    public void warn( Object message, Throwable exception ) {
        log( Project.MSG_WARN, String.valueOf( message ), exception );
    }

    /**
     * Is debug logging currently enabled?
     */
    public boolean isDebugEnabled() {
        return true;
    }

    /**
     * Is error logging currently enabled?
     */
    public boolean isErrorEnabled() {
        return true;
    }

    /**
     * Is fatal logging currently enabled?
     */
    public boolean isFatalEnabled() {
        return true;
    }

    /**
     * Is info logging currently enabled?
     */
    public boolean isInfoEnabled() {
        return true;
    }

    /**
     * Is trace logging currently enabled?
     */
    public boolean isTraceEnabled() {
        return true;
    }

    /**
     * Is warn logging currently enabled?
     */
    public boolean isWarnEnabled() {
        return true;
    }

}
