package com.myarch.antutil;

import java.io.File;
import java.util.Enumeration;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.lang3.StringUtils;
import org.apache.tools.ant.*;
import org.apache.tools.ant.taskdefs.ImportTask;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.PatternSet;

public class AntUtils {

    public static FileSet createFileSetFromPatternSet( Project antProject, File baseDir, PatternSet patternSet ) {
        FileSet fs=new FileSet();
        fs.setProject( antProject );
        fs.setDir( baseDir );
        PatternSet includePS=fs.createPatternSet();
        includePS.addConfiguredPatternset( patternSet );

        return fs;
    }

    public static boolean isPropertyPrefixExist( Project project, String prefix ) {

        Map<String, Object> props=project.getProperties();
        for (String key : props.keySet()) {
            if (StringUtils.startsWithIgnoreCase( key, prefix.trim() ))
                return true;
        }

        return false;
    }

    public static boolean isPropertyTrue( Project project, String name ) {

        boolean bool=false;
        // in some cases we use it from unit tests
        if (project == null)
            return false;

        String val=getPropertyUsingCamelCaseAlias( project, name );
        if (val != null)
            bool=Project.toBoolean( val );
        return bool;
    }

    public static String getPropertyUsingCamelCaseAlias( Project project, String name ) {
        String val=project.getProperty( name );
        if (val == null)
            val=project.getProperty( toCamelCase( name ) );
        return val;
    }

    @SuppressWarnings("unchecked")
    public static <T> T instantiateClass( String className ) {
        T resultingObject;
        try {

            @SuppressWarnings("static-access")
            Class<T> aClass=(Class<T>) AntUtils.class.forName( className );
            resultingObject=aClass.newInstance();

        }
        catch (ClassNotFoundException e) {
            throw new BuildException( "Error attempting to dynamically load a class", e );
        }
        catch (InstantiationException e) {
            throw new BuildException( e );
        }
        catch (IllegalAccessException e) {
            throw new BuildException( e );
        }

        return resultingObject;
    }

    public static void addSystemPropertiesToProject( Project project ) {
        Properties props=System.getProperties();
        for (Map.Entry<Object, Object> prop : props.entrySet()) {
            project.setProperty( prop.getKey().toString(), project.replaceProperties( prop.getValue().toString() ) );
        }
    }

    /**
     * Converts the name from dot-delimited to camel case
     * 
     * @param name
     * @return
     */
    public static String toCamelCase( String name ) {
        name=name.toLowerCase();

        StringBuilder newName=new StringBuilder();
        char prevChar='0';
        for (char c : name.toCharArray()) {
            if (prevChar == '.') {
                c=Character.toUpperCase( c );
            }

            if (c != '.') {
                newName.append( c );
            }
            prevChar=c;
        }
        return newName.toString();
    }

    public static String camelCaseToDotNotation( String name ) {

        StringBuilder newName=new StringBuilder();
        for (char c : name.toCharArray()) {
            if (Character.isUpperCase( c ) && newName.length() > 0) {
                newName.append( '.' );
            }
            newName.append( c );
        }
        return newName.toString().toLowerCase();
    }

    /**
     * Reconfigure -- reevaluate all attributes and reset all instance variables
     * in the task. task.reconfigure does not do it for nested.
     */
    public static void reconfigureWithNested( Task task ) {
        reconfigure( task.getRuntimeConfigurableWrapper(), task.getProject() );
    }

    private static void reconfigure( RuntimeConfigurable runtimeConfigurable, Project project ) {
        runtimeConfigurable.reconfigure( project );
        Enumeration<RuntimeConfigurable> children=runtimeConfigurable.getChildren();

        while (children.hasMoreElements()) {
            RuntimeConfigurable child=children.nextElement();
            // TODO: Reconfiguring XML fragments leads to the issue of text
            // values being appended instead of replaced
            // A better long-term solution is needed
            if (!( child.getProxy() instanceof org.apache.tools.ant.util.XMLFragment.Child ))
                reconfigure( child, project );
        }
    }

    public static String removeTaskPrefix( String taskName ) {
        String nameNoPrefix=taskName;
        int i=taskName.indexOf( ':' );
        if (i >= 0) {
            nameNoPrefix=taskName.substring( i + 1 );
        }

        return nameNoPrefix;
    }

    public static void importFile( Project project, String fileToImport ) {
        // TODO: check if exists?
        // import task does not accept File, has to be a string
        ImportTask importTask=createImportTask(project, fileToImport);
        importTask.execute();
    }

    private static ImportTask createImportTask( Project project, String fileToImport ) {
        ImportTask task=new ImportTask();
        task.setTaskName( "import" );

        task.setFile( fileToImport );
        task.setProject( project );
        Target owningTarget = new Target();
        owningTarget.setName( "" );
        Location dummyLocation=new Location(".");
        task.setLocation( dummyLocation );
        task.setOwningTarget( owningTarget );;

        return task;
    }

}
