package com.myarch.dpbuddy;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.List;

import org.apache.commons.io.IOUtils;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.dpbuddy.xmlutil.SoapUtils;
import com.myarch.dpbuddy.xmlutil.streaming.CombiningInputStream;
import com.myarch.dpbuddy.xmlutil.streaming.ElementMapper;
import com.myarch.dpbuddy.xmlutil.streaming.ExternalContentXmlChunker;

/**
 * Base class for all DataPower request types
 */
public class BaseRequest implements DPRequest {

    private Document doc;
    private Element requestElement;

    public BaseRequest() {
        doc=new Document();
        requestElement=new Element( "request", DPConst.DP_NS );
        doc.setRootElement( requestElement );
    }

    public BaseRequest(String childElementName) {
        this();
        addRequestChild( childElementName );
    }

    public void setDocument( Document doc ) {
        this.doc=doc;
        requestElement=doc.getRootElement();
    }

    public String toString() {
        return JDomUtils.docToString( doc );
    }

    private ElementMapper<InputStream> streamElementMapper;

    public void setStreamElementMapper( ElementMapper<InputStream> streamElementMapper ) {
        this.streamElementMapper=streamElementMapper;
    }

    /**
     * Returns input stream of the request. The request is wrapped in soap
     * header.
     */
    public InputStream getInputStream() {
        InputStream is;

        if (streamElementMapper != null) {
            List<InputStream> streams=ExternalContentXmlChunker.chunk( doc, streamElementMapper );
            // add SOAP 
            streams.add( 0, IOUtils.toInputStream( SoapUtils.SOAP_HEADER, Charset.defaultCharset() ));
            streams.add( IOUtils.toInputStream( SoapUtils.SOAP_FOOTER, Charset.defaultCharset() ));

            is=new CombiningInputStream(streams );
            
        }
        else {
            is=new ByteArrayInputStream( SoapUtils.serializeToBytesWithSoap( getDocument() ) );
        }
        
        return is;
    }

    /**
     * @return requestElement.
     */
    public Element getRequestElement() {
        return requestElement;
    }

    /**
     * Adds a child element to the "request" element of the request
     * 
     * @param eltName
     *            name of the element to add
     * @return newly created element
     */
    public Element addRequestChild( String eltName ) {
        Element requestChild=new Element( eltName, DPConst.DP_NS );
        getRequestElement().addContent( requestChild );

        return requestChild;
    }

    /**
     * Returns the element of the given name which is the immediate child of the
     * 'request' element. If the element is not found, throws
     * IllegalStateException
     * 
     * @param eltName
     *            name of the request's child element
     * @return found element
     */
    public Element getRequestChild( String eltName ) {
        Element childElt=getRequestElement().getChild( eltName, DPConst.DP_JDOM_NS );
        if (childElt == null)
            throw new IllegalStateException(
                            String.format( "'%s' is not found among children of 'request'", eltName ) );
        return childElt;
    }

    public String getDisplayName() {
        String displayName=null;
        List<Element> children=requestElement.getChildren();
        if (children.size() == 0)
            return displayName;
        Element opElt=children.get( 0 );
        displayName=opElt.getName();
        // we have do-action logic here so we can reuse it for AnyRequest
        if (displayName.equals( "do-action" )) {
            displayName+="/" + getActionName( opElt );
        }

        return displayName;
    }

    public String getActionName( Element opElement ) {

        String name=null;
        List<Element> children=opElement.getChildren();
        if (children.size() > 0) {
            Element actionElt=children.get( 0 );
            name=actionElt.getName();
        }

        return name;
    }

    public Document getDocument() {
        return doc;
    }

}
