package com.myarch.dpbuddy;

import java.util.Comparator;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.jdom2.Element;

import com.google.gson.JsonObject;

/**
 * DataPower configuration object. An object always has class (type) and name.
 */
public class DPObject implements Comparable<DPObject> {

    public static final DPObject ALL_OBJECTS=new DPObject(".*", null);
    
    private String clazz;
    private String name;

    public DPObject() {
        super();
    }

    public DPObject(String clazz, String name) {
        super();
        setClass( clazz );
        setName( name );
    }

    public DPObject(DPObject dpObject) {
        this( dpObject.getClazz(), dpObject.getName() );
    }

    
    public DPObject(Pair<String, String> objectTuple) {
        this( objectTuple.getKey(), objectTuple.getValue() );
    }
    
    /**
     * Creates DPObject from an Element representing DataPower object in
     * DataPower configuration format: HTTPUserAgent name="default"
     * 
     * @param dpObjectElt
     */
    public DPObject(Element dpObjectElt) {
        clazz=dpObjectElt.getName();
        name=dpObjectElt.getAttributeValue( "name" );
    }

    public String getClazz() {
        return clazz;
    }

    public void setClass( String clazz ) {
        this.clazz=StringUtils.isNotBlank( clazz )?clazz:null;
    }

    public String getName() {
        return name;
    }

    public void setName( String name ) {
        this.name=StringUtils.isNotBlank( name )?name:null;
    }

    public boolean isEmpty(){
        return StringUtils.isAllEmpty( getClazz(), getName() );
    }
    
    public String toClassObjectStr() {
        StringBuilder str=new StringBuilder();
        if (StringUtils.isNotBlank( clazz))
            str.append( clazz );
        if (StringUtils.isNotBlank( name)) {
            if (str.length()>0)
                str.append( ':' );
            str.append( name );
        }
        
        return str.toString();
    }

    public String toString() {
        return toClassObjectStr();
    }

    public boolean equals( Object obj ) {
        if (this == obj) return true;
        if (!( obj instanceof DPObject )) return false;

        DPObject otherDPObject=(DPObject) obj;

        return toClassObjectStr().equals( otherDPObject.toClassObjectStr() );
    }

    public int hashCode() {
        int hash=1;
        hash=hash * 31 + ( clazz == null ? 0 : clazz.hashCode() );
        hash=hash * 31 + ( name == null ? 0 : name.hashCode() );

        return hash;
    }
    

    public static String objectListToString( Iterable<? extends DPObject> dpObjects ) {
        StringBuilder buf=new StringBuilder();
        for (DPObject dpObject : dpObjects) {
            if (buf.length() > 0) {
                buf.append( ", " );
            }
            buf.append( dpObject.toClassObjectStr() );
        }

        return buf.toString();
    }

    public static boolean objectMatchesFilter( DPObject dpObject, DPObject filter ) {
        return ( filter.getClazz() == null || dpObject.getClazz().matches( "(?i)"+filter.getClazz() ) )
            && ( filter.getName() == null || dpObject.getName().matches( filter.getName() ) );
    }

    
    public Element toXML() {
        Element elt=new Element( "object" );
        if (clazz != null) {
            elt.setAttribute( "class", clazz );
        }
        if (name != null) {
            elt.setAttribute( "name", name );
        }

        return elt;
    }
    
    public JsonObject toJson() {
        JsonObject jsonObj=new JsonObject();
        populateJson(jsonObj);
        return jsonObj;
    }
    
    public void populateJson(JsonObject jsonObj) {
        if (clazz != null) {
            jsonObj.addProperty( "class", clazz );
        }
        if (name != null) {
            jsonObj.addProperty( "object", name );
        }
    }    

    @Override
    /**
     * Compare by type/name (if exists)
     */
    public int compareTo( DPObject anotherDpObject ) {
        int cmp=0;
        cmp=compareWithNulls(name, anotherDpObject.getName());
        if (cmp==0){
            cmp=clazz.compareTo( anotherDpObject.getClazz() );
        }
        return cmp;
    }
    
    private int compareWithNulls(String s1, String s2){
        int cmp = 0;
        if (isStringEmpty(s1)) {
            if (isStringEmpty(s2)) {
                cmp = 0;
            } else {
                cmp = -1;
            }
        } else if (isStringEmpty(s2)) {
            cmp = -1;
        } else {
            cmp = s1.compareTo(s2);
        }
        return cmp;
    }
    
    private boolean isStringEmpty(String str) {
        return str == null || str.isEmpty();
    }
    
    public static class DpObjectElementComparator implements Comparator<Element> {
        
        @Override
        public int compare( Element dpObjElt1, Element dpObjElt2 ) {
            DPObject dpObj1=new DPObject(dpObjElt1);
            DPObject dpObj2=new DPObject(dpObjElt2);
            return dpObj1.compareTo( dpObj2 );
        }

    }    

}
