package com.myarch.dpbuddy;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.stream.util.StreamReaderDelegate;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;

import com.myarch.dpbuddy.action.ActionCommand;
import com.myarch.dpbuddy.client.DPClient;
import com.myarch.dpbuddy.filemanagement.DPFileOrDir;
import com.myarch.dpbuddy.filemanagement.FileStoreInfoResponse;
import com.myarch.dpbuddy.filemanagement.GetFileStoreInfoCommand;
import com.myarch.dpbuddy.status.ObjectStatus;
import com.myarch.dpbuddy.status.ObjectStatusCommand;
import com.myarch.dpbuddy.status.ObjectStatusResponse;
import com.myarch.dpbuddy.status.StatusCommand;
import com.myarch.dpbuddy.status.StatusResponse;
import com.myarch.util.FilePathUtils;

public class Device {

    public final static String DEFAULT_DOMAIN_NAME="default";
    private final Log logger=LogFactory.getLog( getClass() );

    private DPConnection connectionSettings;

    Device(DPConnection connectionSettings) {
        this.connectionSettings=connectionSettings;
    }

    public DPConnection getDPConnection() {
        return connectionSettings;
    }

    /**
     * Create a new device which is a copy of another device. This could be
     * useful when some of the device parameters need to be changed, e.g., we
     * need to connect to the default domain
     * 
     * @param anotherDevice
     */

    public Device(Device anotherDevice) {
        this.connectionSettings=new DPConnection( anotherDevice.connectionSettings );
    }

    public Device switchDefaultDomain() {
        return switchDomain( Device.DEFAULT_DOMAIN_NAME );
    }

    /**
     * Create a new device which is connected to a different domain
     * 
     * @param domain
     * @return new device instance
     */
    public Device switchDomain( String domain ) {
        Device newDevice=new Device( this );
        newDevice.connectionSettings.setDomain( domain );
        return newDevice;
    }

    public String getCurrentDomain() {
        return connectionSettings.getDomain();
    }

    public String getName() {
        return connectionSettings.getDeviceName();
    }

    public Document executeRequest( DPRequest request ) {
        return DPClient.executeRequest( connectionSettings, request );
    }

    public ResultResponse executeRequestWithResultResponse( DPRequest request ) {
        Document responseDoc=executeRequest( request );
        ResultResponse resultResponse=new ResultResponse( responseDoc );

        return resultResponse;
    }

    public Document executeRequest( StreamReaderDelegate responseParser, DPRequest request ) {
        return DPClient.executeRequest( responseParser, connectionSettings, request );
    }

    public Document executeRequest( DPRequest request, boolean use2004URL ) {
        return DPClient.executeRequest( connectionSettings, request, use2004URL );
    }

    public Document executeRequest( String domain, DPRequest request, boolean quiet ) {
        return DPClient.executeRequest( null, null, connectionSettings, domain, request, quiet, false );
    }

    public Document executeRequest( String domain, DPRequest request ) {
        return executeRequest( domain, request, false );
    }

    /**
     * Returns object statuses of all objects
     * 
     * @return
     */
    public Set<ObjectStatus> fetchObjects() {
        return execObjectStatusCommand().getObjectStatuses();
    }

    public Set<ObjectStatus> fetchExistingObjects( Iterable<? extends DPObject> filters ) {
        return execObjectStatusCommand().getObjectStatuses( filters, Collections.<DPObject> emptyList(), true );
    }
    
    /**
     * Fetch objects, error out if none found
     * @param includeFilters
     * @param excludeFilters
     * @return
     */
    public Set<ObjectStatus> fetchExistingObjects( Iterable<? extends DPObject> includeFilters, Iterable<? extends DPObject> excludeFilters) {
        return fetchObjects( includeFilters, excludeFilters, true );
    }

    public Set<ObjectStatus> fetchObjects( Iterable<? extends DPObject> filters,  boolean mustMatch ) {
        return execObjectStatusCommand().getObjectStatuses( filters, Collections.<DPObject> emptyList(), mustMatch );
    }

    public Set<ObjectStatus> fetchObjects( Iterable<? extends DPObject> filters, Iterable<? extends DPObject> excludeFilters, boolean mustMatch ) {
        return execObjectStatusCommand().getObjectStatuses( filters, excludeFilters, mustMatch );
    }

    public Set<ObjectStatus> fetchUnsavedObjects() {
        return execObjectStatusCommand().getUnsavedObjectStatuses();
    }

    private ObjectStatusResponse execObjectStatusCommand() {
        ObjectStatusCommand osc=new ObjectStatusCommand();
        return osc.execute( this );
    }

    public Set<ObjectStatus> fetchObjects( DPObject filter ) {
        List<DPObject> filters=new ArrayList<DPObject>();
        filters.add( filter );
        return fetchExistingObjects( filters );
    }

    public Set<ObjectStatus> fetchObjects( DPObject filter, boolean mustMatch ) {
        List<DPObject> filters=new ArrayList<DPObject>();
        filters.add( filter );
        return fetchObjects( filters, mustMatch );
    }

    public List<DPFileOrDir> fetchMatchingFileInfo( String pattern, boolean filesOnly ) {
        GetFileStoreInfoCommand getFileStoreInfoCmd=new GetFileStoreInfoCommand();
        FileStoreInfoResponse fileStoreInfo=getFileStoreInfoCmd.execute( this );
        return fileStoreInfo.findMatching( pattern, filesOnly );
    }

    public List<DPFileOrDir> fetchMatchingFileInfoSkipOverMatchingDirs( String pattern ) {
        GetFileStoreInfoCommand getFileStoreInfoCmd=new GetFileStoreInfoCommand();
        FileStoreInfoResponse fileStoreInfo=getFileStoreInfoCmd.execute( this );
        return fileStoreInfo.findMatchingSkipOverMatchingDirs( pattern );
    }

    public Set<String> fetchDomains() {
        StatusCommand command=new StatusCommand( "DomainStatus" );
        StatusResponse response=command.execute( this );
        Set<String> domains=new HashSet<String>( response.getStatusValues( "Domain" ) );
        return domains;
    }

    public Set<String> fetchRequiredDomains( Iterable<Pattern> patterns ) {
        return fetchDomains( patterns, true );
    }

    public static Set<String> removeDefaultDomain(Set<String> domains) {
        Set<String> filteredDomains=new HashSet<>();
        for(String domain:domains) {
            if (!DEFAULT_DOMAIN_NAME.equalsIgnoreCase( domain )) {
                filteredDomains.add( domain );
            }
        }
        
        return filteredDomains;
    }
    
    public Set<String> fetchDomains( Iterable<Pattern> patterns, boolean isMatchRequired ) {
        Set<String> allDomains=fetchDomains();
        Set<String> matchedDomains=new LinkedHashSet<String>();
        for (Pattern pattern : patterns) {
            int matchesCount=0;
            for (String domain : allDomains) {
                Matcher matcher=pattern.matcher( domain );
                if (matcher.matches()) {
                    matchedDomains.add( domain );
                    ++matchesCount;
                }
            }
            if (matchesCount == 0) {
                if (isMatchRequired)
                    throw new DPBuddyException( "Could not find a DataPower domain matching the pattern '%s'",
                                    pattern );
                else
                    logger.info( "Didn't find any domains matching the pattern '" + pattern + "', continuing" );
            }
        }

        return matchedDomains;
    }

    public final static String DEFAULT_FILESTORE="local";

    /**
     * If a filesystem in the format filesystem:/ is not specified, insert the
     * default filesystem (local:/)
     * 
     * @param path
     *            path of the resource on the device
     * @return path with the filesystem
     */
    public static String prependDefaultFileStore( String path ) {
        return prependFileStore( DEFAULT_FILESTORE, path );
    }

    public static String prependFileStore( String fileStore, String path ) {
        if (!path.contains( ":/" )) {
            path=fileStore + ":/" + path;
        }
        path=FilePathUtils.normalize( path );

        return path;
    }

    public static boolean isFileStoreRoot( String path ) {
        path=FilenameUtils.normalize( path, true );
        String pathWithoutFileStorePrefix=path;
        if (path.contains( ":" ))
            pathWithoutFileStorePrefix=StringUtils.substringAfter( path, ":" );

        return path.contains( ":" ) && StringUtils.isEmpty( pathWithoutFileStorePrefix )
            || StringUtils.isNotEmpty( pathWithoutFileStorePrefix )
                && StringUtils.containsOnly( pathWithoutFileStorePrefix, '/' );
    }

    public static String getFileStorePrefixFromDPPath( String path ) {
        int colonPos=path.indexOf( ':' );
        if (colonPos < 0)
            throw new DPBuddyException(
                            "Invalid DataPower path '%s'. The path must contain filestore prefix such as 'local://'",
                            path );

        String prefix=path.substring( 0, colonPos + 1 );

        return prefix;
    }

    /**
     * Converts boolean to on/off value for the elements that use dmToggle type.
     * 
     * @param bool
     * @return
     */
    public static String booleanToToggle( boolean bool ) {
        return ( bool ? "on" : "off" );
    }

    public void saveConfiguration() {
        ActionCommand saveCmd=new ActionCommand( "SaveConfig" );
        saveCmd.execute( this );
    }

}
