package com.myarch.dpbuddy.ant;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.logging.Log;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.kohsuke.args4j.Argument;
import org.kohsuke.args4j.Option;

import com.myarch.antutil.AntCommonsLogger;
import com.myarch.antutil.AntUtils;
import com.myarch.antutil.LoggerConfiguration;
import com.myarch.antutil.TaskParameterValidator;
import com.myarch.dpbuddy.*;
import com.myarch.dpbuddy.action.ActionCommand;
import com.myarch.dpbuddy.audit.AuditEvent;
import com.myarch.dpbuddy.audit.AuditEventHandler;
import com.myarch.dpbuddy.checkpoint.RollbackCheckpointCommand;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.client.DPSchemaValidator;
import com.myarch.dpbuddy.license.DPBuddyLicenseManager;
import com.myarch.dpbuddy.license.License;
import com.myarch.dpbuddy.license.LicenseException;
import com.myarch.dpbuddy.xmltransform.FileTransformer;
import com.myarch.dpbuddy.xmlutil.XMLInput;
import com.myarch.environment.EnvironmentAntTask;
import com.myarch.propmanagement.ExpressionEvaluator;
import com.myarch.propmanagement.PropRepository;
import com.myarch.propmanagement.PropSet;
import com.myarch.util.DecryptionUtils;
import com.myarch.util.FilePathUtils;
import com.myarch.util.StringPatternUtils;

public abstract class BaseDPBuddyTask extends Task {

    // only log messages from this root logger
    public final static String ROOT_LOGGER="com.myarch";

    protected ExtensionInitializer extensionInitializer;

    protected PropSet envPropSet;
    protected PropSet allPropSet;

    protected ExpressionEvaluator propExpressionEvaluator;

    protected DeviceFactory deviceFactory;

    protected AuditEventHandler auditEventHandler;
    protected Log logger;

    public BaseDPBuddyTask() {
        logger=AntCommonsLogger.init( ROOT_LOGGER );
    }

    protected void deregisterLogger() {
        AntCommonsLogger.deregister();
    }

    private String urlAttr;
    private String usernameAttr;
    private String passwordAttr;
    private String domainAttr;

    private Boolean isTrustAllCertsAttr;

    private Boolean isContinueOnErrorAttr;
    private boolean isContinueOnErrorEffective=false;

    @Option(name="-save", handler=BoolOptionHandler.class,
                    usage="Save the domain's configuration after the command is finished.")
    private Boolean isSaveAttr;

    private boolean isSaveEffective=false;

    private String schemaFile;

    @Option(name="-validate", handler=BoolOptionHandler.class,
                    usage="Validate all XML SOMA requests against DataPower schema.")
    protected Boolean isValidateAttr;

    @Option(name="-license", usage="Path to the license file")
    private String licenseFile;

    // CLI-only:
    @Option(name="-configFile", aliases="-confFile",
                    usage="Path to the configuration file. This could be properties or .conf file.")
    private String configFile;

    public void setConfigFile( String configFile ) {
        this.configFile=configFile;
    }

    @Option(name="-v", aliases="-verbose", usage="Trigger verbose logging")
    private boolean isVerbose=false;

    @Option(name="-h", aliases="-help", usage="Show command's usage")
    private boolean isHelpOption=false;

    // this is used to capture all unparsed arguments. Task name is always the
    // first
    @Argument
    private List<String> cliArguments=new ArrayList<String>();

    public List<String> getCliArguments() {
        return cliArguments;
    }

    // * end of CLI-only

    @Option(name="-env", aliases= { "-device", "-environment" },
                    usage="Name of the target environment, which is either the path of a HOCON object or a properties prefix")
    private String envPrefix;

    protected String currentEnvPrefix;
    private String savedEnvPrefix;

    public String getConfigFile() {
        return configFile;
    }

    // this is for CLI only
    public boolean isVerbose() {
        return isVerbose;
    }

    // this is for CLI only
    public boolean isHelpOption() {
        return isHelpOption;
    }

    protected FileTransformer fileTransformer;

    public final static String ENV_PREFIX_PROP_NAME="dp.env.prefix";
    public final static String ENV_PROP_NAME="dp.env";
    public final static String ENV_PARAM_PROP_NAME="_dpbuddy.env.param";
    public final static String DEVICE_PROP_NAME="dp.device";

    public final static String USERNAME_PROP_NAME="dp.username";
    public final static String PASSWORD_PROP_NAME="dp.password";

    // These two properties mean the same thing
    public final static String XML_MGM_URL_PROP_NAME="dp.xmlmgm.url";
    public final static String URL_PROP_NAME="dp.url";
    public final static String HOST_PROP_NAME="dp.host";

    public final static String DOMAIN_PROP_NAME="dp.domain";
    public final static String TRUST_ALL_CERTS_PROP_NAME="dp.trust.all.certs";
    public final static String SCHEMA_FILE_PROP_NAME=DPSchemaValidator.DP_SCHEMA_PROP;
    public final static String VALIDATE_PROP_NAME="dp.validate";

    public final static String CONTINUE_ON_ERROR_PROP_NAME="dp.continue.on.error";

    public final static String AUTO_SAVE_PROP_NAME="dp.auto.save";

    public final static String LICENSE_FILE_PROP_NAME="dpbuddy.license.file";
    private final static String LICENSE_FILE_ENV_NAME="DPBUDDY_LICENSE_FILE";

    public final static String GRADLE_LOGGING_PROP_NAME="dpbuddyGradleLogging";

    @Option(name="-trustAllCerts", handler=BoolOptionHandler.class,
                    usage="Ignore SSL certificates validation errors when connecting to a device")
    public void setTrustAllCerts( Boolean ignoreCertError ) {
        this.isTrustAllCertsAttr=ignoreCertError;
    }

    @Option(name="-masterPassword", usage="Master password to decrypt encrypted properties")
    protected void setMasterPassword( String masterPassword ) {
        System.setProperty( DecryptionUtils.DEFAULT_MASTER_PASSWORD_PROP_NAME, masterPassword );
    }

    @Option(name="-password", aliases="-p", usage="Password")
    public void setPassword( String password ) {
        this.passwordAttr=password;
    }

    // Ant alias -- because it was on 3.3, keeping it for backward compat
    public void setPasswd( String password ) {
       this.setPassword( password );
    }

    @Option(name="-url", aliases="-host",
                    usage="URL of the SOAP interface (XML Management interface) or the host name of the target device")
    public void setUrl( String url ) {
        this.urlAttr=url;
    }

    public void setHost( String host ) {
        setUrl( host );
    }

    @Option(name="-user", aliases= { "-u" }, usage="Valid username with administrative privileges.")
    public void setUser( String username ) {
        this.usernameAttr=username;
    }

    // alias for legacy Ant
    public void setUserName( String username ) {
        setUser( username );
    }

    /**
     * @param domain
     *            the domain to set
     */
    @Option(name="-domain", usage="Target domain")
    public void setDomain( String domain ) {
        this.domainAttr=domain;
    }

    public void setSave( Boolean isSave ) {
        this.isSaveAttr=isSave;
    }

    public void setSchemaFile( String schemaFile ) {
        this.schemaFile=schemaFile;
    }

    public void setValidate( Boolean isValidate ) {
        this.isValidateAttr=isValidate;
    }

    public void setLicense( String licenseFile ) {
        this.licenseFile=licenseFile;
    }

    public void setEnv( String envPrefix ) {
        this.envPrefix=envPrefix;
    }

    public void setEnvPrefix( String envPrefix ) {
        setEnv( envPrefix );
    }

    public void setDevice( String device ) {
        setEnv( device );
    }

    @Option(name="-continueOnError", handler=BoolOptionHandler.class,
                    usage="Don't stop if the command failed for one of the devices; attempt to run the command against all the devices regardless of failures.")
    public void setContinueOnError( boolean isContinueOnError ) {
        this.isContinueOnErrorAttr=isContinueOnError;
    }

    private Device currentDevice=null;

    private TaskParameterValidator setFromPropsAndValidateBaseAttributes() throws BuildException {

        String url=urlAttr;
        if (url == null)
            // Legacy property name for the URL
            url=getPropertyUsingCamelCaseAlias( XML_MGM_URL_PROP_NAME );
        if (url == null)
            url=getPropertyUsingCamelCaseAlias( URL_PROP_NAME );
        if (url == null)
            url=getPropertyUsingCamelCaseAlias( HOST_PROP_NAME );

        String username=usernameAttr;
        if (username == null)
            username=getPropertyUsingCamelCaseAlias( USERNAME_PROP_NAME );

        String password=passwordAttr;
        if (password == null)
            password=getPropertyUsingCamelCaseAlias( PASSWORD_PROP_NAME );

        String domain=domainAttr;
        if (domain == null)
            domain=getPropertyUsingCamelCaseAlias( DOMAIN_PROP_NAME );

        Boolean isTrustAllCerts=isTrustAllCertsAttr;
        if (isTrustAllCerts == null)
            isTrustAllCerts=getBoolPropertyUsingCamelCaseAlias( TRUST_ALL_CERTS_PROP_NAME );
        if (isTrustAllCerts == null)
            isTrustAllCerts=true;

        Boolean isSave=this.isSaveAttr;
        if (isSave == null)
            isSave=getBoolPropertyUsingCamelCaseAlias( AUTO_SAVE_PROP_NAME );
        if (isSave == null)
            isSave=false;
        isSaveEffective=isSave;

        String schemaFile=this.schemaFile;
        if (schemaFile == null)
            schemaFile=getPropertyUsingCamelCaseAlias( SCHEMA_FILE_PROP_NAME );

        if (schemaFile != null)
            System.setProperty( DPSchemaValidator.DP_SCHEMA_PROP, schemaFile );

        Boolean isValidate=isValidateAttr;
        if (isValidate == null)
            isValidate=isValidateIfNotSet();
        if (isValidate == null)
            isValidate=getBoolPropertyUsingCamelCaseAlias( VALIDATE_PROP_NAME );
        if (isValidate == null)
            isValidate=false;

        Boolean isContinueOnError=isContinueOnErrorAttr;
        if (isContinueOnErrorAttr == null)
            isContinueOnError=getBoolPropertyUsingCamelCaseAlias( CONTINUE_ON_ERROR_PROP_NAME );
        if (isContinueOnError == null)
            isContinueOnError=false;
        isContinueOnErrorEffective=isContinueOnError;

        // Note: we have to initialize the global variable here
        validator=new TaskParameterValidator( this );
        validator.validateRequired( url, "url", "DataPower url for XML management interface" );
        validator.validateRequired( username, "username", "Valid DataPower user name" );

        if (!isDefaultDomainOverride()) {
            validator.validateRequired( domain, "domain", "Domain name for DataPower request" );
        }

        validator.validateRequired( password, "password", "Valid DataPower password" );
        if (envPrefix == null && validator.getErrors().getErrors().size() > 1) {
            logger.warn( "Environment prefix has not been provided" );
        }
        validator.throwIfErrors();

        currentDevice=createDevice( url, domain, username, DecryptionUtils.decryptIfEncrypted( password ),
                        isTrustAllCerts, isValidate );

        return validator;
    }

    protected Boolean isValidateIfNotSet() {
        return null;
    }

    protected TaskParameterValidator validator;

    protected void initProperties( PropRepository propRepository ) {

        propRepository.initialize( getProject() );

        allPropSet=propRepository.getAll();
        // TODO: refactor propRepository
        envPropSet=allPropSet;
    }

    private final static String LICENSE_REF_NAME="com.myarch.dpbuddy.license";

    protected void validateLicense() {
        Project project=getProject();
        Object licenseObj=project.getReference( LICENSE_REF_NAME );

        if (licenseObj == null) {
            License license;
            if (licenseFile == null) {
                setLicense( getPropertyUsingCamelCaseAlias( LICENSE_FILE_PROP_NAME ) );
                if (licenseFile!=null) {
                    logger.debug( LICENSE_FILE_PROP_NAME
                                + " was provided. We'll attempt to validate the license file specified using this property." );
                }
                
            }
            if (licenseFile == null) {
                setLicense( System.getenv( LICENSE_FILE_ENV_NAME ) );
                if (licenseFile!=null) {
                    logger.debug( LICENSE_FILE_ENV_NAME
                                + " was provided. We'll attempt to validate the license file specified using this environment variable." );
                }
                
            }
            if (licenseFile == null) {
                // See if it's on the classpath, e.g., under conf
                File licenseFileFile=FilePathUtils.getOptionalFileFromResource( DPBuddyLicenseManager.LICENSE_FILE_NAME);
                if (licenseFileFile!=null && licenseFileFile.exists()) {
                    licenseFile=licenseFileFile.getAbsolutePath();
                    logger.debug( "Found "+DPBuddyLicenseManager.LICENSE_FILE_NAME+" on the classpath at "+licenseFile );

                }
            }
                
            DPBuddyLicenseManager licenseManager=new DPBuddyLicenseManager();
            String dpBuddyLicenseFilePath=licenseFile;
            if (dpBuddyLicenseFilePath != null) {
                license=licenseManager.validateLicenseFromFile( new File( dpBuddyLicenseFilePath ) );
            }
            else {
                license=licenseManager.validateDefaultLocationOrTrial();
            }
            project.addReference( LICENSE_REF_NAME, license );
        }
    }

    // Absolute path to the local audit log file
    public final String DP_AUDIT_FILE_LOCAL_LOCATION_PROP_NAME="dp.audit.local.file";
    // Absolute path to the audit log on the device. May contain filesystem
    // prefix, otherwise defaults to local://
    public final String DP_AUDIT_FILE_DP_LOCATION_PROP_NAME="dp.audit.dp.file";
    // Flag to turn on DPBuddy audit. Default: False
    public final String DP_AUDIT_PROP_NAME="dp.audit";
    // Flag to turn off saving the audit log file to DataPower. Default: True
    public final String DP_AUDIT_SAVE_TO_DP_PROP_NAME="dp.audit.save.to.dp";

    // Domain where to save the audit file, e.g., "default". Defaults to the
    // current domain.
    public final String DP_AUDIT_FILE_DOMAIN_PROP_NAME="dp.audit.domain";

    private void initAuditEventHandler() {
        auditEventHandler=new AuditEventHandler();

        // JSON audit is always on
        auditEventHandler.setXmlAuditOn( isAuditOn() );
        if (isAuditOn()) {
            String localAuditLogFile=getPropertyUsingCamelCaseAlias( DP_AUDIT_FILE_LOCAL_LOCATION_PROP_NAME );
            // TODO: default in the eventHandler
            if (localAuditLogFile != null)
                auditEventHandler.setLocalXmlAuditLogFile( new File( localAuditLogFile ) );
            if (!isSaveAuditToDPOn())
                auditEventHandler.setDPAuditLogFile( null );
            else {
                String dpAuditLogFile=getPropertyUsingCamelCaseAlias( DP_AUDIT_FILE_DP_LOCATION_PROP_NAME );
                if (dpAuditLogFile != null)
                    auditEventHandler.setDPAuditLogFile( dpAuditLogFile );
            }

            String dpAuditDomain=getPropertyUsingCamelCaseAlias( DP_AUDIT_FILE_DOMAIN_PROP_NAME );
            if (dpAuditDomain != null)
                auditEventHandler.setAuditLogFileDomain( dpAuditDomain );
        }

    }

    public final String DP_RELEASE_INFO="dp.release.info";
    public final String DP_LOCAL_USER_ID="dp.local.user";

    protected String getPropertyUsingCamelCaseAlias( String name ) {
        String val=envPropSet.get( name );
        if (val == null)
            val=envPropSet.get( AntUtils.toCamelCase( name ) );
        return val;

    }

    protected Boolean getBoolPropertyUsingCamelCaseAlias( String name ) {
        Boolean bool=null;
        String val=getPropertyUsingCamelCaseAlias( name );
        if (val != null)
            bool=Project.toBoolean( val );
        return bool;
    }

    protected String getReleaseInfoForAudit() {
        return getPropertyUsingCamelCaseAlias( DP_RELEASE_INFO );
    }

    protected String getLocalUserIdForAudit() {
        String userId=getPropertyUsingCamelCaseAlias( DP_LOCAL_USER_ID );
        if (StringUtils.isBlank( userId )) {
            userId=System.getProperty( "user.name" );
        }

        return userId;
    }

    private boolean isAuditOn() {
        String dpAuditOnOff=getPropertyUsingCamelCaseAlias( DP_AUDIT_PROP_NAME );
        if (dpAuditOnOff == null)
            return false;

        return Project.toBoolean( dpAuditOnOff );
    }

    private boolean isSaveAuditToDPOn() {
        String dpAuditOnOff=getPropertyUsingCamelCaseAlias( DP_AUDIT_SAVE_TO_DP_PROP_NAME );
        if (dpAuditOnOff == null)
            return true;

        return Project.toBoolean( dpAuditOnOff );
    }

    protected AuditEvent createAuditEvent( String action ) {
        AuditEvent auditEvent=new AuditEvent( getDevice(), action );
        auditEvent.setReleaseInfo( getReleaseInfoForAudit() );
        auditEvent.setLocalUserName( getLocalUserIdForAudit() );
        // FIXME: should go to the device?
        auditEvent.setEnvironmentLabel( getEnvironmentLabel() );

        return auditEvent;
    }

    private static final String ENVIRONMENT_LABEL_PROP_NAME="environment";

    protected String getEnvironmentLabel() {
        // TODO: accommodate tags as well
        // for now it's just the environment
        String environment=getProject().getProperty( ENVIRONMENT_LABEL_PROP_NAME );

        return environment;
    }

    public static boolean isGradleLogging( Project project ) {
        String gradleLoggingString=project.getProperty( GRADLE_LOGGING_PROP_NAME );
        boolean isGradleLogging=false;
        if (gradleLoggingString != null)
            isGradleLogging=Project.toBoolean( gradleLoggingString );

        return isGradleLogging;
    }

    public static void setUpLogging( Project project ) {
        AntCommonsLogger.setProject( project );
        if (isGradleLogging( project ))
            AntCommonsLogger.setLoggerConfiguration( LoggerConfiguration.GRADLE_CONFIG );
    }

    public static String getDPBuddyVersion() {
        String dpbuddyVersion=BaseDPBuddyTask.class.getPackage().getImplementationVersion();
        return dpbuddyVersion;
    }

    @Override
    public void execute() throws BuildException {
        extensionInitializer=new ExtensionInitializer( getProject() );
        setUpLogging( getProject() );

        logger.debug( "DPBuddy version: " + getDPBuddyVersion() );

        initProperties( extensionInitializer.instantiatePropRepository() );
        deviceFactory=extensionInitializer.instantiateDeviceFactory();

        propExpressionEvaluator=envPropSet.getPropExpressionEvaluator();
        fileTransformer=new FileTransformer( getProject(), envPropSet.getPropExpressionEvaluator() );

        try {
            validateLicense();
        }
        catch (LicenseException le) {
            throw new BuildException( le.getMessage(), le );
        }

        savedEnvPrefix=EnvironmentAntTask.getCurrentEnvPrefix( getProject() );
        if (envPrefix == null)
            setEnv( getPropertyUsingCamelCaseAlias( ENV_PARAM_PROP_NAME ) );

        if (envPrefix == null)
            setEnv( getPropertyUsingCamelCaseAlias( ENV_PREFIX_PROP_NAME ) );
        if (envPrefix == null)
            setEnv( getPropertyUsingCamelCaseAlias( ENV_PROP_NAME ) );
        if (envPrefix == null)
            setEnv( getPropertyUsingCamelCaseAlias( DEVICE_PROP_NAME ) );

        // if env prefix is a variable
        if (envPrefix != null) {

            String envPrefixVarValue=getProject().getProperty( envPrefix );
            if (envPrefixVarValue != null && envPrefixVarValue.contains( "," ) && !isHoconConfig( envPrefixVarValue )) {
                logger.debug( "Environment is pointing to the variable. The environment will be set to the value of this variable, which is "
                    + envPrefixVarValue );
                setEnv( envPrefixVarValue );
            }
        }

        logger.debug( "Environment was set to: " + envPrefix );

        try {
            if (envPrefix != null && envPrefix.contains( "," )) {

                String[] envPrefixes=envPrefix.trim().split( "\\s*,\\s*" );
                executeTaskForMultipleEnvs( envPrefixes );
            }
            else {
                executeTaskForSingleEnv();
            }

        }
        finally {
            // restore only if different
            if (savedEnvPrefix != null && envPrefix != null && !savedEnvPrefix.equalsIgnoreCase( envPrefix )) {
                logger.debug( "Restoring env prefix '" + savedEnvPrefix + "'" );
                setEnvironmentPropertyProvider( savedEnvPrefix );
            }

            deregisterLogger();
        }
    }

    private boolean isHoconConfig( String configStr ) {
        return StringUtils.containsAny( configStr, '{', '}' );

    }

    private void executeTaskForMultipleEnvs( String[] envPrefixes ) {
        if (isLogStartAndCompletion) {
            logger.info( String.format( "Executing '%s' against multiple devices '%s' ...", getTaskName(),
                            StringUtils.join( envPrefixes, "," ) ) );
        }

        for (String envPrefix : envPrefixes)
            validatePrefixDefined( envPrefix );

        List<String> successfulEnvs=new ArrayList<String>();
        List<Pair<String, Exception>> failures=new ArrayList<Pair<String, Exception>>();
        onStart();
        
        for (String envPrefix : envPrefixes) {
            setEnvironmentPropertyProvider( envPrefix );
            try {
                executeTaskInEnvContext();
                successfulEnvs.add( envPrefix );
                if (isLogStartAndCompletion) {
                    logger.info( String.format( "Successfully executed '%s' against the device '%s'.\n", getTaskName(),
                                    envPrefix ) );
                }
            }
            catch (Exception e) {
                failures.add( new ImmutablePair<String, Exception>( envPrefix, e ) );
                logger.info( String.format( "'%s' failed for the device '%s' with the following error:\n%s",
                                getTaskName(), envPrefix, exceptionToString( e ) ) );
                if (!isContinueOnErrorEffective) {
                    break;
                }
            }
        }

        if (failures.size() > 0) {
            if (isRollbackAllOnError && successfulEnvs.size() > 0) {
                logger.info( String.format(
                                "\nReverting devices/domains '%s' because of failures with other devices/domains ...",
                                StringUtils.join( successfulEnvs, "," ) ) );

                rollbackEnvs( successfulEnvs, failures );
                successfulEnvs.clear();
            }

            handleMultiEnvFailure( failures, successfulEnvs );
        }
        else {
            onCompletion();
            logSuccessfulCompletion( successfulEnvs );
        }
    }

    private void executeTaskForSingleEnv() {

        if (envPrefix != null) {
            setEnvironmentPropertyProvider( envPrefix );
        }
        onStart();
        try {
            String currEnvPrefix=EnvironmentAntTask.getCurrentEnvPrefix( getProject() );
            if (StringUtils.isNotBlank( currEnvPrefix )) {
                validatePrefixDefined( currEnvPrefix );
            }
            executeTaskInEnvContext();
            onCompletion();
            logSuccessfulCompletion( null );
        }
        catch (Exception e) {
            if (e instanceof BuildException) {
                throw (BuildException) e;
            }
            else {
                // we're wrapping in DPBuddyException to force Ant not to dump
                // the error stack for certain exceptions in non-verbose mode
                throw new BuildException( exceptionToString( e ), new DPBuddyException( e ) );
            }
        }

    }

    protected void onCompletion() {
        
    }
    
    protected void onStart() {
        
    }    
    
    private boolean isRollbackOnError=false;
    private boolean isRollbackAllOnError=false;

    @Option(name="-rollbackOnError", handler=BoolOptionHandler.class,
                    usage="Rollback to the last checkpoint (if exists) in case of any errors.")
    public void setRollbackOnError( boolean isRollbackOnError ) {
        this.isRollbackOnError=isRollbackOnError;
    }

    @Option(name="-rollbackAllOnError", handler=BoolOptionHandler.class,
                    usage="Rollback all devices to the last checkpoint (if exists) in case if the command failed for at least one other device in the group.")
    public void setRollbackAllOnError( boolean isRollbackAllOnError ) {
        this.isRollbackAllOnError=isRollbackAllOnError;
    }

    private void performRollback() {

        RollbackCheckpointCommand rollbackCmd=new RollbackCheckpointCommand();
        rollbackCmd.execute( getDevice() );
        saveDomainConfigIfEnabled();

        logger.info( "Successfully reverted configuration." );

    }

    /**
     * Rollback all envs in the list
     * 
     * @param envs
     */
    private void rollbackEnvs( List<String> envs, List<Pair<String, Exception>> failures ) {
        for (String envPrefix : envs) {
            setEnvironmentPropertyProvider( envPrefix );
            try {
                logger.info( String.format( "Reverting configuration of url/host '%s', domain '%s' ...",
                                getDevice().getDPConnection().getHost(), getActualDomain() ) );
                performRollback();
            }
            catch (Exception e) {
                failures.add( new ImmutablePair<String, Exception>( envPrefix, e ) );
                logger.info( String.format( "Rollback for the device '%s' failed with the following error:\n%s",
                                envPrefix, exceptionToString( e ) ) );
            }
        }
    }

    private void handleMultiEnvFailure( List<Pair<String, Exception>> failures, List<String> successfulEnvs ) {

        StringBuilder errorMsg=new StringBuilder(
                        String.format( "'%s' failed with the following error(s):\n", getTaskName() ) );
        for (Pair<String, Exception> failure : failures) {
            Exception e=failure.getRight();
            errorMsg.append( String.format( "Device %s: %s\n", failure.getLeft(), exceptionToString( e ) ) );

            String stackTrace=ExceptionUtils.getStackTrace( e );
            logger.debug( stackTrace );
        }
        if (successfulEnvs.size() > 0)
            errorMsg.append( String.format( "Note that the command succeeded for the following device(s): '%s'\n",
                            StringUtils.join( successfulEnvs, "'" ) ) );

        throw new BuildException( errorMsg.toString() );

    }

    private void executeTaskInEnvContext() {
        // reset all attributes with the (potentially) new environment
        AntUtils.reconfigureWithNested( this );
        initAuditEventHandler();

        setFromPropsAndValidateBaseAttributes();

        logStartTask();

        printMemory();

        try {
            if (isMultiDomain()) {
                Device device=getDevice();
                Set<String> domains=fetchDomainsForMultiDomainCommand();
                for (String domain : domains) {
                    currentDevice=device.switchDomain( domain );
                    logDomain();
                    executeDPTask();
                }
            }
            else {
                executeDPTask();
            }
        }
        catch (DPBuddyException e) {
            if (e.isRollback() && isRollbackOnError) {
                logger.info( String.format(
                                "'%s' failed with the error:\n%s\nWill try to revert the configuration to the last checkpoint.",
                                getTaskName(), exceptionToString( e ) ) );
                try {
                    performRollback();
                }
                catch (Exception rollbackException) {
                    logger.warn( "Rollback failed with the error:\n" + rollbackException.getMessage() );
                }
            }
            throw e;
        }
    }

    private Set<String> fetchDomainsForMultiDomainCommand() {
        String domainPatternsStr=getDomainPatterns();
        Device device=getDevice();
        Set<String> domains=device.fetchRequiredDomains( StringPatternUtils.stringToPattnernList( domainPatternsStr ) );
        if (!StringUtils.containsIgnoreCase( domainPatternsStr, Device.DEFAULT_DOMAIN_NAME )) {
            domains=Device.removeDefaultDomain( domains );
        }

        return domains;
    }

    private boolean isMultiDomain() {
        return getDomainPatterns() != null;
    }
    
    protected boolean isSuppressDomainLogging() {
        return isMultiDomain();
    }    

    protected String getDomainPatterns() {
        return null;
    }

    private void validatePrefixDefined( String envPrefixParam ) {
        if (!EnvironmentAntTask.isEnvironmentPrefixValid( getProject(), envPrefixParam )) {
            throw new BuildException( "You have not defined any configuration or properties that start with the path '"
                + envPrefixParam + "'" );
        }
    }

    private void setEnvironmentPropertyProvider( String envPrefix ) {
        envPrefix=envPrefix.trim();
        currentEnvPrefix=envPrefix;
        EnvironmentAntTask.setEnvironmentPropertyProvider( getProject(), envPrefix );
    }

    protected abstract void executeDPTask();

    protected Device getDevice() {
        return currentDevice;
    }

    private Device createDevice( String url, String domain, String username, String password, boolean isTrustAllCerts,
                    boolean isValidate ) {
        DPConnection connection=new DPConnection( url, domain, username, password );
        if (currentEnvPrefix != null)
            connection.setDeviceName( currentEnvPrefix );
        connection.setSchemaValidation( isValidate );
        connection.setTrustAllCerts( isTrustAllCerts );
        Device device=deviceFactory.getDevice( connection );

        return device;
    }

    protected void executeRequest( DPCommand<?> command ) {
        command.execute( getDevice() );
    }

    protected void saveDomainConfigIfEnabled() {
        if (isSaveEffective) {
            saveDomainConfig();
        }
    }

    protected void saveDomainConfig() {
        Device device=getDevice();
        if (isDefaultDomainOverride() ) {
            device=getDevice().switchDefaultDomain();
        }
        
        logger.info( String.format("Saving configuration of the domain '%s' ...", device.getCurrentDomain() ));
        ActionCommand saveActionCommand=new ActionCommand( "SaveConfig" );
        saveActionCommand.execute(device);
    }

    protected boolean isSave() {
        return isSaveEffective;
    }

    private boolean isLogStartAndCompletion=true;

    public void setLogStartAndCompletion( boolean isLogStartAndCompletion ) {
        this.isLogStartAndCompletion=isLogStartAndCompletion;
    }

    private void logStartTask() {
        if (isLogStartAndCompletion) {
            String envInfo="";
            if (StringUtils.isNotBlank( currentEnvPrefix )) {
                envInfo=" Device/env: '" + currentEnvPrefix + "', ";
            }
            String domainInfo="";
            if (!isSuppressDomainLogging()) {
                domainInfo=", domain: '" + getActualDomain() + "'";
            }
            logger.info( String.format( "Executing '%s': %surl/host: '%s'%s ...", getTaskName(), envInfo,
                            getDevice().getDPConnection().getHost(), domainInfo ) );
        }
    }

    private void logDomain() {
        if (isLogStartAndCompletion) {
            logger.info( String.format( "Domain '%s' ...", getActualDomain() ) );
        }
    }

    public String getActualDomain() {
        String actualDomain=getDevice().getDPConnection().getDomain();
        if (isDefaultDomainOverride()) {
            actualDomain=Device.DEFAULT_DOMAIN_NAME;
        }

        return actualDomain;
    }

    protected boolean isDefaultDomainOverride() {
        return false;
    }

    private void logSuccessfulCompletion( List<String> successfulEnvs ) {

        if (!isLogStartAndCompletion) {
            return;
        }

        if (successfulEnvs == null || successfulEnvs.size() <= 1) {
            // message for successful completion of a task against one
            // environment
            logger.info( String.format( "Successfully executed '%s'.", getTaskName() ) );
        }
        else {
            String successMsg=String.format( "'%s' was successfully executed against %d devices: '%s'", getTaskName(),
                            successfulEnvs.size(), StringUtils.join( successfulEnvs, "," ) );
            logger.info( successMsg );
        }

    }

    protected void setDeploymentPolicy( DeploymentPolicyRequest deploymentPolicyRequest, File deploymentPolicyFile,
                    String deploymentPolicyName, String deploymentPolicyVariablesName ) {
        if (deploymentPolicyFile != null) {
            if (!deploymentPolicyFile.exists())
                throw new BuildException(
                                String.format( "Deployment policy file '%s' provided to the task '%s'does not exist",
                                                deploymentPolicyFile.getAbsolutePath(), getTaskName() ) );

            XMLInput deplPolicy=new XMLInput( deploymentPolicyFile );

            deplPolicy.resolveVars( envPropSet.getPropExpressionEvaluator() );

            logger.info( "Using deployment policy " + deploymentPolicyFile );
            deploymentPolicyRequest.setDeploymentPolicyFile( deplPolicy );
        }

        if (deploymentPolicyName != null)
            deploymentPolicyRequest.setDeploymentPolicyName( deploymentPolicyName );

        if (deploymentPolicyVariablesName != null)
            deploymentPolicyRequest.setDeploymentPolicyVariablesName( deploymentPolicyVariablesName );

    }

    protected void printMemory() {
        int mb=1024 * 1024;
        Runtime runtime=Runtime.getRuntime();
        logger.debug( "Max JVM Memory:" + runtime.maxMemory() / mb );
    }

    private String exceptionToString( Exception e ) {
        String message=e.toString(); // this will contain the exception class
                                     // name
        if (e instanceof DPBuddyException) {
            Device device=getDevice();
            if (device != null) {
                message=String.format( "DataPower host: '%s', domain: '%s'. %s ", device.getDPConnection().getHost(),
                                device.getCurrentDomain(), e.getMessage() );
            }
            else
                message=e.getMessage();
        }
        else if (e instanceof DPConnectionException) {
            message=e.getMessage();
        }

        Throwable rootCause=ExceptionUtils.getRootCause( e );
        if (rootCause != null) {
            message+="\nCause: " + rootCause.getMessage();
        }

        return message;
    }

}
