package com.myarch.dpbuddy.ant;

import java.io.IOException;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.Ant;
import org.apache.tools.ant.taskdefs.Property;
import org.apache.tools.ant.types.PropertySet;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.environment.EnvironmentAntTask;
import com.myarch.hocon.HoconAntUtils;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigObject;

public class ExecTargetTask extends BaseDPBuddyTask {

    private CustomizedAnt callee;
    // must match the default value of Ant#inheritAll
    private boolean inheritAll=true;
    // must match the default value of Ant#inheritRefs
    private boolean inheritRefs=false;

    private boolean targetSet=false;

    /**
     * If true, pass all properties to the new Ant project. Defaults to true.
     * 
     * @param inherit
     *            <code>boolean</code> flag.
     */
    public void setInheritAll( boolean inherit ) {
        inheritAll=inherit;
    }

    /**
     * If true, pass all references to the new Ant project. Defaults to false.
     * 
     * @param inheritRefs
     *            <code>boolean</code> flag.
     */
    public void setInheritRefs( boolean inheritRefs ) {
        this.inheritRefs=inheritRefs;
    }

    /**
     * Initialize this task by creating new instance of the ant task and
     * configuring it by calling its own init method.
     */
    public void init() {
        callee=new CustomizedAnt( this );
        callee.init();
    }


    @SuppressWarnings("unused")
    private String scopePath;
    
    
    
    public void setScope( String scopePath ) {
        this.scopePath=scopePath;
    }

    @Override
    public void executeDPTask() throws BuildException {
        if (callee == null) {
            init();
        }

        if (!targetSet) {
            throw new BuildException( "Attribute target is required.", getLocation() );
        }
        
        callee.setAntfile( getProject().getProperty( "ant.file" ) );
        callee.setInheritAll( inheritAll );
        callee.setInheritRefs( inheritRefs );

        String savedEnv=getProject().getProperty( ENV_PROP_NAME );
        if (savedEnv==null) {
            savedEnv=getPropertyUsingCamelCaseAlias(ENV_PROP_NAME);
        }
        String prefix=EnvironmentAntTask.getCurrentEnvPrefix( getProject() );
        // FIXME -- add scope to the path
//        if (scopePath!=null) {
//            prefix+=HoconUtils.PATH_SEPARATOR+scopePath;
//        }
        
        callee.getProject().setUserProperty( ENV_PROP_NAME, prefix );
        
        logger.debug( "Property "+ ENV_PROP_NAME+" was set to "+prefix+" before execution of the target ");
        try{
            callee.execute();
        }
        catch (DPBuddyException e) {
            e.setRollback( true );
            throw e;
        }
        finally {
            if (savedEnv==null) {
                // there is no easy way to "unset" the property, but setting it to "" disables property resolution
                savedEnv="";
            }
            getProject().setUserProperty( ENV_PROP_NAME, savedEnv );
            getProject().setUserProperty(getPropertyUsingCamelCaseAlias(ENV_PROP_NAME),savedEnv);
        }
    }

    // not in use
    @SuppressWarnings("unused")
    private void setConfigScope(String scopePath) {
        Config config=HoconAntUtils.getConfig(getProject());
        ConfigObject scope=config.getObject( scopePath ); 
        HoconAntUtils.pushConfig( getProject(), scope.toConfig() );
    }
    
    /**
     * Create a new Property to pass to the invoked target(s).
     * 
     * @return a <code>Property</code> object.
     */
    public Property createParam() {
        if (callee == null) {
            init();
        }
        return callee.createProperty();
    }

    /**
     * Reference element identifying a data type to carry over to the invoked
     * target.
     * 
     * @param r
     *            the specified <code>Ant.Reference</code>.
     * @since Ant 1.5
     */
    public void addReference( Ant.Reference r ) {
        if (callee == null) {
            init();
        }
        callee.addReference( r );
    }

    /**
     * Set of properties to pass to the new project.
     * 
     * @param ps
     *            the <code>PropertySet</code> to pass.
     * @since Ant 1.6
     */
    public void addPropertyset( PropertySet ps ) {
        if (callee == null) {
            init();
        }
        callee.addPropertyset( ps );
    }

    /**
     * Set target to execute.
     * 
     * @param target
     *            the name of the target to execute.
     */
    public void setTarget( String target ) {
        if (callee == null) {
            init();
        }
        callee.setTarget( target );
        targetSet=true;
    }

    /**
     * Add a target to the list of targets to invoke.
     * 
     * @param t
     *            <code>Ant.TargetElement</code> representing the target.
     * @since Ant 1.6.3
     */
    public void addConfiguredTarget( Ant.TargetElement t ) {
        if (callee == null) {
            init();
        }
        callee.addConfiguredTarget( t );
        targetSet=true;
    }

    /**
     * Handles output. Send it the the new project if is present, otherwise call
     * the super class.
     * 
     * @param output
     *            The string output to output.
     * @see Task#handleOutput(String)
     * @since Ant 1.5
     */
    // when the output is routed to the new project, it doesn't print anything.
/*    
    public void handleOutput( String output ) {


        if (callee != null) {
            callee.handleOutput( output );
        }
        else {
            super.handleOutput( output );
        }
    }
*/    

    /**
     * Handles input. Delegate to the created project, if present, otherwise call
     * the super class.
     * 
     * @param buffer
     *            the buffer into which data is to be read.
     * @param offset
     *            the offset into the buffer at which data is stored.
     * @param length
     *            the amount of data to read.
     *
     * @return the number of bytes read.
     *
     * @exception IOException
     *                if the data cannot be read.
     * @see Task#handleInput(byte[], int, int)
     * @since Ant 1.6
     */
    public int handleInput( byte[] buffer, int offset, int length ) throws IOException {
        if (callee != null) {
            return callee.handleInput( buffer, offset, length );
        }
        return super.handleInput( buffer, offset, length );
    }

    /**
     * Handles output. Send it the the new project if is present, otherwise call
     * the super class.
     * 
     * @param output
     *            The string to output.
     * @see Task#handleFlush(String)
     * @since Ant 1.5.2
     */
    public void handleFlush( String output ) {
        if (callee != null) {
            callee.handleFlush( output );
        }
        else {
            super.handleFlush( output );
        }
    }

    /**
     * Handle error output. Send it the the new project if is present, otherwise
     * call the super class.
     * 
     * @param output
     *            The string to output.
     *
     * @see Task#handleErrorOutput(String)
     * @since Ant 1.5
     */
    public void handleErrorOutput( String output ) {
        if (callee != null) {
            callee.handleErrorOutput( output );
        }
        else {
            super.handleErrorOutput( output );
        }
    }

    /**
     * Handle error output. Send it the the new project if is present, otherwise
     * call the super class.
     * 
     * @param output
     *            The string to output.
     * @see Task#handleErrorFlush(String)
     * @since Ant 1.5.2
     */
    public void handleErrorFlush( String output ) {
        if (callee != null) {
            callee.handleErrorFlush( output );
        }
        else {
            super.handleErrorFlush( output );
        }
    }

}
