package com.myarch.dpbuddy.audit;

import java.net.MalformedURLException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.*;

import org.apache.commons.lang.StringUtils;
import org.jdom2.Element;

import com.google.gson.JsonArray;
import com.google.gson.JsonObject;
import com.myarch.dpbuddy.*;
import com.myarch.dpbuddy.event.EventCreator;
import com.myarch.dpbuddy.event.EventUtils;
import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.eventmodel.Event;
import com.myarch.eventmodel.EventType;
import com.myarch.util.JsonUtils;

public class AuditEvent {

    public final static String TIMESTAMP_FORMAT="yyyy-MM-dd HH:mm:ss";

    public final static String LOCAL_USERNAME_NAME="localUserName";
    public final static String RELEASE_INFO="releaseInfo";
    public final static String TIMESTAMP_NAME="timestamp";

    public final static String DP_URL_NAME="dpUrl";
    public final static String DP_USERNAME_NAME="dpUserName";
    public final static String DEVICENAME_NAME="deviceName";
    public final static String DOMAIN_NAME="domain";
    public final static String ACTION_NAME="action";
    public final static String AFFECTED_OBJECTS_NAME="objects";

    private String localUserName;
    private Date timestamp;
    private String releaseInfo;

    private DPConnection dpConnection;
    private Device device;

    private String action;
    private String environmentLabel;

    private List<? extends DPObject> objects=new ArrayList<DPObject>();

    private List<AuditEventFile> auditEventFiles=new ArrayList<AuditEventFile>();
    private AuditEventFile archiveFile;

    public AuditEvent() {
        timestamp=new Date();
        localUserName=System.getProperty( "user.name" );
    }

    public AuditEvent(Device device, String action) {
        this();
        this.device=device;
        // TODO: get rid of connection
        this.dpConnection=device.getDPConnection();
        this.action=action;
    }

    public DPConnection getDPConnection() {
        return dpConnection;
    }

    public void setObjects( List<? extends DPObject> dpObjects ) {
        objects=dpObjects;
    }
    
    public void setObjects( Set<? extends DPObject> dpObjects ) {
        objects=new ArrayList<DPObject>(dpObjects);
    }
    public void addFile( AuditEventFile auditEventFile ) {
        auditEventFiles.add( auditEventFile );
    }

    public void setArchiveFile( AuditEventFile archiveFile ) {
        this.archiveFile=archiveFile;
    }

    public void setReleaseInfo( String releaseInfo ) {
        this.releaseInfo=releaseInfo;
    }

    public String getEnvironmentLabel() {
        return environmentLabel;
    }

    public void setEnvironmentLabel( String environmentLabel ) {
        this.environmentLabel=environmentLabel;
    }

    public Element toXMLElement() {
        Element eventElt=new Element( "event" );
        toXmlOrJson( eventElt, null );

        return eventElt;
    }

    // old JSON format, to delete
    @SuppressWarnings("unused")
    private JsonObject toJson() {
        JsonObject jsonObj=new JsonObject();
        toXmlOrJson( null, jsonObj );

        return jsonObj;
    }

    private void toXmlOrJson( Element eventElt, JsonObject eventJsonObj ) {

        String isoTimestampString=JsonUtils.dateToIsoStr( timestamp );
        addToContainer( null, eventJsonObj, TIMESTAMP_NAME, isoTimestampString );
        // Using the custom format for XML only for compatibility
        String timestampString=new SimpleDateFormat( TIMESTAMP_FORMAT ).format( timestamp );
        addToContainer( eventElt, null, TIMESTAMP_NAME, timestampString );

        addToContainer( eventElt, eventJsonObj, ACTION_NAME, action );
        addToContainer( eventElt, eventJsonObj, LOCAL_USERNAME_NAME, localUserName );
        if (releaseInfo != null) {
            addToContainer( eventElt, eventJsonObj, RELEASE_INFO, releaseInfo );
        }

        if (dpConnection == null)
            throw new IllegalStateException( "dpConnection must be set, it is null!" );
        if (dpConnection.getDeviceName() != null) {
            addToContainer( eventElt, eventJsonObj, DEVICENAME_NAME, dpConnection.getDeviceName() );
        }
        addToContainer( eventElt, eventJsonObj, DP_URL_NAME, dpConnection.getUrl() );
        addToContainer( eventElt, eventJsonObj, DOMAIN_NAME, dpConnection.getDomain() );
        addToContainer( eventElt, eventJsonObj, DP_USERNAME_NAME, dpConnection.getUsername() );

        if (objects.size() > 0) {

            if (eventElt != null) {
                Element affectedObjectElt=new Element( AFFECTED_OBJECTS_NAME );
                eventElt.addContent( affectedObjectElt );
                for (DPObject dpObject : objects) {
                    affectedObjectElt.addContent( dpObject.toXML() );
                }
            }

            if (eventJsonObj != null) {
                JsonArray arr=new JsonArray();
                for (DPObject dpObject : objects) {
                    arr.add( dpObject.toJson() );
                }
                eventJsonObj.add( AFFECTED_OBJECTS_NAME, arr );
            }
        }

        if (auditEventFiles.size() > 0) {
            if (eventElt != null) {
                Element filesElt=new Element( "files" );
                eventElt.addContent( filesElt );
                for (AuditEventFile auditEventFile : auditEventFiles) {
                    filesElt.addContent( auditEventFile.toXML() );
                }
            }

            if (eventJsonObj != null) {
                JsonArray arr=new JsonArray();
                for (AuditEventFile auditEventFile : auditEventFiles) {
                    arr.add( auditEventFile.toJson() );
                }
                eventJsonObj.add( "files", arr );
            }
        }
    }

    // TODO: update_id
    public List<Event> asJsonEvents() {
        List<Event> events=new ArrayList<>();

        String changeId=genUniqueId();

        if (objects.size() > 0) {
            for (DPObject dpObject : objects) {
                // TODO: reuse
                Event event=createAuditJsonEvent(changeId);
                JsonObject dpJson=EventUtils.getDpJson( event );
                dpObject.populateJson( dpJson );
                events.add( event );
            }
        }
        if (auditEventFiles.size() > 0) {
            for (AuditEventFile auditEventFile : auditEventFiles) {
                Event event=createAuditJsonEvent(changeId);
                // TODO: underscore notation for files
                event.json().add( "file", auditEventFile.toJson() );
                events.add( event );
            }
        }

        if (objects.size() == 0 && auditEventFiles.size() == 0) {
            // event that did not affect any objects or files, just a single
            // action
            Event event=createAuditJsonEvent(changeId);
            events.add( event );
        }

        return events;
    }

    private String genUniqueId() {
        String id=StringUtils.remove( UUID.randomUUID().toString(), '-');
        
        return id;
    }
    
    private Event createAuditJsonEvent(String changeId) {
        EventCreator eventCreator=new EventCreator();
        Event event=eventCreator.createDataPowerEvent( device, device.getCurrentDomain(), EventType.AUDIT_CHANGE );
        event.setType( "dp.audit" );

        JsonObject json=event.json();
        json.addProperty( "change_id", changeId );
        json.addProperty( "action", action );
        json.addProperty( "user", dpConnection.getUsername() );
        if (localUserName != null) {
            json.addProperty( "local_user", localUserName );
        }
        if (releaseInfo != null) {
            json.addProperty( "release_info", releaseInfo );
        }
        if (environmentLabel != null) {
            json.addProperty( "environment", environmentLabel );
        }
        if (archiveFile != null) {
            json.add( "archive_file", archiveFile.toJson() );
        }

        return event;

    }

    private void addToContainer( Element parentElt, JsonObject parentJsonObj, String name, String val ) {
        if (parentElt != null) {
            addElt( parentElt, name, val );
        }
        if (parentJsonObj != null) {
            parentJsonObj.addProperty( name, val );
        }

    }

    private Element addElt( Element parent, String name, String text ) {
        return JDomUtils.addOrUpdateElement( parent, name, text );
    }

    public boolean urlEquals( String url ) {
        URL urlToCompareWith;
        try {
            urlToCompareWith=new URL( url );
        }
        catch (MalformedURLException e) {
            throw new DPBuddyException( "Malformed URL %s", e, url );
        }

        URL eventUrl;
        try {
            eventUrl=new URL( dpConnection.getUrl() );
        }
        catch (MalformedURLException e) {
            throw new DPBuddyException( "Audit event contains malformed URL %s !", e, dpConnection.getUrl() );
        }

        return eventUrl.getHost().equalsIgnoreCase( urlToCompareWith.getHost() );
    }

    public boolean domainEquals( String domain ) {
        return dpConnection.getDomain().equalsIgnoreCase( domain );
    }

    public void setLocalUserName( String localUserName ) {
        this.localUserName=localUserName;
    }

    public static AuditEvent fromXMLElement( Element eventElt ) {

        String localUserName=JDomUtils.getRequiredChildElementValue( eventElt, LOCAL_USERNAME_NAME );
        String dpUsername=JDomUtils.getRequiredChildElementValue( eventElt, DP_USERNAME_NAME );
        String dpUrl=JDomUtils.getRequiredChildElementValue( eventElt, DP_URL_NAME );
        String deviceName=eventElt.getChildText( DEVICENAME_NAME );
        String domain=JDomUtils.getRequiredChildElementValue( eventElt, DOMAIN_NAME );
        String action=JDomUtils.getRequiredChildElementValue( eventElt, ACTION_NAME );

        DPConnection dpConnection=new DPConnection( deviceName, dpUrl, domain, dpUsername, null );
        Device device=DeviceFactoryImpl.newInstance().getDevice( dpConnection );
        AuditEvent event=new AuditEvent( device, action );
        event.setLocalUserName( localUserName );

        /* FIXME: complete
        Element affectedObjectsElt=eventElt.getChild(AFFECTED_OBJECTS_NAME);
        if (affectedObjectsElt!=null) {
            
            List<? extends DPObject> affectedObjects=new ArrayList<DPObject>();
            
        }
        */

        // FIXME: set timestamp, currently the timestamp is the current
        // timestamp.

        return event;
    }

}
