package com.myarch.dpbuddy.backup;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPConst;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.export.ExportBackupBase;
import com.myarch.dpbuddy.filemanagement.FileResponse;
import com.myarch.dpbuddy.xmlutil.FileSavingStreamParser;
import com.myarch.util.StringPatternUtils;

/**
 * Exports configuration and files from the device. Represents "do-export" SOMA
 * operation.
 */
public class BackupCommand extends ExportBackupBase implements DPCommand<FileResponse> {

    private final Log logger=LogFactory.getLog( getClass() );

    private List<Pattern> domainPatterns=new ArrayList<Pattern>();
    private boolean isDomainMatchRequired=true;

    public BackupCommand() {
        super( "do-backup" );
    }

    private void addDomain( String name ) {
        Element domainElement=new Element( "domain", DPConst.DP_NS );

        domainElement.setAttribute( "name", name );
        getOperationElement().addContent( domainElement );
    }

    public void addDomainPatterns( String namePatterns ) {
        domainPatterns.addAll( StringPatternUtils.stringToPattnernList( namePatterns ) );
    }

    public void addDomainPatterns( List<Pattern> namePatterns ) {
        this.domainPatterns.addAll( namePatterns );
    }

    public void addDomainPattern( Pattern namePattern ) {
        domainPatterns.add( namePattern );
    }

    public void setDomainMatchRequired( boolean isDomainMatchRequired ) {
        this.isDomainMatchRequired=isDomainMatchRequired;
    }

    private void removeAllDomains() {
        getOperationElement().removeChildren( "domain", DPConst.DP_JDOM_NS );
    }

    public List<String> getDomains() {
        List<Element> domainElts=getOperationElement().getChildren( "domain", DPConst.DP_JDOM_NS );
        List<String> domains=new ArrayList<String>();
        for (Element domainElt : domainElts) {
            domains.add( domainElt.getAttributeValue( "name" ) );
        }
        return domains;
    }

    public FileResponse execute( Device device ) {
        if (getToFile() == null)
            throw new DPBuddyException( "You must provide a file to save to save the backup to" );

        if (domainPatterns.size() == 0) {
            addDomainPatterns( device.getCurrentDomain() );
        }
        logger.info( "Querying the device to determine domains to backup ..." );
        removeAllDomains();
        Set<String> domains=device.fetchDomains( domainPatterns, isDomainMatchRequired );
        if (domains.size()==0)
            return null;
        
        for (String domain : domains) {
            addDomain( domain );
        }
            

        logger.info( "Performing backup of the following domain(s): " + getDomains() );
        FileSavingStreamParser parser=new FileSavingStreamParser( new File( getToFile() ) );
        Document responseDoc=device.executeRequest( parser, this );

        FileResponse response=new FileResponse( responseDoc );
        response.validate();
        logger.info( "Backup was saved to '" + getToFile() + "'" );
        return response;
    }

}
