package com.myarch.dpbuddy.backup;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.tools.ant.BuildException;
import org.jdom2.Document;
import org.jdom2.Element;
import org.kohsuke.args4j.Option;
import org.kohsuke.args4j.spi.ExplicitBooleanOptionHandler;

import com.myarch.antutil.TaskParameterValidator;
import com.myarch.dpbuddy.DPConst;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.audit.AuditEvent;
import com.myarch.dpbuddy.audit.AuditEventFile;
import com.myarch.dpbuddy.dpimport.ImportCommand;
import com.myarch.dpbuddy.dpimport.ImportResponse;
import com.myarch.dpbuddy.inpututil.Input;
import com.myarch.dpbuddy.inpututil.InputFileWrapper;
import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.dpbuddy.xmlutil.XMLParsingUtils;

/**
 * Restore multi-domain backup
 */
public class RestoreAntTask extends BaseDPBuddyTask {

    private File fileToRestore;

    private File deploymentPolicyFile;
    private String deploymentPolicyName;
    private String deploymentPolicyVariablesName;

    private Boolean overwriteFiles;
    private Boolean overwriteObjects;
    private Boolean rewriteLocalIp;
    private Boolean dryRun;
    
    private boolean reset=false;
    
    
    @Option(name="-reset", handler=ExplicitBooleanOptionHandler.class,
                    usage="Reset the domain's configuration before restore.")
    public void setReset( boolean reset ) {
        this.reset=reset;
    }    

    @Override
    protected boolean isDefaultDomainOverride() {
        return true;
    }

    @Override
    public void executeDPTask() throws BuildException {

        TaskParameterValidator validator=new TaskParameterValidator( this );
        validator.validateExistingFile( fileToRestore, "file",
                        "ZIP file containing the backup to restore" );
        validator.throwIfErrors();

        Input fileInput=new InputFileWrapper( fileToRestore );

        ImportCommand importCommand=new ImportCommand( true );
        importCommand.setFile( fileInput );

        setDeploymentPolicy( importCommand, deploymentPolicyFile, deploymentPolicyName,
                        deploymentPolicyVariablesName );

        if (overwriteFiles != null)
            importCommand.setOverwriteFiles( overwriteFiles );
        if (overwriteObjects != null)
            importCommand.setOverwriteObjects( overwriteObjects );
        if (rewriteLocalIp != null)
            importCommand.setRewriteLocalIp( rewriteLocalIp );
        if (dryRun != null) {
            if (dryRun)
                logger.info( "Executing 'import' in 'dry-run' mode, no changes will be made" );
            importCommand.setDryRun( dryRun );
        }

        Element doRestoreElt=importCommand.getRequestChild( ImportCommand.DO_RESTORE );
        addDomainsToRequestElt(doRestoreElt, fileToRestore);
        importCommand.setQuiet(true);
        performRestore( importCommand, getDevice().switchDefaultDomain() );

    }

    
    private void addDomainsToRequestElt(Element doRestoreElt, File zipFile){
        List<String> domainNames=extractDomains(zipFile);
        for(String domainName:domainNames) {
            addDomainToRequestElt( doRestoreElt, domainName );
        }
        logger.info("The following domains will be restored: "+domainNames);
    }
     
    
    private void addDomainToRequestElt(Element doRestoreElt, String domainName) {
        Element domainElt=new Element("domain", DPConst.DP_NS );
        domainElt.setAttribute( "name", domainName );
        if (reset) {
            domainElt.setAttribute( "reset-domain", "true" );
        }
        doRestoreElt.addContent( domainElt );
    }
    
    
    private List<String> extractDomains(File zipFile) {
        Document exportXml=XMLParsingUtils.parseFromZip( zipFile, "export.xml");
        Element domainsElt=JDomUtils.getRequiredChildElement( exportXml.getRootElement(),"domains" );
        List<String> domainNames=new ArrayList<>();
        for(Element domainElt:domainsElt.getChildren()) {
            domainNames.add( domainElt.getAttributeValue( "name" ) );
        }
        
        return domainNames;
    }
    
    private void performRestore( ImportCommand importCommand, Device device ) {

        // in case if we have audit, don't fail right away
        importCommand.setFailOnError( false );
        ImportResponse response=importCommand.execute( device );

        if (auditEventHandler != null) {
            AuditEvent auditEvent=createRestoreAuditEvent( response );
            auditEventHandler.handleAuditEvent( auditEvent, device );
        }
        response.throwIfFailure();

        if (!isDryRun())
            saveDomainConfigIfEnabled();

        // note - we don't want to automatically rollback in case of errors since there are could be multiple domains involved
    }

    
    private boolean isDryRun() {
        return dryRun != null && dryRun == true;
    }

    private AuditEvent createRestoreAuditEvent( ImportResponse importResponse ) {
        AuditEvent auditEvent=createAuditEvent( "restore" );
        auditEvent.addFile( new AuditEventFile( fileToRestore ) );

        return auditEvent;
    }

    @Option(name="-file", required=true, usage="ZIP file containing the backup to restore.")
    public void setFile( File file ) {
        this.fileToRestore=file;
    }

    public File getImportFile() {
        return fileToRestore;
    }

    @Option(name="-deploymentPolicyFile",
                    usage="Path to the deployment policy file on the local filesystem.")
    public void setDeploymentPolicyFile( File deploymentPolicyFile ) {
        this.deploymentPolicyFile=deploymentPolicyFile;
    }

    @Option(name="-deploymentPolicyName",
                    usage="Name of the existing deployment policy to use for the import operation.")
    public void setDeploymentPolicyName( String deploymentPolicy ) {
        this.deploymentPolicyName=deploymentPolicy;
    }

    @Option(name="-deploymentPolicyVars", aliases="-dpVars",
                    usage="Name of the deployment policy variables object to be applied to this deployment policy.")
    public void setDeploymentPolicyVars( String deploymentPolicyVariablesName ) {
        this.deploymentPolicyVariablesName=deploymentPolicyVariablesName;
    }

    @Option(name="-overwriteFiles", handler=ExplicitBooleanOptionHandler.class,
                    usage="Overwrite existing files with the files included with the import.")
    public void setOverwriteFiles( Boolean overwriteFiles ) {
        this.overwriteFiles=overwriteFiles;
    }

    @Option(name="-overwriteObjects", handler=ExplicitBooleanOptionHandler.class,
                    usage="Overwrite existing objects with the objects defined as part of this import.")
    public void setOverwriteObjects( Boolean overwriteObjects ) {
        this.overwriteObjects=overwriteObjects;
    }

    @Option(name="-rewriteLocalIp", handler=ExplicitBooleanOptionHandler.class,
                    usage="If set to 'true', the local address bindings of services contained in the import package will be rewritten to match the equivalent interfaces of the device.")
    public void setRewriteLocalIp( Boolean rewriteLocalIp ) {
        this.rewriteLocalIp=rewriteLocalIp;
    }

    @Option(name="-dryRun", handler=ExplicitBooleanOptionHandler.class,
                    usage="Run import in the 'dry run' mode.")
    public void setDryRun( Boolean dryRun ) {
        this.dryRun=dryRun;
    }
    
}
