package com.myarch.dpbuddy.backup;

import java.io.File;
import java.util.Collection;

import org.apache.commons.io.FileUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Element;

import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.action.ActionCommand;
import com.myarch.dpbuddy.filemanagement.CopyFilesCommand;
import com.myarch.dpbuddy.inpututil.InputFileWrapper;
import com.myarch.dpbuddy.restart.DPPoller;
import com.myarch.dpbuddy.restart.RestartDPService;
import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.util.ThreadUtils;

/**
 * Perform secure backup
 */
public class SecureRestoreCommand extends ActionCommand {

    // Location on the device we'll be restoring from
    public final static String DEFAULT_DP_DIR="local:/secure-restore";
    private final static String DP_DIR_ELT_NAME="source";

    private final Log logger=LogFactory.getLog( getClass() );

    private String dpDir=DEFAULT_DP_DIR;

    private File backupLocation;

    private int restartTimeout=RestartDPService.DEFAULT_TIMEOUT;
    private boolean waitForReboot=false;
    // just in case we'll wait a little longer for shutdown to start
    private int shutdownWaitTime=RestartDPService.DEFAULT_SHUTDOWN_WAIT_TIME + 5000;
    // TODO: machine type

    public SecureRestoreCommand(String cred, File backupLocation, boolean waitForReboot, int timeout) {
        super( "SecureRestore" );
        addActionChildElement( "cred", cred );
        // Validate is required even though it is optional in the schema
        addActionChildElement( DP_DIR_ELT_NAME, DEFAULT_DP_DIR );
        addActionChildElement( "validate", "off" );

        this.backupLocation=backupLocation;
        this.waitForReboot=waitForReboot;
        this.restartTimeout=timeout;
    }

    /**
     * The directory used to store the backup internally on the device
     * 
     * @param dir
     */
    public void setDPDir( String dpDir ) {
        this.dpDir=dpDir;
        Element actionElt=getActionElement();
        JDomUtils.addOrUpdateElement( actionElt, DP_DIR_ELT_NAME, dpDir );
    }

    private CopyFilesCommand createCopyCommand( File backupLocation, String targetDpDir ) {
        CopyFilesCommand copyCommand=new CopyFilesCommand();

        String[] extensions=new String[] { "xml", "tgz" };
        Collection<File> files=FileUtils.listFiles( backupLocation, extensions, false );
        for (File file : files) {
            copyCommand.addFile( new InputFileWrapper( file ), targetDpDir );
        }

        return copyCommand;
    }

    @Override
    public ResultResponse execute( Device device ) {

        Device defaultDomainDevice=device.switchDefaultDomain();

        logger.info( String.format( "Copying backup files to the device  from '%s' ...",
                        backupLocation.getAbsolutePath() ) );

        CopyFilesCommand copyCommand=createCopyCommand( backupLocation, dpDir );
        copyCommand.execute( defaultDomainDevice );

        logger.info( "Performing secure restore of the device. This will take a few minutes ..." );

        ResultResponse sbackupResponse=super.execute( defaultDomainDevice );
        logger.info( "Secure restore has been initiated." );
        if (waitForReboot) {
            long startTime=System.nanoTime();

            logger.info( String.format( "Waiting for %d seconds for the reboot process to start ...",
                            shutdownWaitTime / 1000 ) );
            ThreadUtils.delay( shutdownWaitTime );

            DPPoller poller=new DPPoller();
            logger.info( String.format(
                            "Waiting for the device to complete reboot after restore. "
                                + "This will take some time (up to %d seconds timeout) ...",
                            restartTimeout ) );

            poller.pollUntilSuccess( defaultDomainDevice, restartTimeout );
            long elapsedTime=System.nanoTime() - startTime;
            logger.info( String.format(
                            "Reboot after secure restore has been completed in %d sconds, DataPower is up and running.",
                            elapsedTime / 1000 / 1000 / 1000 ) );
        }

        return sbackupResponse;

    }

}
