package com.myarch.dpbuddy.checkpoint;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.action.ActionCommand;
import com.myarch.dpbuddy.status.StatusCommand;
import com.myarch.dpbuddy.status.StatusResponse;

public class RemoveCheckpointCommand extends ActionCommand {
    private final Log logger=LogFactory.getLog( getClass() );

    private String checkpointName;
    private Pattern checkpointPattern;
    private boolean isMatchRequired;

    public final static String ACTION_NAME="RemoveCheckpoint";
    public final static String ACTION_ELEMENT_NAME="ChkName";

    

    
    public RemoveCheckpointCommand(String checkpointName, String checkpointNamePattern,
                    boolean isMatchRequired) {
        super();
        this.checkpointName=checkpointName;
        if (checkpointNamePattern != null)
            this.checkpointPattern=Pattern.compile( checkpointNamePattern, Pattern.CASE_INSENSITIVE );
        this.isMatchRequired=isMatchRequired;
    }

    @Override
    public ResultResponse execute( Device device ) {
        if (StringUtils.isNotBlank( checkpointName ))
            addRemoveCheckpointAction( checkpointName );

        if (checkpointPattern != null) {
            List<String> matchingCheckpoints=findMatchingCheckpoints( device, checkpointPattern );
            if (isMatchRequired && matchingCheckpoints.size() == 0)
                throw new DPBuddyException(
                                "Domain '%s' does not contain any checkpoints matching '%s'",
                                device.getCurrentDomain(), checkpointPattern.toString() );
            logger.info( String.format(
                            "Following checkpoints matched the pattern '%s' and will be deleted: %s",
                            checkpointPattern, StringUtils.join( matchingCheckpoints, ", " ) ) );
            for (String matchingCheckpointName : matchingCheckpoints)
                addRemoveCheckpointAction( matchingCheckpointName );
            if (matchingCheckpoints.size() == 0)
                logger.info( "No checkpoints were found for the name pattern " + checkpointPattern );
        }

        ResultResponse response=null;
        if (getActionElements().size() > 0) {
            response=super.execute( device );
        }
        else {
            logger.debug( "No checkpoints to delete" );
        }
        return response;
    }

    private void addRemoveCheckpointAction( String checkpointName ) {
        logger.debug( String.format( "Checkpoint '%s' will be removed if exists", checkpointName ) );
        addAction( ACTION_NAME, ACTION_ELEMENT_NAME, checkpointName );
    }

    private List<String> findMatchingCheckpoints( Device device, Pattern namePattern ) {
        StatusCommand statusCmd=new StatusCommand( "DomainCheckpointStatus" );
        StatusResponse statusResponse=statusCmd.execute( device );
        List<String> existingCheckpoints=statusResponse.getStatusValues( "ChkName" );
        List<String> matchingCheckpoints=new ArrayList<String>();
        for (String checkpointName : existingCheckpoints) {
            Matcher matcher=namePattern.matcher( checkpointName );
            if (matcher.matches())
                matchingCheckpoints.add( checkpointName );
        }

        return matchingCheckpoints;
    }
}
