package com.myarch.dpbuddy.cli;

import java.io.File;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.tools.ant.DefaultLogger;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.ProjectHelper;
import org.kohsuke.args4j.CmdLineException;

import com.myarch.antutil.AntCommonsLogger;
import com.myarch.antutil.LoggerConfiguration;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.hocon.HoconAntUtils;
import com.myarch.util.FilePathUtils;

public class DPBuddyCliExecutor {

    private final Log logger=LogFactory.getLog( getClass() );

    // private Map<String, Task> commandTaskMap=new HashMap<String, Task>();
    private DPBuddyCliParser parser;
    
    public DPBuddyCliExecutor(CliCommandDefContainer cmdDefs) {
        parser=new DPBuddyCliParser( cmdDefs );
    }

    public final int FAILURE_EXIT_CODE=12;
    public final int SUCCESS_EXIT_CODE=0;

    public int execTask( String[] args ) {
        try {
            Project project=createProject( args );
            args=resolveArgVars( project, args );

            BaseDPBuddyTask task=parser.parseCommandLine( args );
            // if we just printed help, the task will be null
            if (task != null) {
                task.setProject( project );
                project.registerThreadTask( Thread.currentThread(), task );

                AntCommonsLogger.setLoggerConfiguration( LoggerConfiguration.CLI_CONFIG );
                BaseDPBuddyTask.setUpLogging( project );

                // load properties
                setProjectProperties( task );
                // task.reconfigure();

                task.execute();
            }
            return SUCCESS_EXIT_CODE;
        }
        catch (Exception e) {
            processException( args, e );
            return FAILURE_EXIT_CODE;
        }
    }

    public BaseDPBuddyTask instantiateTask( String commandName )  {
        // TODO: handle verbose
        String[] projectArgs=new String[0];
        
        Project project=createProject( projectArgs );
        BaseDPBuddyTask task=parser.getTaskInstance( commandName);
        if (task == null) {
            throw new IllegalArgumentException( "Invalid DPBuddy command name " + commandName );
        }
        task.setProject( project );
        project.registerThreadTask( Thread.currentThread(), task );

        AntCommonsLogger.setLoggerConfiguration( LoggerConfiguration.CLI_CONFIG );
        BaseDPBuddyTask.setUpLogging( project );

        return task;

    }

    private String[] resolveArgVars( Project project, String[] args ) {
        String[] resolvedArgs=new String[args.length];
        for (int i=0; i < args.length; ++i) {
            resolvedArgs[i]=project.replaceProperties( args[i] );
        }
        return resolvedArgs;
    }

    private void processException( String[] args, Exception e ) {
        String commandName=null;
        if (args.length >= 1 && !args[0].startsWith( "-" ))
            commandName=args[0];

        if (e instanceof CmdLineException) {
            System.err.println( e.getMessage() );
            if (commandName != null) {
                System.err.println( "Try 'dpbuddy " + commandName + " -h' for more information" );
            }
            else {
                System.err.println( "Try 'dpbuddy -h' or 'dpbuddy <command> -h' for more information" );
            }
        }
        else {
            String message=e.getMessage();
            if (commandName != null)
                message=String.format( "Command '%s' failed:\n%s", commandName, message );
            System.err.println( message );
        }
        if (DPBuddyCliParser.isVerbose( args )) {
            e.printStackTrace();
        }
    }

    public void setProjectProperties( BaseDPBuddyTask task ) {
        Project project=task.getProject();

        File configFile=locateConfigFile( task );
        if (configFile != null) {
            logger.debug( "Loading configuration file from " + configFile.getAbsolutePath() );
            // load properties file using Apache commons config for backward
            // compatibility
            if (FilenameUtils.isExtension( configFile.getName(), "properties" )) {
                Config config=new Config( configFile );
                config.addConfigPropsToAntProject( project );
            }
            // Hocon should handle most of the extensions
            else {
                HoconAntUtils.addConfig( project, configFile );
            }
        }

    }

    private Project createProject( String[] args ) {
        Project project=new Project();
        project.init();
        DefaultLogger antLogger=new DefaultLogger();
        antLogger.setOutputPrintStream( System.out );
        antLogger.setErrorPrintStream( System.err );
        if (DPBuddyCliParser.isVerbose( args ))
            antLogger.setMessageOutputLevel( Project.MSG_VERBOSE );
        else
            antLogger.setMessageOutputLevel( Project.MSG_INFO );
        project.addBuildListener( antLogger );
        setAntProps( project );
        ProjectHelper projectHelper=new ProjectHelper();
        projectHelper.getImportStack().add( new File("dpbuddy-cli-ant-bridge.xml") );
        project.addReference( ProjectHelper.PROJECTHELPER_REFERENCE, projectHelper );
//        ProjectHelper helper =
//                        (ProjectHelper) getProject().
//                            getReference(ProjectHelper.PROJECTHELPER_REFERENCE);

        return project;

    }

    private void setAntProps( Project project ) {
        File basedir=new File( "./" );
        project.setBaseDir( basedir );
    }

    public final static String DEFAULT_PROPERTIES_FILE_NAME="dpbuddy.properties";

    public final static String DEFAULT_CONF_FILE_NAME="dpbuddy.conf";

    public final static String CONFIG_ENV_VAR_NAME="DPBUDDY_CONFIG";

    private File locateConfigFile( BaseDPBuddyTask task ) {
        File configFile=null;
        // get it from parameter
        String configFilePath=task.getConfigFile();
        String configEnvVar=System.getenv( CONFIG_ENV_VAR_NAME );

        if (configFilePath != null) {
            // the full path file was provided on the command line
            configFile=new File( configFilePath );
            if (!configFile.exists()) {
                throw new CliException(
                                "The DPBuddy config file path '%s' that you provided does not exist",
                                configFile.getAbsolutePath() );
            }
        }
        else if (StringUtils.isNotBlank( configEnvVar )) {
            // the full path to file was provided by the env var
            configFile=new File( configEnvVar.trim() );

            if (!configFile.exists()) {
                throw new CliException( "The DPBuddy config file '%s' set by '%s' does not exist",
                                configFile.getAbsolutePath(), CONFIG_ENV_VAR_NAME );
            }
        }
        else {
            // try current dir or home or etc
            configFile=loadConfigFileFromDefault( DEFAULT_CONF_FILE_NAME );
            if (configFile == null) {
                configFile=loadConfigFileFromDefault( DEFAULT_PROPERTIES_FILE_NAME );
                if (configFile == null) {
                    logger.warn( "Did not find DPBuddy config file. The config file could reside in the current directory or in home/$username/.dpbuddy or in /etc or in $DPBUDDY_HOME/conf" );
                }
            }
        }

        if (configFile != null) {
            logger.debug( "Found DPBuddy config file " + configFile.getAbsolutePath() );
        }

        return configFile;
    }

    private File loadConfigFileFromDefault( String defaultConfigFileName ) {
        // try current dir
        File configFile=new File( "./" + defaultConfigFileName );
        if (!configFile.exists()) {
            // try classpath
            configFile=FilePathUtils.getOptionalFileFromResource( defaultConfigFileName );
            if (configFile==null || ! configFile.exists()) {
                // try home
                configFile=new File( getUserHomeLocation(), defaultConfigFileName );
                if (!configFile.exists()) {
                    // try etc
                    File etc=getEtcLocation();
                    if (etc.canRead()) {
                        configFile=new File( etc, defaultConfigFileName );
                    }
                }
            }
        }

        if (configFile!=null && !configFile.exists()) {
            configFile=null;
        }

        return configFile;
    }

    private File getUserHomeLocation() {
        String userHome=System.getProperty( "user.home" );
        File location=new File( userHome + "/.dpbuddy" );
        return location;
    }

    private File getEtcLocation() {
        File location=new File( "/etc" );
        return location;
    }

}
