package com.myarch.dpbuddy.cli;

import java.io.PrintWriter;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.apache.tools.ant.RuntimeConfigurable;
import org.apache.tools.ant.Task;
import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.OptionHandlerFilter;
import org.kohsuke.args4j.spi.OptionHandler;

import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.ant.NoOpTask;

public class DPBuddyCliParser {

    private CliCommandDefContainer cmdDefs;

    public DPBuddyCliParser(CliCommandDefContainer cmdDefs) {
        this.cmdDefs=cmdDefs;
    }

    @SuppressWarnings("deprecation")
    public BaseDPBuddyTask parseCommandLine( String[] args ) throws CmdLineException {
        BaseDPBuddyTask task=null;
        if (isGeneralHelp( args )) {
            printGeneralHelp();
        }
        else if (isVersionArg( args[0] )) {
            System.out.println( "DPBuddy version " + BaseDPBuddyTask.getDPBuddyVersion() );
        }
        else {
            task=determineTask( args );
            CmdLineParser parser=new CmdLineParser( task );
            if (isHelp( args )) {
                printCommandHelp( parser, task );
                task=null;
            }
            else {
                parser.parseArgument( args );
                List<String> cliArguments=task.getCliArguments();
                if (cliArguments.size() > 1) {
                    String unrecognizedArgs=StringUtils
                                    .join( cliArguments.subList( 1, cliArguments.size() ), "," );
                    String errorMessage="You provided too many arguments. The following arguments were not recognized: "
                        + unrecognizedArgs;
                    throw new CmdLineException( parser, errorMessage );
                }

                // create runtime configurable with the variables that still
                // need to be resolved - they will be resolved for each env
                // FIXME: this doesn't take into account -name=opt syntax
                RuntimeConfigurable runtimeconfigurable=task.getRuntimeConfigurableWrapper();
                int i=0;
                for (String arg : args) {
                    if (arg.contains( "${" ) && i > 0 && args[i - 1].startsWith( "-" )
                        && args[i - 1].length() > 1) {
                        String attrName=args[i - 1].substring( 1 );
                        runtimeconfigurable.setAttribute( attrName, arg );
                    }
                    ++i;
                }
            }
        }
        return task;
    }

    static boolean isHelp( String[] args ) {
        for (String arg : args) {
            if (isHelpArg( arg ))
                return true;
        }
        return false;
    }

    static boolean isAllOptions( String[] args ) {
        for (String arg : args) {
            if (!arg.startsWith( "-" )) {
                return false;
            }
        }
        return true;
    }

    static boolean isGeneralHelp( String[] args ) {
        return args.length < 1 || isHelpArg( args[0] );
    }

    private BaseDPBuddyTask determineTask( String[] args ) {

        String taskName=getTaskNameFromArgs( args );

        BaseDPBuddyTask task=getTaskInstance(taskName);

        return task;
    }
    
    public BaseDPBuddyTask getTaskInstance(String taskName) {
        CliCommandDef cmdDef=cmdDefs.get( taskName );

        if (cmdDef == null) {
            throw new CliException(
                            "Command '%s' is not a valid DPBuddy command. Following is the list of supported commands:\n%s",
                            taskName, getSupportedCommandStrList() );
        }

        BaseDPBuddyTask task=cmdDef.getAntTask();
        
        return task;
    }
    

    private static final CmdLineParser COMMON_OPTIONS_PARSER=new CmdLineParser( new NoOpTask() );

    private void printGeneralHelp() {
        System.out.println( "Usage: dpbuddy <command> [options]" );
        System.out.println(
                        "\nDPBuddy supports the following commands:\n" + getSupportedCommandStrList() );
        System.out.println( "\nThe following options are common for all commands:\n" );
        printCommonOptions();
    }

    private void printCommandHelp( CmdLineParser parser, Task task ) {
        System.out.println( String.format( "Usage: dpbuddy %s [options]\n", task.getTaskName() ) );
        // TODO: description and example
        parser.printUsage( new PrintWriter( System.out, true ), null, COMMAND_OPTIONS_FILTER );
    }

    private void printCommonOptions() {
        COMMON_OPTIONS_PARSER.printUsage( System.out );

    }

    private String getTaskNameFromArgs( String[] args ) {
        String taskNameCandidate=args[0];
        if (args.length > 1 && isVerboseArg( args[0] ))
            taskNameCandidate=args[1];

        if (taskNameCandidate.startsWith( "-" ))
            throw new CliException(
                            "DPBuddy command cannot start with '-'. All options must follow the command name. You provided: '%s'",
                            taskNameCandidate );
        return taskNameCandidate;

    }

    private String getSupportedCommandStrList() {
        return cmdDefs.getSupportedCommandStrList();
    }

    static boolean isHelpArg( String arg ) {
        return "-h".equalsIgnoreCase( arg ) || "--help".equalsIgnoreCase( arg )
            || "-help".equalsIgnoreCase( arg );
    }

    static boolean isVersionArg( String arg ) {
        return "-version".equalsIgnoreCase( arg ) || "--version".equalsIgnoreCase( arg );
    }

    static boolean isVerbose( String[] args ) {
        for (String arg : args) {
            if (isVerboseArg( arg ))
                return true;
        }
        return false;
    }

    static boolean isVerboseArg( String arg ) {
        return "-v".equalsIgnoreCase( arg ) || "-verbose".equalsIgnoreCase( arg );
    }

    static final CommandOptionsFilter COMMAND_OPTIONS_FILTER=new CommandOptionsFilter();

    public final static String[] COMMON_OPTIONS=new String[] { "-h", "-v", "-configFile" };

    static class CommandOptionsFilter implements OptionHandlerFilter {

        @Override
        public boolean select( @SuppressWarnings("rawtypes" ) OptionHandler optionHandler) {
            String optionName=optionHandler.getNameAndMeta( null );
            return !isCommon( optionName );
        }

        @SuppressWarnings("rawtypes")
        private boolean isCommon( String optionName ) {
            List<OptionHandler> commonOptions=COMMON_OPTIONS_PARSER.getOptions();

            for (OptionHandler option : commonOptions) {
                String commonOptionName=option.getNameAndMeta( null );
                // System.out.println( "Comparing "+optionName+" with
                // "+commonOptionName);
                if (optionName.equalsIgnoreCase( commonOptionName )) {
                    return true;
                }

            }

            return false;
        }

    }
}
