package com.myarch.dpbuddy.client;

import java.io.IOException;
import java.io.InputStream;
import java.net.ConnectException;

import javax.net.ssl.SSLHandshakeException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.util.StreamReaderDelegate;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.HttpHost;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPConnection;
import com.myarch.dpbuddy.DPConnectionException;
import com.myarch.dpbuddy.DPRequest;
import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.dpbuddy.xmlutil.SoapUtils;
import com.myarch.dpbuddy.xmlutil.XMLParsingUtils;

/**
 * Communicate with DP device using XML management interface. Takes care of
 * putting the request into SOAP envelope. Checks the response for SOAP faults
 */
public class DPClient {

    private static Log logger=LogFactory.getLog( DPClient.class );

    public static Document executeRequest( DPConnection connection, DPRequest request ) {
        return executeRequest( null, null, connection, connection.getDomain(), request, false, false );
    }

    public static Document executeRequest( StreamReaderDelegate responseParser, DPConnection connection,
                    DPRequest request ) {
        return executeRequest( null, responseParser, connection, connection.getDomain(), request, false, false );
    }

    public static Document executeRequest( DPConnection connection, DPRequest request, boolean use2004URL ) {
        return executeRequest( null, null, connection, connection.getDomain(), request, false, use2004URL );
    }

    // note that quiet is no longer in use
    // httpClient is not currently in use. It's just a way to provide client
    // with special settings
    public static Document executeRequest( SecureHTTPClient httpClient, StreamReaderDelegate responseParser,
                    DPConnection connection, String domain, DPRequest request, boolean quiet, boolean use2004URL ) {

        // reinitialize logger in case if we changed the level from unit test
        logger=LogFactory.getLog( DPClient.class );

        String msg=String.format( "Executing XML management request '%s' against '%s' domain", request.getDisplayName(),
                        domain );

        logger.debug( msg );

        Element requestElement=request.getRequestElement();
        // FIXME: non-thread-safe -- implement as part of serialization instead
        String savedDomain=requestElement.getAttributeValue( "domain" );
        if (domain != null) {
            requestElement.setAttribute( "domain", domain );
        }

        // byte[] serializedRequest=serializeToBytes( request.getDocument() );
        // logger.debug( "Serialized request's size is " +
        // serializedRequest.length + " bytes" );

        if (logger.isDebugEnabled()) {
            String requestString=JDomUtils.docToCutBase64String( request.getDocument() );
            logger.debug( "Request:\n" + SoapUtils.wrapInSoap( requestString ) );
        }

        // note that file upload requests are not validated because of the 2004
        // schema
        if (connection.isSchemaValidation() && !use2004URL) {
            new DPSchemaValidator().validate( request.getRequestElement() );
            logger.debug( "Successfully validated " + request.getDisplayName() + " against DataPower schema" );
        }

        if (httpClient == null) {
            httpClient=connection.getHttpClient( use2004URL );
        }

        String url=httpClient.getUrlString();

        httpClient.setCredentials( connection.getUsername(), connection.getPassword() );

        if (logger.isDebugEnabled()) {
            String connectionMessage="Connecting to ";
            if (connection.getDeviceName() != null)
                connectionMessage+="the device (environment) '" + connection.getDeviceName() + "' at ";
            connectionMessage+=url;

            logger.debug( connectionMessage );
        }

        InputStream responseInputStream=null;
        try (InputStream requestInputStream=request.getInputStream()) {

            responseInputStream=httpClient.postInputStream( requestInputStream );

            if (savedDomain != null) {
                requestElement.setAttribute( "domain", savedDomain );
            }
        }
        catch (IOException e) {
            String urlMessage=url;
            HttpHost proxy=httpClient.getProxy();
            if (proxy != null)
                urlMessage=urlMessage + " (Proxy: " + proxy + ")";
            if (e instanceof ConnectException || e instanceof java.net.UnknownHostException)
                throw new DPConnectionException(
                                "Unable to connect to DataPower at %s, please check that host and port are correct. You need to specify the port of XML management interface, 5550 is the default",
                                e, urlMessage );
            else if (e instanceof SSLHandshakeException)
                throw new DPConnectionException( "Unable to establish SSL connection with DataPower at %s, "
                    + "please make sure that you imported DataPower certificate into your JDK's keystore "
                    + "or use dpTrustAllCerts property/attribute. This is caused by this error:\n%s", e, urlMessage,
                                e.getMessage() );
            else
                throw new DPConnectionException(
                                "Unable to connect to DataPower at %s, please check your connection parameters", e,
                                urlMessage );
        }

        logger.debug( "Parsing the response ..." );
        XMLStreamReader responseStreamReader=XMLParsingUtils.getDefaulStreamReader( responseInputStream );
        if (responseParser != null) {
            responseParser.setParent( responseStreamReader );
            responseStreamReader=responseParser;
        }

        Document doc=XMLParsingUtils.parseUsingStax( responseStreamReader );

        try {
            responseInputStream.close();
            httpClient.closeResponse();
        }
        catch (IOException e) {
            logger.error( "Error trying to close input stream", e );
        }

        if (logger.isDebugEnabled()) {
            logger.debug( "Response:\n" + JDomUtils.docToCutBase64String( doc ) );
        }

        String faultReason=SOAPUtil.extractSOAPFaultReason( doc );
        if (faultReason != null) {

            String errorMessage="Request to DataPower at %s, domain '%s' failed with the error '%s'.";
            String faultExplanation=produceExplanationForFault( faultReason, connection, domain );
            if (faultExplanation != null)
                errorMessage+=" " + faultExplanation;
            throw new DPBuddyException( errorMessage, url, domain, faultReason );
        }

        return doc;
    }

    private static String produceExplanationForFault( String faultReason, DPConnection connection, String domain ) {
        String explanation=null;
        if (StringUtils.containsIgnoreCase( faultReason, "authentication" )) {
            explanation=String.format(
                            "Please make sure that the domain '%s' exists and that the user '%s' has access to this domain.",
                            domain, connection.getUsername() );
        }
        return explanation;

    }

}
