package com.myarch.dpbuddy.client;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocket;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpHost;
import org.apache.http.HttpVersion;
import org.apache.http.auth.AuthenticationException;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.ssl.SSLContextBuilder;
import org.apache.http.conn.ssl.TrustSelfSignedStrategy;
import org.apache.http.conn.ssl.X509HostnameVerifier;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.impl.NoConnectionReuseStrategy;
import org.apache.http.impl.auth.BasicScheme;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.client.HttpClients;

import com.myarch.dpbuddy.DPBuddyException;

@SuppressWarnings("deprecation")
public class SecureHTTPClient {

    private final Log logger=LogFactory.getLog( getClass() );

    private CloseableHttpClient httpClient;
    private HttpPost httpPost;
    private CloseableHttpResponse response;

    public SecureHTTPClient(String url, boolean isTrustAllCerts) {
        long startTime=System.nanoTime();
        httpClient=createTrustingClient( isTrustAllCerts );

        httpPost=new HttpPost( url );
        httpPost.setProtocolVersion( HttpVersion.HTTP_1_1 );

        long elapsedTime=System.nanoTime() - startTime;

        logger.debug( "Established connection in " + elapsedTime / 1000 / 1000 + " ms" );
    }

    public void setTimeouts( int socketTimeout, int connectionTimeout ) {

        RequestConfig requestConfig=RequestConfig.copy( RequestConfig.DEFAULT )
                        .setSocketTimeout( socketTimeout ).setConnectTimeout( connectionTimeout )
                        .build();

        setRequestConfig( requestConfig );
    }

    public void setRequestConfig( RequestConfig requestConfig ) {

        httpPost.setConfig( requestConfig );
    }

    public String getUrlString() {
        return httpPost.getURI().toString();
    }

    private CloseableHttpClient createTrustingClient( boolean isTrustAllCerts ) {

        HttpClientBuilder hcb=HttpClients.custom();
        // Attempt to deal with the bug in httpClient throwing NohttpResponse
        // connection becomes stale otherwise
        hcb.setConnectionReuseStrategy( NoConnectionReuseStrategy.INSTANCE );

        /* 
        hcb.setRetryHandler( 
                        new HttpRequestRetryHandler() {
                            @Override
                            public boolean retryRequest( IOException exception, int executionCount,
                                            HttpContext context ) {
                                logger.info("retrying "+executionCount);
                                if (executionCount > 3) {
                                    logger.warn( "Maximum tries reached for client http pool " );
                                    return false;
                                }
                                if (exception instanceof org.apache.http.NoHttpResponseException) {
                                    logger.warn( "No response from server on " + executionCount
                                        + " call" );
                                    return true;
                                }
                                return false;
                            }
                        } );
        */
        if (isTrustAllCerts) {
            SSLContextBuilder sslctxb=new SSLContextBuilder();
            SSLContext sslctx=null;
            try {
                sslctxb.loadTrustMaterial( KeyStore.getInstance( KeyStore.getDefaultType() ),
                                new TrustSelfSignedStrategy() {
                                    @Override
                                    public boolean isTrusted( X509Certificate[] chain, String authType )
                                                    throws CertificateException {
                                        return true;
                                    }
                                } );

                sslctx=sslctxb.build();
            }
            catch (NoSuchAlgorithmException e) {
                handleTrustingClientException( e );
            }
            catch (KeyStoreException e) {
                handleTrustingClientException( e );
            }
            catch (KeyManagementException e) {
                handleTrustingClientException( e );
            }

            hcb.setSslcontext( sslctx );
        }
        // note: we always trust all hosts so we can accept self-signed certs
        hcb.setHostnameVerifier( new X509HostnameVerifier() {
            @Override
            public void verify( String host, SSLSocket ssl ) throws IOException {
            }

            @Override
            public void verify( String host, X509Certificate cert ) throws SSLException {
            }

            @Override
            public void verify( String host, String[] cns, String[] subjectAlts ) throws SSLException {
            }

            @Override
            public boolean verify( String hostname, SSLSession session ) {
                return true;
            }
        } );

        HttpHost proxy=getProxy();
        if (proxy != null)
            hcb.setProxy( proxy );

        CloseableHttpClient c=hcb.build();

        return c;
    }

    public HttpHost getProxy() {
        HttpHost proxy=null;
        String proxyHost=System.getProperty( "https.proxyHost" );
        // System.out.println("Proxy host: "+proxyHost);
        if (proxyHost != null) {
            String proxyPortStr=System.getProperty( "https.proxyPort" );
            if (proxyPortStr == null)
                proxyPortStr="443";
            if (!StringUtils.isNumeric( proxyPortStr ))
                throw new DPBuddyException(
                                "You supplied https.proxyPort property, however, the port is not numeric: %s",
                                proxyPortStr );

            int proxyPort=Integer.parseInt( proxyPortStr );
            proxy=new HttpHost( proxyHost, proxyPort, "http" );
            logger.debug( String.format( "Using the following https proxy settings: %s:%d", proxyHost,
                            proxyPort ) );
        }

        return proxy;
    }

    private void handleTrustingClientException( Exception e ) {
        throw new DPBuddyException(
                        "We were unable to set up our SSL connection to accept self-signed certificates. Please try running DPBuddy with dpTrustAllCerts='false'",
                        e );

    }

    public void closeResponse() throws IOException {
        if (response != null)
            response.close();
    }

    public void setCredentials( String username, String password ) {

        configureAuth( username, password );
    }

    protected void configureAuth( String username, String password ) {

        UsernamePasswordCredentials creds=new UsernamePasswordCredentials( username, password );
        BasicScheme basicScheme=new BasicScheme();
        try {
            httpPost.setHeader( basicScheme.authenticate( creds, httpPost, null ) );
        }
        catch (AuthenticationException e) {
            throw new DPBuddyException( "Unable to set the basic authentication header", e );
        }

    }

    public InputStream postBytes( byte[] content ) throws ClientProtocolException, IOException {
        InputStreamEntity entity=new InputStreamEntity( new ByteArrayInputStream( content ),
                        content.length, ContentType.TEXT_XML );
        return post( entity );
    }

    public InputStream postInputStream( InputStream content )
                    throws ClientProtocolException, IOException {
        InputStreamEntity entity=new InputStreamEntity( content, ContentType.TEXT_XML );
        return post( entity );
    }

    private InputStream post( HttpEntity requestEntity ) throws ClientProtocolException, IOException {

        httpPost.setEntity( requestEntity );

        // logger.info("Headers: "+Arrays.asList( httpPost.getAllHeaders() ));

        logger.debug( "Executing HTTP POST request ..." );
        long startTime=System.nanoTime();
        CloseableHttpResponse response;
        try {
            response=httpClient.execute( httpPost );
        }
        catch (IOException ioe) {
            long elapsedTime=System.nanoTime() - startTime;
            logger.debug( "Received IOException from the device in " + elapsedTime / 1000 / 1000
                + " ms" );
            throw ioe;
        }

        int statusCode=response.getStatusLine().getStatusCode();

        long elapsedTime=System.nanoTime() - startTime;
        logger.debug( "Received response from the device in " + elapsedTime / 1000 / 1000
            + " ms. Status code: " + statusCode );

        if (statusCode < 200 || statusCode >= 300) {
            String msg="HTTP Request failed. Code: " + statusCode;
            logger.error( msg );
            throw new DPBuddyException( "Received invalid status code '%d' from the device at '%s'",
                            statusCode, getUrlString() );
        }

        HttpEntity entity=response.getEntity();

        Header responseContentTypeHeader=entity.getContentType();
        if (responseContentTypeHeader != null) {
            String responseContentType=responseContentTypeHeader.getValue();
            if (!StringUtils.containsIgnoreCase( responseContentType, "xml" )) {
                throw new DPBuddyException(
                                "Received invalid content type '%s' from the device at '%s', was expecting 'text/xml'",
                                responseContentType, getUrlString() );
            }

        }

        // TODO: handle entity being null
        InputStream responseInputStream=entity.getContent();

        return responseInputStream;
    }

}
