package com.myarch.dpbuddy.config;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.kohsuke.args4j.Option;
import org.kohsuke.args4j.spi.ExplicitBooleanOptionHandler;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.ant.DataPowerConfigurationType;
import com.myarch.dpbuddy.config.ConfigCommand.ConfigOperation;
import com.myarch.dpbuddy.filemanagement.ant.TransformableFileSet;
import com.myarch.dpbuddy.xmltransform.TransformActionContainer;
import com.myarch.dpbuddy.xmlutil.XMLInput;

public class ConfigAntTask extends BaseDPBuddyTask {

    private List<TransformableFileSet> fileSets=new ArrayList<TransformableFileSet>();
    private File configFile;
    private List<DataPowerConfigurationType> dpConfigs=new ArrayList<DataPowerConfigurationType>();

    private boolean resolveVars=true;

    @Option(name="-file", usage="Path to the configuration file.")
    public void setFile( File file ) {
        this.configFile=file;
    }

    @Option(name="-resolveVars", handler=ExplicitBooleanOptionHandler.class,
                    usage="Resolve variables (${}) inside the config file.")
    public void setResolveVars( boolean resolveVars ) {
        this.resolveVars=resolveVars;
    }

    public void addFileset( TransformableFileSet fileSet ) {
        fileSets.add( fileSet );
    }

    public void addConfiguration( DataPowerConfigurationType config ) {
        dpConfigs.add( config );
    }

    private TransformActionContainer transformActionContainer;

    // nested transform element, using Ant conventions
    public TransformActionContainer createTransform() {
        transformActionContainer=TransformActionContainer
                        .createTransformOnlyIfNull( transformActionContainer );
        return transformActionContainer;
    }

    @Override
    protected Boolean isValidateIfNotSet() {
        return true;
    }


    @Override
    public void executeDPTask() throws BuildException {

        ConfigOperation op=ConfigOperation.SET;
        if (getTaskName().toLowerCase().contains( "modify" ))
            op=ConfigOperation.MODIFY;

        ConfigCommand config=new ConfigCommand( op );

        if (transformActionContainer != null && configFile == null && dpConfigs.size() == 0)
            throw new DPBuddyException(
                            "You specified 'transform' element without defining nested 'configuration' or specifying 'file' attribute" );

        if (configFile != null) {
            addConfigFileToConfigCommandWithTransform( config, new XMLInput( configFile ) );
        }

        for (DataPowerConfigurationType dpConfig : dpConfigs) {
            addConfigFileToConfigCommandWithTransform( config, dpConfig.getConfigurationXml() );
        }

        processFileSets( config, fileSets );

        if (config.isEmpty())
            throw new BuildException(
                            "You must specify 'file' attribute and/or nested filesets elements" );

        executeRequest( config );
        saveDomainConfigIfEnabled();
    }
    

    private void processFileSets( ConfigCommand config, List<TransformableFileSet> fileSets ) {
        for (TransformableFileSet fileset : fileSets) {
            DirectoryScanner ds=fileset.getDirectoryScanner();
            String[] files=ds.getIncludedFiles();
            for (String file : files) {
                String filePath=ds.getBasedir() + System.getProperty( "file.separator" ) + file;
                XMLInput transformedFile=fileTransformer.transformXMLFileIfTransformProvided( new File(
                                filePath ), fileset.getTransform(), getDevice().getDPConnection() );
                addConfigFileToConfigCommand( config, transformedFile );
            }
            if (files.length == 0)
                throw new DPBuddyException(
                                "Could not find any files matching the fileset with the base dir '%s'",
                                fileset.getDirectoryScanner().getBasedir() );
        }
    }

    private void addConfigFileToConfigCommandWithTransform( ConfigCommand configCommand, XMLInput input ) {
        if (transformActionContainer != null) {
            input.transform( transformActionContainer );
        }

        if (resolveVars) {
            input.resolveVars( propExpressionEvaluator );
        }
        addConfigFileToConfigCommand( configCommand, input );
    }

    private void addConfigFileToConfigCommand( ConfigCommand configCommand, XMLInput input ) {
        logger.debug( "Adding this file to the config request: " + input.getName() );
        configCommand.addConfiguration( input );
    }

}
