package com.myarch.dpbuddy.config;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.BaseRequest;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPConst;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.dpbuddy.xmlutil.XMLInput;



/**
 * Performs set-config or modify-config operation.
 * Loads configuration file in datapower XML format, forms set-config/modify-config request
 * and executes it.  
 */
// TODO: de-couple request and the command
public class ConfigCommand extends BaseRequest implements DPCommand<ResultResponse>{
    
    @SuppressWarnings("unused")
    private Log logger = LogFactory.getLog(getClass());
    
    // set-config or modify-config
    private Element configOpElement;
   
    private ConfigOperation operation; 
    
    public ConfigCommand( ConfigOperation operation ) {
        this.operation = operation;
        createConfigOpElement();
    }
    
    public ConfigCommand( ) {
        this( ConfigOperation.SET );
    }
    
    public boolean isEmpty() {
        return configOpElement.getChildren().size()==0;
    }
    
    public void addConfiguration(XMLInput xmlSource){
        addConfigurationDoc( xmlSource.getDocument() );
    }

    public void addConfiguration(DPConfiguration dpConfiguration){
        addConfigurationDoc( dpConfiguration.getDocument() );
    }

    private void addConfigurationDoc( Document inputConfigDoc) {
        // we need to get to "configuration" element and ignore everything else
        Element rootConfigElement;
        rootConfigElement = inputConfigDoc.getRootElement().getChild("configuration" );
        if (rootConfigElement==null)
            throw new DPBuddyException("You must provide 'configuration' element that contains the objects that are being configured");
        // Add all elements contained within "configuration" 
        for(Object configuration:rootConfigElement.cloneContent() ) {
            if ( configuration instanceof Element ) {
                Element configElement = (Element)configuration; 
                configOpElement.addContent(configElement);
            }
        }
    }
    
    
    public String getConfigTypesString() {
        
        List<String> types= getConfigTypes();
        StringBuilder typesStr = new StringBuilder();
        for( String name : types){
            if (typesStr.length()>0){
                typesStr.append(", ");
            }
            typesStr.append(name);
        }
        
        return typesStr.toString();
    }
    
    public String getConfigName(){
        List<String> names = getConfigNames();
        if (names.size()==0)
            throw new IllegalStateException("No XML configuration files has been added to the Config object");
        if (names.size()>1)
            throw new IllegalStateException("Config object contains multiple configurations. " +
            		"We don't know what name to return");
        return names.get(0);
    }
    
    public List<String> getConfigNames(){
        List<Element> configRoots = (List<Element>)configOpElement.getChildren();
        List<String> names = new ArrayList<String>();
        for( Element config : configRoots ){
            // TODO: this check should be done during parsing
            String name = config.getAttributeValue("name");
            if (name==null)
                throw new DPBuddyException("Configuration "+config.getName()+" is missing 'name' attribute." +
                		" This attributed is required as per DataPower schema.");
            names.add(name);
        }
        return names;
    }

    public List<String> getConfigTypes(){
        List<Element> configRoots = (List<Element>)configOpElement.getChildren();
        List<String> types = new ArrayList<String>();
        for( Element config : configRoots ){
            types.add(config.getName());
        }
        return types;
    }


    /**
     * Adds the content of the provided configure request
     * to this object
     * @param conf
     */
    public void addConfigureRequest( ConfigCommand conf ){
        Element fromRoot = conf.getConfigOpElement();
        for( Element elt:(List<Element>)fromRoot.getChildren()) {
            configOpElement.addContent( (Element)elt.clone());
        }
    }
    
    private Element createConfigOpElement(){
        configOpElement = new Element(operation.getName(), DPConst.DP_NS);
        getRequestElement().addContent(configOpElement);
        return configOpElement;
    }
    
    private Element getConfigOpElement() {
        return configOpElement;
    }
 
    
    
    public ResultResponse execute( Device device ) {

        if ( getRequestElement().getChildren().size()==0)
            throw new DPBuddyException("You must provide at least one configuration item");
        
        Document responseDoc = device.executeRequest( this );
        ResultResponse dpResponse = new ResultResponse( responseDoc );
        dpResponse.validate();
        return dpResponse;
    }

    public String toXmlString() {
        return JDomUtils.docToString( getDocument() );    
    }

    public static enum ConfigOperation {
        SET("set-config"),
        MODIFY("modify-config");
        
        private String name;
        
        ConfigOperation(String name) {
            this.name = name;
        }

        public String getName() {
            return name;
        }
    }
    
    
}
