package com.myarch.dpbuddy.config;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.BaseRequest;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.ResultResponseItem;
import com.myarch.dpbuddy.status.ObjectStatus;

/**
 * Deletes configuration
 */
public class DelConfigCommand implements DPCommand<ResultResponse> {

    private Log logger=LogFactory.getLog( getClass() );

    private List<DPObject> objectPatterns=new ArrayList<DPObject>();
    private List<DPObject> excludePatterns=new ArrayList<DPObject>();
    
    // do not error out if nothing is found
    private boolean isQuiet=false;
    private boolean isDryRun=false;

    private BaseRequest request=new BaseRequest();
    private Element delConfigElt;
     
    
    public DelConfigCommand() {
        this( false );
    }

    public DelConfigCommand(boolean isQuiet) {
        this.isQuiet=isQuiet;
        delConfigElt=request.addRequestChild( "del-config" );
    }

    public boolean isDryRun() {
        return isDryRun;
    }

    public void setDryRun( boolean isDryRun ) {
        this.isDryRun=isDryRun;
    }

    public void addObjectPattern( DPObject dpObject ) {
        objectPatterns.add( dpObject );
    }

    public void addObjectPatterns( List<DPObject> dpObjects ) {
        objectPatterns.addAll( dpObjects );
    }

    public void setExcludePatterns( List<DPObject> dpObjects ) {
        excludePatterns=dpObjects;
    }

    public ResultResponse execute( Device device ) {
        if (objectPatterns.size() == 0)
            throw new DPBuddyException(
                            "You must provide at least one object or object pattern to delete" );

        logger.info( "Querying the device to determine what objects to delete ..." );

        // second param mustMatch. Has to be true if NOT quiet. in this case,
        // the exc is raised if at least one patter did not match
        Set<ObjectStatus> objectsToDelete=device.fetchObjects( objectPatterns, excludePatterns, !isQuiet );
        // this could only happen if quiet=true
        if (objectsToDelete.size() == 0) {
            logger.info( String
                            .format( "Object patterns that you provided (%s) did not match any existing DataPower objects, there is nothing to delete",
                                            DPObject.objectListToString( objectPatterns ) ) );
            return null;
        }

        logger.info( String.format( "The following objects will be deleted:\n%s",
                        DPObject.objectListToString( objectsToDelete ) ) );

        if (isDryRun) {
            logger.info( "Dry-run mode is set, the objects won't actually be deleted." );
            return null;
        }


        ResultResponse dpResponse=performDelete(device, objectsToDelete);
        List<ResultResponseItem> responseErrorItems=dpResponse.getErrorResultItems();
        
        List<DPObject> objectsToDeleteList=new ArrayList<DPObject>();
        objectsToDeleteList.addAll( objectsToDelete );
        
        StringBuilder errorMsgBuf=new StringBuilder();
        for (ResultResponseItem errorItem : responseErrorItems) {
            errorMsgBuf.append( String.format( "Deletion of '%s' failed",
                            objectsToDeleteList.get( errorItem.getIndex() ) ) );
            
            if (errorItem.getLogEventsStr() != null) {
                errorMsgBuf.append( " with the error: " + errorItem.getLogEventsStr() );
            }
            errorMsgBuf.append( "\n" );
        }
        
        if (errorMsgBuf.length() > 0) {
            if (isQuiet)
                logger.info( "Encountered the following errors during delete; ignoring because 'quiet' is set to True:\n"+errorMsgBuf  );
            else
                throw new DPBuddyException( errorMsgBuf.toString() );           
        }
            


        return dpResponse;
    }
    
  

    private ResultResponse performDelete(Device device,  Set<ObjectStatus> objectsToDelete) {

        ResultResponse dpResponse;
        Set<ObjectStatus> inUseObjectsToDelete = new HashSet<ObjectStatus> ();

        do {
            if (inUseObjectsToDelete.size()>0) {
                objectsToDelete.clear();
                objectsToDelete.addAll( inUseObjectsToDelete );
                logger.info( String.format( "The following objects were referenced by other objects and will be retried:\n%s",
                                DPObject.objectListToString( objectsToDelete ) ) );
            }
                
            addObjectElts( delConfigElt, objectsToDelete );
            Document responseDoc=device.executeRequest( request );
            
            dpResponse=new ResultResponse( responseDoc );
    
            List<ResultResponseItem> resultItems=dpResponse.getResultItems();
            int i=0;
            inUseObjectsToDelete.clear();
            for (ObjectStatus dpObject:objectsToDelete) {
                ResultResponseItem resultItem=resultItems.get( i );
                if (isInUseByAnotherObjectError(resultItem)) {
                    logger.debug( "Encountered 'in use' error, will attempt to delete this object again: "+dpObject.toString() );
                    inUseObjectsToDelete.add( dpObject );
                }
                // if any other failure
                else if (! resultItem.isSuccess())
                    break;
                
                ++i;
            }
            // we didn't delete anything, give up
            if (inUseObjectsToDelete.size()>0 && inUseObjectsToDelete.size()==objectsToDelete.size() ) {
                logger.debug( "Didn't delete any objects, giving up. Objects to delete count: "+ inUseObjectsToDelete.size());
                break;
            }
        }
        while (inUseObjectsToDelete.size()>0);
        
        return dpResponse;
    }
    
    
    
    private boolean isInUseByAnotherObjectError(ResultResponseItem errorItem) {
        String errorMessage=errorItem.getLogEventsStr();
        return ( !errorItem.isSuccess() && errorMessage!=null && StringUtils.containsIgnoreCase( errorMessage, "in use by:"));
    }
    
    private void addObjectElts( Element parentElt, Set<? extends DPObject> objectsToDelete ) {
        
        parentElt.removeContent();
        
        for (DPObject dpObj : objectsToDelete) {
            Element objElt=new Element( dpObj.getClazz() );
            objElt.setAttribute( "name", dpObj.getName() );

            parentElt.addContent( objElt );
        }
    }

}
