package com.myarch.dpbuddy.config;

import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.tools.ant.BuildException;
import org.jdom2.Element;
import org.kohsuke.args4j.Option;

import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.ObjectListParser;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.audit.AuditEvent;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.config.ConfigCommand.ConfigOperation;
import com.myarch.dpbuddy.status.ObjectStatus;
import com.myarch.hocon.HoconUtils;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import com.typesafe.config.ConfigList;
import com.typesafe.config.ConfigObject;
import com.typesafe.config.ConfigValue;
import com.typesafe.config.ConfigValueType;

public class UpdateConfigAntTask extends BaseDPBuddyTask {

    public UpdateConfigAntTask() {
        super();
        // Enforce schema validation
        isValidateAttr=true;
    }

    
    private String domainPatternsString;

    @Option(name="-domainPatterns", aliases="-domains",
                    usage="Comma-delimited list of regular expression patterns defining what domains to apply the command to. Use '.*' for all domains except the default. Use '.*,default' to include the default.")
    public void setDomainPatterns( String domainPatterns ) {
        this.domainPatternsString=domainPatterns;
    }
    
    @Override
    protected String getDomainPatterns() {
        return domainPatternsString;
    }    
    
    private String namePattern;
    private String classPattern;
    private List<DPObject> objects=new ArrayList<>();
    private List<DPObject> excludeObjects=new ArrayList<>();

    @Option(name="-namePattern", aliases="-name", usage="Regular expression defining what objects to modify")
    public void setNamePattern( String namePattern ) {
        this.namePattern=namePattern;
    }

    @Option(name="-classPattern", aliases="-class", usage="Regular expression defining what classes (types) to modify.")
    public void setClassPattern( String classPattern ) {
        this.classPattern=classPattern;
    }

    @Option(name="-objects",
                    usage="Comma-delimited list of object patterns to update in the format 'classPatter:namePattern, classPatter:namePattern'.")
    public void setObjects( String includeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( includeObjectsList );
        objects.addAll( objs );
    }

    @Option(name="-exclude",
                    usage="Comma-delimited list of object patterns to exclude in the format 'classPatter:namePattern, classPatter:namePattern'.")
    public void setExclude( String excludeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( excludeObjectsList );
        excludeObjects.addAll( objs );
    }
    
    
    private String hoconStr;

    @Option(name="-updates", required=true,
                    usage="Elements to modify in HOCON format, e.g., DoHostRewrite:on, Priority: normal, HTTPCompression: off. Elements must match the DataPower schema, see the corresponding Modify type definition in xml-mgmt.xsd")
    public void setUpdates( String hoconStr ) {
        this.hoconStr=hoconStr;
    }

    private boolean isDryRun=false;

    @Option(name="-dryRun", handler=BoolOptionHandler.class,
                    usage="Do not update anything, just print the matched objects.")
    public void setDryRun( boolean isDryRun ) {
        this.isDryRun=isDryRun;
    }

    @Override
    public void executeDPTask() throws BuildException {

        if (namePattern == null && classPattern == null && objects.isEmpty()) {
            throw new BuildException( "You must provide namePattern or classPattern or 'objects'" );
        }

        if (isDryRun) {
            logger.info( "Executing 'udateConfig' in 'dry-run' mode, no changes will be made" );
        }

        if (namePattern != null || classPattern != null) {
            objects.add( new DPObject( classPattern, namePattern ) );
        }

        logger.info( "Querying the device to determine what objects to modify" );
        Set<ObjectStatus> objectsToModify=getDevice().fetchExistingObjects( objects, excludeObjects );

        Config modifications=ConfigFactory.parseString( hoconStr );
        modifications=modifications.resolve();
        
        ConfigOperation op=ConfigOperation.MODIFY;
        ConfigCommand configCommand=new ConfigCommand( op );

        logger.info( String.format( "The following objects will be updated:\n%s",
                        DPObject.objectListToString( objectsToModify ) ) );
        logger.info( "The following configuration elements will be updated:\n"
            + HoconUtils.configToString( modifications ) );

        for (ObjectStatus dpObj : objectsToModify) {
            DPConfiguration dpConfiguration=buildDpConfiguration( dpObj.getClazz(), dpObj.getName(), modifications );
            configCommand.addConfiguration( dpConfiguration );
        }

        if (configCommand.isEmpty())
            throw new BuildException( "Invalid config" );

        if (!isDryRun) {
            executeRequest( configCommand );
            AuditEvent auditEvent=createAuditEvent( "updateConfig" );
            auditEvent.setObjects( objectsToModify );
            if (auditEventHandler != null) {
                auditEventHandler.handleAuditEvent( auditEvent, getDevice() );
            }

            saveDomainConfigIfEnabled();
        }
    }

    private DPConfiguration buildDpConfiguration( String clazz, String objectName, Config hoconModifications ) {

        Element classNameElt=new Element( clazz );
        classNameElt.setAttribute( "name", objectName );
        addHoconEntriesToElement( classNameElt, hoconModifications.root() );

        DPConfiguration dpConfiguration=new DPConfiguration();
        dpConfiguration.addElement( classNameElt );

        return dpConfiguration;
    }

    private void addHoconEntriesToElement( Element parentElt, ConfigObject hoconObject ) {
        Set<Entry<String, ConfigValue>> rootEntries=hoconObject.entrySet();
        
        for (Entry<String, ConfigValue> entry : rootEntries) {
            ConfigValue configVal=entry.getValue();
            addHoconConfigValToElement(parentElt, entry.getKey(), configVal);
        }
    }
    
    private void addHoconConfigValToElement( Element parentElt, String hoconKey, ConfigValue configVal) {
        if (HoconUtils.isNull( configVal )) {
            Element nameValElt=new Element( hoconKey );
            parentElt.addContent( nameValElt );
        }
        else if (configVal.valueType() == ConfigValueType.OBJECT) {
            ConfigObject configObj=(ConfigObject) configVal;
            Element newParentElt=new Element( hoconKey );
            addHoconEntriesToElement( newParentElt, configObj );
            parentElt.addContent( newParentElt );
        }
        else if (configVal.valueType() == ConfigValueType.LIST) {
            List<ConfigValue> configValues=(ConfigList) configVal;
            Element newParentElt=new Element( hoconKey );
            parentElt.addContent( newParentElt );
            for(ConfigValue configListVal:configValues) {
                // NOTE: currently only objects are supported inside arrays
                if (configListVal.valueType() == ConfigValueType.OBJECT) {
                    addHoconEntriesToElement( newParentElt, (ConfigObject)configListVal );    
                }
            }
        }        
        else if (HoconUtils.isPrimitiveType( configVal )) {
            String val=configVal.unwrapped().toString();
            Element nameValElt=new Element( hoconKey );
            nameValElt.setText( val );
            parentElt.addContent( nameValElt );
        }
       
    }
    

}
