package com.myarch.dpbuddy.configreport;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.tools.ant.BuildException;
import org.jdom2.Document;
import org.kohsuke.args4j.Option;

import com.myarch.dpbuddy.*;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.export.ExportCommand;
import com.myarch.dpbuddy.export.ExportObject;
import com.myarch.dpbuddy.filemanagement.FileResponse;
import com.myarch.dpbuddy.xmltransform.DeleteTransformAction;
import com.myarch.dpbuddy.xmltransform.TransformAction;
import com.myarch.dpbuddy.xmltransform.TransformActionContainer;
import com.myarch.dpbuddy.xmlutil.DpObjectXmlSorter;
import com.myarch.dpbuddy.xmlutil.XMLInput;
import com.myarch.dpbuddy.xmlutil.XMLParsingUtils;

public class ConfigReportTask extends BaseDPBuddyTask {

    public final static String DEFAULT_EXPORT_FILE_NAME="export.xml";
    public final static String DEFAULT_EXCEL_FILE_NAME="DataPowerConfigurationReport.xlsx";

    private File toDir;
    private String reportFileName=DEFAULT_EXCEL_FILE_NAME;
    private String exportFileName=DEFAULT_EXPORT_FILE_NAME;

    private Boolean persisted;

    private boolean refObjects=true;
    private boolean includeDebug=false;

    private boolean diff=false;
    private boolean isExcel=false;
    private boolean isCleanToDir=true;

    // TODO: to complete
    private boolean isSaveExport=true;

    private List<DPPattern> namePatterns=new ArrayList<>();

    private List<DPObject> objectPatterns=new ArrayList<>();
    private List<DPObject> excludePatterns=new ArrayList<>();

    private List<String> includeElements=new ArrayList<>();
    private List<String> excludeElements=new ArrayList<>(Arrays.asList( "GeneratedPolicies" ));

    private TransformActionContainer transformActionContainer;

    private boolean isDefaultTransform=true;

    private String namePattern;
    private String classPattern;

    private boolean sortDpObjects=false;

    private String domainPatternsString;

    @Option(name="-domainPatterns", aliases="-domains",
                    usage="Comma-delimited list of regular expression patterns defining what domains to apply the command to. Use '.*' for all domains except the default. Use '.*,default' to include the default.")
    public void setDomainPatterns( String domainPatterns ) {
        this.domainPatternsString=domainPatterns;
    }

    @Option(name="-includeElements",
                    usage="Comma-delimited list of configuration elements to include into the report. Defaults to all elements.")
    public void setIncludeElements( String elementNames ) {
        this.includeElements=Arrays.asList( StringUtils.splitByWholeSeparator( elementNames, "," ) );
    }

    @Option(name="-excludeElements", usage="Comma-delimited list of configuration elements to exclude from the report.")
    public void setExcludeElements( String elementNames ) {
        this.excludeElements=Arrays.asList( StringUtils.splitByWholeSeparator( elementNames, "," ) );
    }

    @Override
    protected String getDomainPatterns() {
        return domainPatternsString;
    }

    // nested transform element, using Ant conventions
    public TransformActionContainer createTransform() {
        transformActionContainer=TransformActionContainer.createTransformOnlyIfNull( transformActionContainer );
        return transformActionContainer;

    }

    private File getFileToExportTo( Device device ) {
        File f=new File( toDir, device.getName() );
        f=new File( f, device.getCurrentDomain() );
        f=new File( f, exportFileName );

        return f;
    }

    private DPObjectsReportMaker dpReportMaker;

    protected void onStart() {
        dpReportMaker=new DPObjectsReportMaker( includeElements, excludeElements );
        if (isCleanToDir) {
            try {
                logger.info("Cleaning directory "+toDir.getAbsolutePath()+" before generating reports");
                FileUtils.deleteDirectory( toDir );
            }
            catch (IOException e) {
                logger.warn( String.format( "Unable to delete directory '%s' prior to generating reports", e, toDir ) );
            }
        }
    }

    // TODO: constants
    private ConfReport confReport=new ConfReport( Arrays.asList( "Device", "Domain", "Name" ) );

    @Override
    public void executeDPTask() throws BuildException {

        File file=getFileToExportTo( getDevice() );

        ExportCommand exportCommand=new ExportCommand();

        // we need to provide the file so we can determine the export type
        exportCommand.setToFile( file.getAbsolutePath() );

        if (persisted != null) {
            exportCommand.setPersisted( persisted );
        }

        // options on the command line or attributes (as opposed to nested
        // objects)
        if (namePattern != null || classPattern != null || !objectPatterns.isEmpty()) {
            ExportObject whatToExport=new ExportObject( classPattern, namePattern, refObjects, false, includeDebug );
            whatToExport.setObjects( objectPatterns );
            whatToExport.setExclude( excludePatterns );
            exportCommand.addExportObject( whatToExport );
        }

        // nested name patterns
        if (namePatterns.size() > 0) {
            for (DPPattern dpPattern : namePatterns) {
                exportCommand.addExportObject( new ExportObject( null, dpPattern.toString() ) );
            }
        }

        if (exportCommand.getExportObjects().size() == 0) {
            logger.debug( "No nested exportObject element was provided; all objects will be included into export" );
            exportCommand.addExportObject( ExportObject.createAllObjectsExportObject() );
        }
        try {
            FileResponse exportResponse=exportCommand.execute( getDevice() );
            transformAndSave( exportResponse, file );

            if (sortDpObjects) {
                logger.info( "Sorting DataPower objects ..." );
                DpObjectXmlSorter sorter=new DpObjectXmlSorter();
                sorter.sort( file, file );
            }
        }
        catch (DPObjectsNotFoundException notFoundEx) {
            logger.info( "Didn't find any matching objects" );
        }

    }

    private void transformAndSave( FileResponse exportResponse, File file ) {
        XMLInput transformedXml=exportResponse.getFileAsXML();
        TransformActionContainer currentTransformActionContaner=transformActionContainer;
        if (isDefaultTransform)
            currentTransformActionContaner=injectDefaultTransformActions();

        if (currentTransformActionContaner != null) {
            logger.debug( "Transforming exported configuration ..." );
            transformedXml.transform( currentTransformActionContaner );
        }
        transformedXml.removeIntroducedNamespaces();
        if (isSaveExport) {
            XMLParsingUtils.persistXML( file, transformedXml.getDocument() );
        }
        populateReport( confReport, transformedXml.getDocument(), getDevice() );

    }

    private void populateReport( ConfReport confReport, Document dpXml, Device device ) {
        dpReportMaker.populateReports( confReport, dpXml, getDevice() );
    }

    @Override
    protected void onCompletion() {
        File output=toDir;
        ReportRenderer reportRenderer=new CsvReportRenderer();
        if (isExcel) {
            output=new File( toDir, reportFileName );
            reportRenderer=new ExcelReportRenderer();
        }
        reportRenderer.render( confReport, output, diff );
    }

    public final static String FILES_XPATH="/datapower-configuration/files";
    public final static String IFACE_XPATH="/datapower-configuration/interface-data";
    public final static String EXPORT_DETAILS_XPATH="/datapower-configuration/export-details";
    public final static String XPATH_OR=" | ";

    private TransformActionContainer injectDefaultTransformActions() {
        TransformActionContainer transformer=transformActionContainer;
        if (transformer == null) {
            transformer=new TransformActionContainer();
        }

        String xpath=EXPORT_DETAILS_XPATH + XPATH_OR + IFACE_XPATH + XPATH_OR + FILES_XPATH;
        TransformAction transformAction=new DeleteTransformAction( xpath );
        transformAction.setMatchRequired( false );
        transformer.insertFirst( transformAction );

        return transformer;
    }

    @Option(name="-objects",
                    usage="Comma-delimited list of object patterns to export in the format 'class:name, class:name'.")
    public void setObjects( String includeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( includeObjectsList );
        objectPatterns.addAll( objs );
    }

    @Option(name="-exclude",
                    usage="Comma-delimited list of object patterns to exclude from export in the format 'class:name, class:name'.")
    public void setExclude( String excludeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( excludeObjectsList );
        excludePatterns.addAll( objs );
    }

    @Option(name="-toDir", aliases="-dir", required=true,
                    usage="Folder/directory to save the reports and the exported files to.")
    public void setToDir( File dir ) {
        this.toDir=dir;
    }

    @Option(name="-reportFileName", aliases="-fileName", usage="File name of the report relative to 'toDir'. Excel report defaults to "
        + DEFAULT_EXCEL_FILE_NAME)
    public void setReportFileName( String fileName ) {
        this.reportFileName=fileName;
    }

    @Option(name="-persisted", handler=BoolOptionHandler.class, usage="Export only the persisted configuration.")
    public void setPersisted( Boolean persisted ) {
        this.persisted=persisted;
    }

    @Option(name="-namePattern", aliases="-name", usage="Regular expression defining what object names to export.")
    public void setNamePattern( String namePattern ) {
        this.namePattern=namePattern;
    }

    @Option(name="-classPattern", aliases="-class", usage="Regular expression defining what classes (types) to export.")
    public void setClassPattern( String classPattern ) {
        this.classPattern=classPattern;
    }

    @Option(name="-defaultTransform", handler=BoolOptionHandler.class,
                    usage="Apply default transformations to the exported XML configuration.")
    public void setDefaultTransform( boolean isDefaultTransform ) {
        this.isDefaultTransform=isDefaultTransform;
    }

    @Option(name="-sortObjects", handler=BoolOptionHandler.class, usage="Sort DataPower objects alphabetically.")
    public void setSortObjects( boolean sortDpObjects ) {
        this.sortDpObjects=sortDpObjects;
    }

    @Option(name="-includeDebug", handler=BoolOptionHandler.class, usage="Include probe data into export.")
    public void setIncludeDebug( final boolean includeDebug ) {
        this.includeDebug=includeDebug;
    }

    @Option(name="-diff", handler=BoolOptionHandler.class,
                    usage="Only show differences in the report. Defaults to false.")
    public void setDiff( final boolean isDiff ) {
        this.diff=isDiff;
    }

    @Option(name="-cleanToDir", handler=BoolOptionHandler.class,
                    usage="Remove all files and subdirectories in 'toDir' prior to generating reports. Defaults to true.")
    public void setCleanToDir( final boolean isCleanDir ) {
        this.isCleanToDir=isCleanDir;
    }

    @Option(name="-excel", handler=BoolOptionHandler.class,
                    usage="Produce report in Excel format (CSV is the default).")
    public void setExcel( final boolean isExcel ) {
        this.isExcel=isExcel;
    }

    @Option(name="-refObjects", handler=BoolOptionHandler.class,
                    usage="Include all objects referenced/required by exported objects.")
    public void setRefObjects( Boolean refObjects ) {
        this.refObjects=refObjects;
    }

}
