package com.myarch.dpbuddy.dpimport;

import java.io.File;

import org.apache.commons.lang3.StringUtils;
import org.apache.tools.ant.BuildException;
import org.jdom2.Document;
import org.kohsuke.args4j.Option;

import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.config.DPConfiguration;
import com.myarch.dpbuddy.config.DelConfigCommand;
import com.myarch.dpbuddy.quiesce.QuiesceAction;
import com.myarch.dpbuddy.quiesce.QuiesceUnquiesceDomainCommand;
import com.myarch.dpbuddy.status.StatusCommand;
import com.myarch.dpbuddy.status.StatusResponse;
import com.myarch.dpbuddy.xmltransform.ReplaceTextTransformAction;
import com.myarch.dpbuddy.xmlutil.XMLInput;

public class CreateDomainAntTask extends BaseDPBuddyTask {

    public final static String DOMAIN_TEMPLATE_RESOURCE="/com/myarch/dpbuddy/dpconfig/Domain.xml";

    private File configFile;
    private int maxCheckpoints=3;
    private boolean isDelete=false;

    // fudge because of data power bug
    private int delayAfterUnquiesce=1000;

    @Option(name="-file", required=false, usage="Domain configuration to import in xml/xcfg file.")
    public void setFile( File configFile ) {
        this.configFile=configFile;
    }

    @Option(name="-maxCheckpoints", required=false,
                    usage="Maximum  number of checkpoints for the domain")
    public void setMaxCheckpoints( int maxCheckpoints ) {
        this.maxCheckpoints=maxCheckpoints;
    }

    @Option(name="-delete", handler=BoolOptionHandler.class,
                    usage="If the domain with the same name already exists, delete it.")
    public void setDelete( boolean isDelete ) {
        this.isDelete=isDelete;
    }

    public void setDelayAfterUnquiesce( int delayAfterUnquiesce ) {
        this.delayAfterUnquiesce=delayAfterUnquiesce;
    }

    @Override
    public void executeDPTask() throws BuildException {

        XMLInput templateXml=null;
        if (configFile == null) {
            templateXml=XMLInput.fromResource( DOMAIN_TEMPLATE_RESOURCE );
        }
        else {
            templateXml=new XMLInput( configFile );
            templateXml.setName( configFile.getAbsolutePath() );
        }

        templateXml.resolveVars( envPropSet.getPropExpressionEvaluator() );

        DPConfiguration domainConfig=new DPConfiguration( templateXml );
        logger.debug( String.format( "Setting number of checkpoints for the domain to %d",
                        maxCheckpoints ) );
        domainConfig.setElementValue( "MaxChkpoints", new Integer( maxCheckpoints ).toString() );

        String domainNameToReplace=domainConfig.getName();

        String newDomain=getDevice().getCurrentDomain();
        logger.debug( String.format( "Replacing domain name '%s' with the name '%s'",
                        domainNameToReplace, newDomain ) );

        XMLInput targetDomainXml=replaceDomainName( templateXml, newDomain, domainNameToReplace );
        targetDomainXml.setName( templateXml.getName() );
        ImportCommand importCommand=new ImportCommand();
        importCommand.setFile( targetDomainXml );

        if (isDelete)
            deleteDomain( newDomain );
        
        if (logger.isDebugEnabled())
            logger.debug( "Creating domain using the following configuration:\n"+targetDomainXml.getXMLAsString() );
        
        importCommand.execute( getDevice().switchDefaultDomain() );
        if (isDelete) {
            reuiesceDomainIfNeeded( newDomain );
        }

        saveDomainConfigIfEnabled();
    }

    private XMLInput replaceDomainName( XMLInput domainTemplate, String newDomainName,
                    String domainNameToReplace ) {
        ReplaceTextTransformAction replacer=new ReplaceTextTransformAction();
        replacer.setTextToReplace( domainNameToReplace );
        replacer.setReplaceWith( newDomainName );

        Document transformedDoc=replacer.transform( domainTemplate.getDocument() );

        return new XMLInput( transformedDoc );
    }

    private void deleteDomain( String domainToDelete ) {
        DelConfigCommand delConfigCommand=new DelConfigCommand( true );
        delConfigCommand.addObjectPattern( new DPObject( "Domain", domainToDelete ) );
        delConfigCommand.execute( getDevice().switchDefaultDomain() );
    }

    private void reuiesceDomainIfNeeded( String domainName ) {
        if (isDomainQuiesced( domainName )) {
            logger.info( "The newly created domain is in the quiesced state. DPBuddy has to re-quiesce it, otherwise it could remain in an inconsistent state" );
            QuiesceUnquiesceDomainCommand udCommand=new QuiesceUnquiesceDomainCommand(
                            QuiesceAction.UNQUIESCE, domainName, null );
            udCommand.execute( getDevice() );

            try {
                Thread.sleep( delayAfterUnquiesce );
            }
            catch (InterruptedException e) {
                logger.debug( "Delay didn't complete!", e );
            }

            QuiesceUnquiesceDomainCommand qdCommand=new QuiesceUnquiesceDomainCommand(
                            QuiesceAction.QUIESCE, domainName, null );
            qdCommand.execute( getDevice() );
        }
    }

    private boolean isDomainQuiesced( String domainName ) {
        StatusCommand sc=new StatusCommand( "DomainStatus" );
        StatusResponse statusResponse=sc.execute( getDevice().switchDefaultDomain() );
        String quiesceState=statusResponse.getStatusValue( "Domain", domainName, "QuiesceState" );

        // <QuiesceState>quiesced</QuiesceState>
        if (StringUtils.startsWithIgnoreCase( quiesceState, "qu" ))
            return true;
        else
            return false;
    }
    
    @Override
    protected boolean isDefaultDomainOverride(){
        return true;
    }
}
