package com.myarch.dpbuddy.dpimport;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.types.Reference;
import org.kohsuke.args4j.Option;

import com.myarch.antutil.AntUtils;
import com.myarch.antutil.TaskParameterValidator;
import com.myarch.antutil.ZipHelper;
import com.myarch.dpbuddy.*;
import com.myarch.dpbuddy.action.ActionCommand;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.audit.AuditEvent;
import com.myarch.dpbuddy.audit.AuditEventFile;
import com.myarch.dpbuddy.checkpoint.CheckpointService;
import com.myarch.dpbuddy.checkpoint.SaveCheckpointCommand;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.config.ConfigCommand;
import com.myarch.dpbuddy.config.ConfigCommand.ConfigOperation;
import com.myarch.dpbuddy.config.DPConfiguration;
import com.myarch.dpbuddy.config.DPObjectConfiguration;
import com.myarch.dpbuddy.inpututil.Input;
import com.myarch.dpbuddy.inpututil.InputFileWrapper;
import com.myarch.dpbuddy.inpututil.InputUtils;
import com.myarch.dpbuddy.quiesce.QuiesceAction;
import com.myarch.dpbuddy.quiesce.QuiesceUnquiesceDomainCommand;
import com.myarch.dpbuddy.status.AssertOpenPortsAntTask;
import com.myarch.dpbuddy.status.AssertOperationalStateCommand;
import com.myarch.dpbuddy.status.AssertStateAntTask;
import com.myarch.dpbuddy.status.OperationalState;
import com.myarch.dpbuddy.xmltransform.TransformActionContainer;
import com.myarch.dpbuddy.xmltransform.TransformablePatternSet;
import com.myarch.dpbuddy.xmltransform.UpdateValuesTransformAction;
import com.myarch.dpbuddy.xmlutil.XMLInput;

/**
 * Imports configuration (XML or zip) the device
 */
public class ImportAntTask extends BaseDPBuddyTask {

    private File importFile;
    // used by subtasks
    private XMLInput importXmlInput;

    private File deploymentPolicyFile;
    private String deploymentPolicyName;
    private String deploymentPolicyVariablesName;

    private Boolean overwriteFiles;
    private Boolean overwriteObjects;
    private Boolean rewriteLocalIp;
    private Boolean dryRun;

    private boolean isAssertObjectsUp=false;

    private boolean transformOnly=false;

    private boolean resolveVars=false;

    private File workDir;

    private List<TransformablePatternSet> transformFiles=new ArrayList<TransformablePatternSet>();
    private TransformActionContainer transformActionContainer;

    private String checkpoint;

    private boolean reset=false;

    private boolean isQuiesce=false;
    private boolean isUnQuiesce=false;

    private String domainComments;

    private String portsToAssert;

    private boolean failOnError=true;
    private boolean reimportOnError=false;

    private String transformDefsFile;
    private String transformId;
    private String updatesHoconStr;
    private String applyUpdatesToHoconStr;
    
    private String domainPatternsString;
    
    @Option(name="-domainPatterns", aliases="-domains",
                    usage="Comma-delimited list of regular expression patterns defining what domains to apply the command to. Use '.*' for all domains except the default. Use '.*,default' to include the default.")
    public void setDomainPatterns( String domainPatterns ) {
        this.domainPatternsString=domainPatterns;
    }
    
    @Override
    protected String getDomainPatterns() {
        return domainPatternsString;
    }      
    
    // nested transform element, using Ant conventions
    public TransformActionContainer createTransform() {
        transformActionContainer=TransformActionContainer.createTransformOnlyIfNull( transformActionContainer );
        transformActionContainer.setProject( getProject() );
        return transformActionContainer;
    }

    // fileset
    public void addTransformFiles( TransformablePatternSet update ) {
        transformFiles.add( update );
    }

    @Override
    public void executeDPTask() throws BuildException {

        if (importXmlInput == null) {
            TaskParameterValidator validator=new TaskParameterValidator( this );
            validator.validateExistingFile( importFile, "file", "XML or ZIP file to import" );
            validator.throwIfErrors();
        }

        if (transformOnly) {
            logger.info( "Executing 'import' in 'transform-only' mode, files will be transformed, but the import command won't be executed" );
        }

        if (resolveVars && transformActionContainer == null) {
            // create an empty transform so we can resolve the vars
            createTransform();
        }

        Input fileInput=determineInputAndTransform();
        ImportCommand importCommand=createImportCommand( fileInput );

        if (!transformOnly) {
            performImportForDevice( importCommand, fileInput, getDevice() );
        }
        
        InputUtils.close( fileInput );
    }

    
    
    private ImportCommand createImportCommand( Input fileInput ) {

        ImportCommand importCommand=new ImportCommand();
        importCommand.setFile( fileInput );

        setDeploymentPolicy( importCommand, deploymentPolicyFile, deploymentPolicyName, deploymentPolicyVariablesName );

        if (overwriteFiles != null)
            importCommand.setOverwriteFiles( overwriteFiles );
        if (overwriteObjects != null)
            importCommand.setOverwriteObjects( overwriteObjects );
        if (rewriteLocalIp != null)
            importCommand.setRewriteLocalIp( rewriteLocalIp );
        if (dryRun != null) {
            if (dryRun)
                logger.info( "Executing 'import' in 'dry-run' mode, no changes will be made" );
            importCommand.setDryRun( dryRun );
        }
        return importCommand;
    }

    private Input determineInputAndTransform() {
        addInlineTranformsToTransformContainer();
        processTransformDefsFile();
        Input fileInput;
        if (!transformFiles.isEmpty() || transformActionContainer != null) {

            if (importXmlInput != null)
                fileInput=processTransforms( importXmlInput );
            else
                fileInput=processTransforms( importFile );

        }
        else {
            if (importXmlInput == null)
                fileInput=new InputFileWrapper( importFile );
            else
                fileInput=importXmlInput;
        }

        return fileInput;
    }

    
    private void processTransformDefsFile() {
        if (transformDefsFile != null) {
            
            AntUtils.importFile( getProject(), transformDefsFile );
            if (transformActionContainer!=null) {
                throw new DPBuddyException("You can't combine transform element and transform definition file");
            }
            createTransform();
            if (transformId==null) {
                throw new DPBuddyException("You must provide 'transformId' if you're using transform definition file");
            }
            logger.info( String.format( "Using transform definition '%s' from file '%s'",transformId, transformDefsFile) );
            Reference ref=new Reference(getProject(), transformId);
            
            transformActionContainer.setRefid( ref );
        }
    }

    private void addInlineTranformsToTransformContainer(){
        if (updatesHoconStr!=null) {
            // create transform container if doesn't exist
            createTransform();
            UpdateValuesTransformAction action = new UpdateValuesTransformAction( updatesHoconStr, applyUpdatesToHoconStr );
            transformActionContainer.addUpdateValues( action );
            transformActionContainer.setVerbose( true );
        }
        
    }
    
    
    private void performImportForDevice( ImportCommand importCommand, Input fileInput, Device device ) {

        if (isQuiesce) {
            QuiesceUnquiesceDomainCommand quiesceCommand=new QuiesceUnquiesceDomainCommand( QuiesceAction.QUIESCE,
                            device.getCurrentDomain(), null );
            quiesceCommand.execute( device );
        }

        if (numberOfOldCheckpointToDelete > 0) {
            logger.info( "Will try to remove up to " + numberOfOldCheckpointToDelete + " oldest checkpoints" );
            CheckpointService.deleteOldCheckpoints( getDevice(), numberOfOldCheckpointToDelete );
        }

        if (checkpoint != null && !isDryRun()) {
            SaveCheckpointCommand checkpointCmd=new SaveCheckpointCommand( checkpoint );
            checkpointCmd.execute( device );
        }

        if (reset && !isDryRun()) {
            logger.info( "Resetting target domain, domain's configuration will be deleted ..." );
            ActionCommand actionCommand=new ActionCommand( "ResetThisDomain" );
            actionCommand.execute( device );
        }

        try {
            // in case if we have audit, don't fail right away
            importCommand.setFailOnError( false );
            ImportResponse response=importCommand.execute( device );
            if (reimportOnError && response.getFailedObjects().size() > 0) {
                logger.info( "Re-running import because of 'reimportOnError=true'" );
                if (fileInput instanceof InputFileWrapper) {
                    // FIXME
                    ( (InputFileWrapper) fileInput ).reopen();
                }
                importCommand=createImportCommand( fileInput );
                response=importCommand.execute( device );
            }

            if (auditEventHandler != null) {
                AuditEvent auditEvent=createImportAuditEvent( response );
                auditEventHandler.handleAuditEvent( auditEvent, device );
            }

            if (!failOnError && response.getFailedObjects().size() > 0) {
                logger.info( "Ignoring import failures because of 'failOnError=false'" );
            }
            if (failOnError) {
                response.throwIfFailure();
            }

            if (isUnQuiesce) {
                QuiesceUnquiesceDomainCommand quiesceCommand=new QuiesceUnquiesceDomainCommand( QuiesceAction.UNQUIESCE,
                                device.getCurrentDomain(), null );
                quiesceCommand.execute( device );
            }

            if (isAssertObjectsUp) {
                AssertOperationalStateCommand assertObjectsUpCommand=new AssertOperationalStateCommand(
                                OperationalState.UP );
                String ignoreList=getProject().getProperty( AssertStateAntTask.IGNORE_OBJECTS_PROP_NAME );
                if (ignoreList != null) {
                    List<DPObject> objs=ObjectListParser.parsePatternList( ignoreList );
                    assertObjectsUpCommand.setExcludePatterns( objs );
                }
                assertObjectsUpCommand.execute( getDevice() );
            }

            if (StringUtils.isNotBlank( portsToAssert )) {
                assertPorts( getDevice(), portsToAssert );
            }

            if (domainComments != null && !isDryRun()) {
                updateDomainComments( getDevice() );
            }

            if (!isDryRun())
                saveDomainConfigIfEnabled();

        }
        catch (DPBuddyException e) {
            if (!isDryRun())
                e.setRollback( true );
            throw e;
        }

    }

    private boolean isDryRun() {
        return dryRun != null && dryRun == true;
    }

    private AuditEvent createImportAuditEvent( ImportResponse importResponse ) {
        AuditEvent auditEvent=createAuditEvent( "import" );
        List<? extends DPObject> importedObjects=importResponse.getImportedObjects();
        auditEvent.setObjects( importedObjects );
        auditEvent.setArchiveFile( new AuditEventFile( importFile, null ) );

        List<ImportedFile> importedFiles=importResponse.getFiles();
        for (ImportedFile importedFile : importedFiles) {
            auditEvent.addFile( new AuditEventFile( importedFile.getPath() ) );
        }
        return auditEvent;
    }

    private Input processTransforms( File file ) {

        Input updatedInputFile=null;
        if (!ExportImportFileType.isZipExtension( file )) { // XML
            if (transformFiles.size() > 0)
                throw new DPBuddyException(
                                "You can use 'transformFiles' only when you're importing ZIP files. Use 'transform' when importing XML" );

            if (transformActionContainer != null) {
                transformActionContainer.setDPConnection( getDevice().getDPConnection() );
                logger.debug( String.format( "Transforming import file '%s' using provided XPath overrides",
                                file.getAbsoluteFile() ) );
                updatedInputFile=fileTransformer.transformXMLFile( file, transformActionContainer );
            }

        }
        else {
            updatedInputFile=processZip( file );
        }

        return updatedInputFile;
    }

    private Input processTransforms( XMLInput xmlInput ) {
        Input updatedInputFile=null;
        if (transformFiles.size() > 0)
            throw new DPBuddyException(
                            "You can use 'transformFiles' only when you're importing ZIP files. Use 'transform' when importing XML" );

        if (transformActionContainer != null) {
            transformActionContainer.setDPConnection( getDevice().getDPConnection() );
            logger.debug( String.format( "Transforming import file '%s' using provided XPath overrides",
                            xmlInput.getName() ) );
            updatedInputFile=fileTransformer.transformXMLInput( xmlInput, transformActionContainer );
        }
        return updatedInputFile;
    }

    private Input processZip( File zipFile ) {
        if (transformActionContainer != null && transformFiles.size() > 0)
            throw new DPBuddyException( "Specifying both 'transformFiles' and 'transform' is not allowed" );

        Input transformedZipFile=null;

        if (transformActionContainer != null) {
            TransformablePatternSet tranformExportPatternSet=new TransformablePatternSet();
            tranformExportPatternSet.setIncludes( "export.xml" );

            transformActionContainer.setDPConnection( getDevice().getDPConnection() );
            tranformExportPatternSet.setTransform( transformActionContainer );

            List<TransformablePatternSet> tranformExportPatternSets=new ArrayList<TransformablePatternSet>();
            tranformExportPatternSets.add( tranformExportPatternSet );

            transformedZipFile=transformZipFile( zipFile, tranformExportPatternSets );
        }
        else if (transformFiles.size() > 0) {
            for (TransformablePatternSet transformablePatternSet : transformFiles) {
                TransformActionContainer transformActionContainer=transformablePatternSet.getTransform();
                if (transformActionContainer == null)
                    throw new DPBuddyException( "You must provide nested 'transform' inside 'transformFiles'" );
                transformActionContainer.setDPConnection( getDevice().getDPConnection() );
            }
            transformedZipFile=transformZipFile( zipFile, transformFiles );
        }
        else {
            transformedZipFile=new InputFileWrapper( zipFile );
        }

        return transformedZipFile;
    }

    private Input transformZipFile( File zipFile, List<TransformablePatternSet> transformablePatternSets ) {
        File unzipDir=createUnzipDir();

        try {
            FileUtils.copyFileToDirectory( zipFile, unzipDir );
        }
        catch (IOException e) {
            throw new BuildException( e );
        }

        File updatableZipFile=new File( FilenameUtils.concat( unzipDir.getAbsolutePath(), zipFile.getName() ) );

        ZipHelper zipHelper=new ZipHelper( getProject() );
        zipHelper.unzip( updatableZipFile, unzipDir );

        fileTransformer.transformAndPersistPatternSets( unzipDir, transformablePatternSets );

        zipHelper.zip( unzipDir, updatableZipFile );

        return new InputFileWrapper( updatableZipFile );
    }

    private File createUnzipDir() {
        if (workDir == null) {
            workDir=new File( FilenameUtils.concat( System.getProperty( "java.io.tmpdir" ), "dpbuddy/import" ) );
        }

        File unzipDir=workDir;

        try {
            FileUtils.deleteQuietly( unzipDir );
            FileUtils.forceMkdir( unzipDir );
        }
        catch (IOException e) {
            throw new BuildException( e );
        }

        return unzipDir;
    }

    private void updateDomainComments( Device device ) {
        ConfigCommand cmd=new ConfigCommand( ConfigOperation.MODIFY );

        DPConfiguration domainConf=new DPConfiguration();
        DPObjectConfiguration domain=new DPObjectConfiguration( "Domain", device.getCurrentDomain() );
        domain.setComments( domainComments );
        domainConf.addObject( domain );
        cmd.addConfiguration( domainConf );

        Device defaultDomainDevice=device.switchDefaultDomain();

        logger.info( "Setting domain's comments to '" + domainComments + "'" );

        cmd.execute( defaultDomainDevice );

        if (isSave()) {
            defaultDomainDevice.saveConfiguration();
        }
    }

    private void assertPorts( Device device, String ports ) {

        logger.info( String.format( "Checking the state of the following port(s): %s ...", ports.trim() ) );

        AssertOpenPortsAntTask.assertPorts( device, ports );

        logger.info( String.format( "Port(s) %s is open on the device", ports.trim() ) );
    }

    @Option(name="-file", required=true, usage="File to import. This could be a zip or an xml/xcfg file.")
    public void setFile( File file ) {
        this.importFile=file;
    }

    public File getImportFile() {
        return importFile;
    }

    protected void setImportXmlInput( XMLInput importXmlInput ) {
        this.importXmlInput=importXmlInput;
    }

    @Option(name="-deploymentPolicyFile", usage="Path to the deployment policy file on the local filesystem.")
    public void setDeploymentPolicyFile( File deploymentPolicyFile ) {
        this.deploymentPolicyFile=deploymentPolicyFile;
    }

    @Option(name="-deploymentPolicyName",
                    usage="Name of the existing deployment policy to use for the import operation.")
    public void setDeploymentPolicyName( String deploymentPolicy ) {
        this.deploymentPolicyName=deploymentPolicy;
    }

    @Option(name="-deploymentPolicyVars", aliases="-dpVars",
                    usage="Name of the deployment policy variables object to be applied to this deployment policy.")
    public void setDeploymentPolicyVars( String deploymentPolicyVariablesName ) {
        this.deploymentPolicyVariablesName=deploymentPolicyVariablesName;
    }

    @Option(name="-overwriteFiles", handler=BoolOptionHandler.class,
                    usage="Overwrite existing files with the files included with the import.")
    public void setOverwriteFiles( Boolean overwriteFiles ) {
        this.overwriteFiles=overwriteFiles;
    }

    @Option(name="-overwriteObjects", handler=BoolOptionHandler.class,
                    usage="Overwrite existing objects with the objects defined as part of this import.")
    public void setOverwriteObjects( Boolean overwriteObjects ) {
        this.overwriteObjects=overwriteObjects;
    }

    @Option(name="-rewriteLocalIp", handler=BoolOptionHandler.class,
                    usage="If set to 'true', the local address bindings of services contained in the import package will be rewritten to match the equivalent interfaces of the device.")
    public void setRewriteLocalIp( Boolean rewriteLocalIp ) {
        this.rewriteLocalIp=rewriteLocalIp;
    }

    @Option(name="-dryRun", handler=BoolOptionHandler.class, usage="Run import in the 'dry run' mode.")
    public void setDryRun( Boolean dryRun ) {
        this.dryRun=dryRun;
    }

    @Option(name="-workDir", usage="Path to the directory used to hold temporary files.")
    public void setWorkDir( File workDir ) {
        this.workDir=workDir;
    }
    
    @Option(name="-transformOnly", usage="Run transformations of the configuration, but do not execute import.", handler=BoolOptionHandler.class)
    public void setTransformOnly( boolean transformOnly ) {
        this.transformOnly=transformOnly;
    }

    @Option(name="-assertObjectsUp", handler=BoolOptionHandler.class,
                    usage="If set to 'true', upon completing the import, DPBuddy will invoke assertState task to verify that all objects in the domain are in 'up' state.")
    public void setAssertObjectsUp( boolean isAssertObjectsUp ) {
        this.isAssertObjectsUp=isAssertObjectsUp;
    }

    @Option(name="-resolveVars", handler=BoolOptionHandler.class,
                    usage="Resolve variables (${}) inside the import file.")
    public void setResolveVars( boolean resolveVars ) {
        this.resolveVars=resolveVars;
    }

    @Option(name="-checkpoint", aliases="-cp", usage="Create the checkpoint with this name before the import.")
    public void setCheckpoint( String checkpoint ) {
        this.checkpoint=checkpoint;
    }

    private int numberOfOldCheckpointToDelete=-1;

    @Option(name="-removeCheckpoints", aliases="-removeCPs", usage="Number of old checkpoints to delete.")
    public void setRemoveCheckpoints( int numberOfOldCheckpointToDelete ) {
        this.numberOfOldCheckpointToDelete=numberOfOldCheckpointToDelete;
    }

    @Option(name="-reset", handler=BoolOptionHandler.class, usage="Reset the domain's configuration before the import.")
    public void setReset( boolean reset ) {
        this.reset=reset;
    }

    @Option(name="-quiesce", handler=BoolOptionHandler.class, usage="Quiesce the domain before the import.")
    public void setQuiesce( boolean quiesce ) {
        this.isQuiesce=quiesce;
    }

    @Option(name="-unquiesce", handler=BoolOptionHandler.class, usage="Un-quiesce the domain after the import.")
    public void setUnquiesce( boolean unQuiesce ) {
        this.isUnQuiesce=unQuiesce;
    }

    @Option(name="-domainComments", usage="Comments, such as a version number, to be set at the domain level.")
    public void setDomainComments( String domainComment ) {
        this.domainComments=domainComment;
    }

    @Option(name="-assertPorts", usage="Comma-delimited list of ports to check after import.")
    public void setAssertPorts( String assertPorts ) {
        this.portsToAssert=assertPorts;
    }

    @Option(name="-failOnError", usage="Fail in case of import errors. Defaults to true.",
                    handler=BoolOptionHandler.class)
    public void setFailOnError( boolean failOnError ) {
        this.failOnError=failOnError;
    }

    @Option(name="-reimportOnError", usage="Re-run import in case of import errors.", handler=BoolOptionHandler.class)
    public void setReimportOnError( boolean reimportOnError ) {
        this.reimportOnError=reimportOnError;
    }

    @Option(name="-transformDefs", usage="File to load transform definitions from. Definitions are specified in XML (Apache Ant) format. "
        + "The file could contain multiple \"transform\" elements, each with its own \"id\" attribute.")
    public void setTransformDefs( String transformDefsFile ) {
        this.transformDefsFile=transformDefsFile;
    }
    
    @Option(name="-transformId", usage="ID of the transform definition as defined on the 'transform' element in the transform definition file.")
    public void setTransformId( String transformId ) {
        this.transformId=transformId;
    }
    
    @Option(name="-updates", usage="Updates to apply to elements of the objects specified by 'applyUpdatesTo'. HOCON format is used to specify element names and values, e.g., 'LocalPort:8090, RemoteEndpointHostname: myHost'. This means that the \"LocalPort\" element of all objects matched by \"applyUpdatesTo\" will be set to \"8090\"")
    public void setUpdates( String updatesHoconStr ) {
        this.updatesHoconStr=updatesHoconStr;
    }
    
    
    @Option(name="-applyUpdatesTo", usage="List of DataPower configuration class names and object names to update in HOCON format, e.g., 'XMLFirewallService: fw1, WSGateway: proxy2'. The actual updates (what element to update and the values) are defined by the \"updates\" option")
    public void setApplyUpdatesTo( String applyUpdatesToHoconStr ) {
        this.applyUpdatesToHoconStr=applyUpdatesToHoconStr;
    }    

}
