package com.myarch.dpbuddy.dpimport;

import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPConst;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.DeploymentPolicyRequest;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ExportImportFileType;
import com.myarch.dpbuddy.inpututil.Input;
import com.myarch.dpbuddy.inpututil.InputFileWrapper;
import com.myarch.dpbuddy.xmlutil.streaming.InputEncodingElementMapper;

/**
 * Performs import of the configuration and files. Represents do-import SOMA
 * request
 */
public class ImportCommand extends DeploymentPolicyRequest implements DPCommand<ImportResponse> {

    private Log logger=LogFactory.getLog( getClass() );

    private String inputName;
    private ExportImportFileType fileType;

    private Element doImportElement;

    private boolean isQuiet=false;

    private boolean failOnError=true;

    public final static String DO_IMPORT="do-import";
    public final static String DO_RESTORE="do-restore";

    public ImportCommand() {
        this( false );
    }

    public ImportCommand(boolean isBackupMode) {
        String eltName=DO_IMPORT;
        if (isBackupMode) {
            eltName=DO_RESTORE;
        }
        createRootDoElement( eltName );
    }

    public void setFile( String file ) {
        InputFileWrapper inputFile=new InputFileWrapper( file );
        setFile( inputFile );

    }

    public void setFile( Input input ) {
        this.inputName=input.getName();
        // This is required in 3.8.2.3 firmware
        fileType=ExportImportFileType.fromFile( inputName );
        setSourceType( fileType );

        Element inputFileElement=createInputFileElement();

        InputEncodingElementMapper fileMapper=new InputEncodingElementMapper();
        fileMapper.add( inputFileElement, input );
        setStreamElementMapper( fileMapper );

    }

    public boolean isFailOnError() {
        return failOnError;
    }

    public void setFailOnError( boolean failOnError ) {
        this.failOnError=failOnError;
    }

    private Element createRootDoElement( String eltName ) {
        doImportElement=new Element( eltName, DPConst.DP_NS );
        setOverwriteObjects( true );

        getRequestElement().addContent( doImportElement );

        return doImportElement;
    }

    private Element createInputFileElement() {
        Element inputFileElement=new Element( "input-file", DPConst.DP_NS );
        doImportElement.addContent( inputFileElement );

        return inputFileElement;
    }

    public String toString() {
        StringBuilder buf=new StringBuilder();

        buf.append( super.toString() );
        return buf.toString();
    }

    private Boolean isOverwriteFiles;

    public void setOverwriteFiles( boolean overwriteFiles ) {
        doImportElement.setAttribute( "overwrite-files", Boolean.toString( overwriteFiles ) );
        isOverwriteFiles=overwriteFiles;
    }

    public void setOverwriteObjects( boolean overwriteObjects ) {
        doImportElement.setAttribute( "overwrite-objects", Boolean.toString( overwriteObjects ) );
    }

    public void setRewriteLocalIp( boolean rewriteLocalIp ) {
        doImportElement.setAttribute( "rewrite-local-ip", Boolean.toString( rewriteLocalIp ) );
    }

    public void setSourceType( ExportImportFileType fileType ) {
        doImportElement.setAttribute( "source-type", fileType.toString() );
    }

    public void setDryRun( boolean dryRun ) {
        doImportElement.setAttribute( "dry-run", Boolean.toString( dryRun ) );
    }

    public boolean isDryRun() {
        return Boolean.valueOf( doImportElement.getAttributeValue( "dry-run" ) );
    }

    public boolean isQuiet() {
        return isQuiet;
    }

    /**
     * If set to true, only the number of updated objects will be reported;
     * object classes/names and files won't be shown
     * 
     * @param isQuiet
     */
    public void setQuiet( boolean isQuiet ) {
        this.isQuiet=isQuiet;
    }

    private boolean isImportedFileZip() {
        return fileType == ExportImportFileType.ZIP;
    }

    public ImportResponse execute( Device device ) {
        if (inputName == null)
            throw new DPBuddyException( "You must provide a file to the import command" );

        String statusMessage=String.format( "Importing '%s'", inputName );
        if (hasDeploymentPolicy())
            statusMessage+=String.format( " using deployment policy '%s'", getDeploymentPolicyName() );
        logger.info( statusMessage );

        if (hasDeploymentPolicy()) {
            ImportCommand policyConfigCommand=getDeploymentPolicyConfigCommand();
            policyConfigCommand.execute( device );
        }

        setDefaults();

        Document doc=device.executeRequest( this );
        ImportResponse response=new ImportResponse( doc );

        if (isDryRun()) {
            logger.info( "Import was executed in 'dry-run' mode; status of configuration changes was not provided" );
        }
        else {
            response.validate();

            List<ImportedFile> ignoredFiles=response.getFilesForStatus( ImportedFile.Status.IGNORED );
            // don't print file status for XML import
            if (ignoredFiles.size() > 0 && isImportedFileZip()) {
                logIfNoisy( "Following files were skipped (most likely because thery were identical to the ones in the import file):\n"
                    + ImportedFile.fileListToString( ignoredFiles ) );
            }

            List<ImportedFile> changedFiles=response.getFilesForStatus( ImportedFile.Status.OK );
            if (changedFiles.size() > 0) {
                if (!isImportedFileZip())
                    logger.warn( "Because of 'overwriteFiles=\"true\"' used with XML import, zero byte-size files were created." );

                logIfNoisy( String.format( "%d file(s) were successfully created or updated:\n%s",
                                changedFiles.size(), ImportedFile.fileListToString( changedFiles ) ) );
            }

            List<ImportedObject> skippedObjs=response
                            .getObjectsForStatus( ImportedObject.Status.SKIPPED );
            if (skippedObjs.size() > 0)
                logger.warn( "Following objects were skipped during import:\n"
                    + DPObject.objectListToString( skippedObjs ) );

            List<ImportedObject> successObjs=response.getImportedObjects();
            String successObjMsg=String.format( "%d object(s) were imported", successObjs.size() );
            logger.info( successObjMsg );
            
            if (successObjs.size() > 0) {
                logIfNoisy( ImportedObject.objectListToString( successObjs ));
            }

            List<ImportedObject> failedObjs=response
                            .getObjectsForStatus( ImportedObject.Status.FAILURE );
            if (failedObjs.size() > 0) {
                String failedObjMsg=String.format( "%d object(s) failed", failedObjs.size() );
                failedObjMsg+=":\n" + ImportedObject.objectListToString( failedObjs );
                logger.error( failedObjMsg );
                if (failOnError) {
                    response.throwIfFailure();
                }
            }

        }

        return response;
    }

    private void logIfNoisy( String msg ) {
        if (isQuiet) {
            logger.debug( msg );
        }
        else {
            logger.info( msg );
        }
    }

    private void setDefaults() {
        if (isOverwriteFiles == null) {
            if (isImportedFileZip())
                setOverwriteFiles( true );
            else
                setOverwriteFiles( false );
        }
    }

}
