package com.myarch.dpbuddy.export;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.tools.ant.BuildException;
import org.kohsuke.args4j.Option;
import org.kohsuke.args4j.spi.ExplicitBooleanOptionHandler;

import com.myarch.dpbuddy.*;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.filemanagement.FileResponse;
import com.myarch.dpbuddy.xmltransform.DeleteTransformAction;
import com.myarch.dpbuddy.xmltransform.TransformAction;
import com.myarch.dpbuddy.xmltransform.TransformActionContainer;
import com.myarch.dpbuddy.xmlutil.DpObjectXmlSorter;
import com.myarch.dpbuddy.xmlutil.XMLInput;
import com.myarch.dpbuddy.xmlutil.XMLParsingUtils;

public class ExportAntTask extends BaseDPBuddyTask {

    private File file;
    private File deploymentPolicyFile;
    private String deploymentPolicyName;
    private Boolean persisted;
    private Boolean allFiles;

    private boolean refObjects=true;
    private boolean refFiles=true;
    private boolean includeDebug=false;
    
    private String namePatternsString;
    private List<DPPattern> namePatterns=new ArrayList<>();
    
    private List<DPObject> objectPatterns=new ArrayList<>();
    private List<DPObject> excludePatterns=new ArrayList<>();
    
    private List<ExportObject> exportObjects=new ArrayList<>();

    private TransformActionContainer transformActionContainer;

    private boolean isDefaultTransform=true;

    private String namePattern;
    private String classPattern;
    
    private boolean sortDpObjects=false;

    // nested transform element, using Ant conventions
    public TransformActionContainer createTransform() {
        transformActionContainer=TransformActionContainer
                        .createTransformOnlyIfNull( transformActionContainer );
        return transformActionContainer;

    }

    @Override
    public void executeDPTask() throws BuildException {

        validator.validateRequired( this.file, "file",
                        "File to save the results of the export to (XML or ZIP)" );
        validator.throwIfErrors();
        ExportCommand exportCommand=new ExportCommand();

        if (transformActionContainer != null && ExportImportFileType.isZipExtension( file ))
            throw new DPBuddyException(
                            "You can specify 'transform' in the export task only if you're exporting configuration in XML format" );

        // we need to provide the file so we can determine the export type
        exportCommand.setToFile( file.getAbsolutePath() );

        setDeploymentPolicy( exportCommand, deploymentPolicyFile, deploymentPolicyName, null );

        if (persisted != null) {
            exportCommand.setPersisted( persisted );
        }

        if (allFiles != null) {
            exportCommand.setAllFiles( allFiles );
        }

        if (exportObjects.size() > 0) {
            exportCommand.addExportObjects( exportObjects );
        }

        // this is deprecated
        if (namePatternsString != null) {
            exportCommand.addExportObject( new ExportObject( null, namePatternsString ) );
        }

        // options on the command line or attributes (as opposed to nested objects)
        if (namePattern != null || classPattern != null || !objectPatterns.isEmpty() || !excludePatterns.isEmpty()) {
            ExportObject whatToExport=new ExportObject( classPattern, namePattern, refObjects, refFiles, includeDebug );
            
            if (StringUtils.isAllBlank( classPattern, namePattern) && objectPatterns.isEmpty() && !excludePatterns.isEmpty()) {
                objectPatterns.add( DPObject.ALL_OBJECTS );
            }
            whatToExport.setObjects( objectPatterns );
            whatToExport.setExclude( excludePatterns );
            exportCommand.addExportObject( whatToExport );
        }
        

        // nested name patterns
        if (namePatterns.size() > 0) {
            for (DPPattern dpPattern : namePatterns) {
                exportCommand.addExportObject( new ExportObject( null, dpPattern.toString() ) );
            }
        }

        if (exportCommand.getExportObjects().size() == 0) {
            logger.debug( "No nested exportObject element was provided; all objects will be included into export" );
            exportCommand.addExportObject( ExportObject.createAllObjectsExportObject() );
        }
        
        if (ExportImportFileType.isZipExtension( file ) || transformActionContainer == null
            && isDefaultTransform == false) {
            exportCommand.setSaveToFile( file );
            exportCommand.execute( getDevice() );
        }
        else {
            FileResponse exportResponse=exportCommand.execute( getDevice() );
            transformAndSave( exportResponse );
        }
        
        if (!ExportImportFileType.isZipExtension( file ) && sortDpObjects) {
            logger.info( "Sorting DataPower objects ..." );
            DpObjectXmlSorter sorter= new DpObjectXmlSorter();
            sorter.sort( file, file );
        }
 
    }

    private void transformAndSave( FileResponse exportResponse ) {
        XMLInput transformedXml=exportResponse.getFileAsXML();
        TransformActionContainer currentTransformActionContaner=transformActionContainer;
        if (isDefaultTransform)
            currentTransformActionContaner=injectDefaultTransformActions();
        
        if (currentTransformActionContaner!=null) {
            logger.info( "Transforming exported configuration ..." );
            transformedXml.transform( currentTransformActionContaner );
        }
        transformedXml.removeIntroducedNamespaces();
        XMLParsingUtils.persistXML( file, transformedXml.getDocument() );
        
    }

    public final static String FILES_XPATH="/datapower-configuration/files";
    public final static String IFACE_XPATH="/datapower-configuration/interface-data";
    public final static String EXPORT_DETAILS_XPATH="/datapower-configuration/export-details";
    public final static String XPATH_OR=" | ";
    
    private TransformActionContainer injectDefaultTransformActions() {
        TransformActionContainer transformer=transformActionContainer;
        if (transformer == null) {
            transformer=new TransformActionContainer();
        }
        
        String xpath=EXPORT_DETAILS_XPATH+XPATH_OR+IFACE_XPATH+XPATH_OR+FILES_XPATH;
        TransformAction transformAction=new DeleteTransformAction( xpath );
        transformAction.setMatchRequired( false );
        transformer.insertFirst( transformAction );

        return transformer;
    }
    
    @Option(name="-objects",
                    usage="Comma-delimited list of object patterns to export in the format 'class:name, class:name'.")
    public void setObjects( String includeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( includeObjectsList );
        objectPatterns.addAll( objs );
    }
    
    
    @Option(name="-exclude",
                    usage="Comma-delimited list of object patterns to exclude from export in the format 'class:name, class:name'.")
    public void setExclude( String excludeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( excludeObjectsList );
        excludePatterns.addAll( objs );
    }
    
    @Option(name="-allFiles", handler=ExplicitBooleanOptionHandler.class, usage="Export all files from the local:/ filesystem.")
    public void setAllFiles( Boolean allFiles ) {
        this.allFiles=allFiles;
    }

    public void addConfiguredExportObject( ExportObject exportObject ) {
        exportObjects.add( exportObject );
    }

    @Option(
                    name="-file",
                    required=true,
                    usage="Name of the file to save the exported configuration to. Must have zip, xml or xcfg extensions.")
    public void setFile( File file ) {
        this.file=file;
    }

    @Option(name="-persisted", handler=BoolOptionHandler.class, usage="Export only the persisted configuration.")
    public void setPersisted( Boolean persisted ) {
        this.persisted=persisted;
    }

    @Option(name="-deploymentPolicyFile",
                    usage="Path to the deployment policy file on the local machine")
    public void setDeploymentPolicyFile( File deploymentPolicy ) {
        this.deploymentPolicyFile=deploymentPolicy;
    }

    @Option(name="-deploymentPolicyName",
                    usage="Name of the deployment policy to include with the export/backup.")
    public void setDeploymentPolicyName( String deploymentPolicy ) {
        this.deploymentPolicyName=deploymentPolicy;
    }

    @Deprecated
    public void setNamePatterns( String namePatternsString ) {
        this.namePatternsString=namePatternsString;
    }

    @Deprecated
    public void addConfiguredNamePattern( DPPattern namePattern ) {
        namePatterns.add( namePattern );
    }

    @Option(name="-namePattern", aliases="-name", usage="Regular expression defining what object names to export.")
    public void setNamePattern( String namePattern ) {
        this.namePattern=namePattern;
    }


    @Option(name="-classPattern", aliases="-class", usage="Regular expression defining what classes (types) to export.")
    public void setClassPattern( String classPattern ) {
        this.classPattern=classPattern;
    }

    @Option(
                    name="-defaultTransform", handler=BoolOptionHandler.class,
                    usage="Apply default transformations to the exported XML configuration.")
    public void setDefaultTransform( boolean isDefaultTransform ) {
        this.isDefaultTransform=isDefaultTransform;
    }

    @Option(
                    name="-sortObjects", handler=BoolOptionHandler.class,
                    usage="Sort DataPower objects alphabetically.")
    public void setSortObjects( boolean sortDpObjects ) {
        this.sortDpObjects=sortDpObjects;
    }

    @Option(name="-includeDebug", handler=BoolOptionHandler.class, usage="Include probe data into export.")
    public void setIncludeDebug( final boolean includeDebug ) {
        this.includeDebug=includeDebug;
    }   

    @Option(name="-refObjects", handler=BoolOptionHandler.class, usage="Include all objects referenced/required by exported objects.")
    public void setRefObjects( Boolean refObjects ) {
        this.refObjects=refObjects;
    }

    @Option(name="-refFiles", handler=BoolOptionHandler.class, usage="Include all files referenced by exported objects")
    public void setRefFiles( Boolean refFiles ) {
        this.refFiles=refFiles;
    }
    
}
