package com.myarch.dpbuddy.export;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.*;
import com.myarch.dpbuddy.dpimport.ImportCommand;
import com.myarch.dpbuddy.filemanagement.FileResponse;
import com.myarch.dpbuddy.status.ObjectStatus;
import com.myarch.dpbuddy.xmlutil.FileSavingStreamParser;

/**
 * Exports configuration and files from the device. Represents "do-export" SOMA
 * operation.
 */
public class ExportCommand extends ExportBackupBase implements DPCommand<FileResponse> {

    private final Log logger=LogFactory.getLog( getClass() );

    private File saveToFile=null;

    private List<ExportObject> exportObjectPatterns=new ArrayList<ExportObject>();

    public ExportCommand() {
        super( "do-export" );
    }

    public void setSaveToFile( File saveToFile ) {
        this.saveToFile=saveToFile;
    }

    public void setExportObjects( List<ExportObject> exportObjects ) {
        exportObjectPatterns=exportObjects;
    }

    public void addExportObject( ExportObject exportObject ) {
        exportObjectPatterns.add( exportObject );
    }

    public void addExportObjects( List<ExportObject> exportObjects ) {
        exportObjectPatterns.addAll( exportObjects );
    }

    public List<ExportObject> getExportObjects() {
        return exportObjectPatterns;
    }

    private void addExportObjectXML( ExportObject exportObject ) {
        Element exportObjectElt=new Element( "object", DPConst.DP_NS );

        exportObjectElt.setAttribute( "class", exportObject.getClazz() );
        exportObjectElt.setAttribute( "name", exportObject.getName() );

        if (exportObject.getRefFiles() != null)
            exportObjectElt.setAttribute( "ref-files", exportObject.getRefFiles().toString() );

        if (exportObject.getRefObjects() != null)
            exportObjectElt.setAttribute( "ref-objects", exportObject.getRefObjects().toString() );

        if (exportObject.getIncludeDebug() != null)
            exportObjectElt.setAttribute( "include-debug", exportObject.getIncludeDebug().toString() );

        getOperationElement().addContent( exportObjectElt );
    }

    public void setAllFiles( boolean allFiles ) {
        getOperationElement().setAttribute( "all-files", new Boolean( allFiles ).toString() );
    }

    public String getExportObjectDisplayList() {
        StringBuilder buf=new StringBuilder();
        for (Element elt : getOperationElement().getChildren( "object", DPConst.DP_JDOM_NS )) {
            if (buf.length() > 0)
                buf.append( ", " );
            buf.append( elt.getAttributeValue( "class" ) + ":" + elt.getAttributeValue( "name" ) );
        }

        return buf.toString();
    }

    public FileResponse execute( Device device ) {
        if (needsToQuery())
            logger.info( "Querying the device to determine objects to export" );
        addExportObjectsFromPatterns( device );
        logger.info( String.format( "Exporting configuration to '%s'", getToFile() ) );
        if (hasDeploymentPolicy()) {
            ImportCommand policyConfigCommand=getDeploymentPolicyConfigCommand();
            policyConfigCommand.execute( device );
        }
        Document responseDoc=null;
        if (saveToFile != null) {
            FileSavingStreamParser parser=new FileSavingStreamParser( saveToFile );
            responseDoc=device.executeRequest( parser, this );
        }
        else {
            responseDoc=device.executeRequest( this );
        }

        FileResponse response=new FileResponse( responseDoc );
        response.validate();

        return response;
    }

    private boolean needsToQuery() {
        if (exportObjectPatterns.size() == 0
            || exportObjectPatterns.size() == 1 && exportObjectPatterns.get( 0 ).isAllObjects())
            return false;
        return true;
    }

    private void addExportObjectsFromPatterns( Device device ) {
        cleanExportObjectXML();
        
        for (ExportObject exportConfig : exportObjectPatterns) {
            if (exportConfig.isAllObjects()) {
                addExportObjectXML( exportConfig.copyAsAllObjectsExportObject() );
                logger.info( String.format( "All objects in the domain '%s' are going to be exported",
                                device.getCurrentDomain() ) );
                break;
            }
            else {
                Set<ObjectStatus> matchedObjects=findMatchingObjects( device, exportConfig );
                if (matchedObjects.isEmpty()) {
                    throw new DPObjectsNotFoundException( "Object patterns that you provided (%s) did not match any DataPower objects",
                                    DPObject.objectListToString( exportConfig.getObjects() ) );
                } 
                
                logger.info( String.format( "Following objects in the domain '%s' are going to be exported:",
                                device.getCurrentDomain() ) );
                logger.info( DPObject.objectListToString( matchedObjects ) );
                addExportObjectsFromMatches( exportConfig, matchedObjects );
            }
        }
    }

    private Set<ObjectStatus> findMatchingObjects( Device device, ExportObject exportConfig ) {
        boolean mustMatch=false;
        Set<ObjectStatus> matchingObjects=device.fetchObjects( exportConfig.getObjects(), exportConfig.getExclude(), mustMatch );
        return matchingObjects;
    }

    private void addExportObjectsFromMatches( ExportObject exportObjectPattern, Set<ObjectStatus> matchedObjects ) {
        
        for (ObjectStatus matchedObject : matchedObjects) {
            ExportObject objectToAdd=new ExportObject( matchedObject );
            // configure object that matched with the flags that were defined at
            // the exporetObject level
            objectToAdd.populateFlags( exportObjectPattern );
            addExportObjectXML( objectToAdd );
        }
    }

    private void cleanExportObjectXML() {
        getOperationElement().removeChildren( "object", DPConst.DP_JDOM_NS );
    }
}
