package com.myarch.dpbuddy.filemanagement;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.BaseRequest;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPConst;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.inpututil.Input;
import com.myarch.dpbuddy.xmlutil.streaming.InputEncodingElementMapper;

/**
 * Uploads (copies) files to datapower
 */
public class CopyFilesCommand extends BaseRequest implements DPCommand<ResultResponse> {
    
	private Log logger = LogFactory.getLog(getClass());
    
 
    private MkdirCommand createDirCommand;
    private RmdirCommand removeDirCommand;
    //private DeleteMultipleFilesCommand deleteFilesCommand;

    
    private boolean cleanDirectories;

    // suppress printing log file locations and other info
    private boolean isQuiet=false;
    
    
    public void setCleanDirectories(boolean deleteDirs) {
        this.cleanDirectories = deleteDirs;
    }
   
    
    public void setQuiet(boolean isQuiet) {
        this.isQuiet=isQuiet;
    }


    public void addFile(Input input, String targetDirOrPath) {
        addFile(input, targetDirOrPath, null);
    }
    
    
    private boolean isCleanRootWarningFlag=false;
    
    private InputEncodingElementMapper fileMapper=new InputEncodingElementMapper();
    
    public void addFile(Input input, String targetDir, String targetFileName){
        
        targetDir=Device.prependDefaultFileStore(targetDir);
        // handle directory cleaning
        if (! Device.isFileStoreRoot( targetDir ) ) {
            
            if (createDirCommand==null) {
                createDirCommand= new MkdirCommand();
                
                if (cleanDirectories) {
                    removeDirCommand=new RmdirCommand();
                    removeDirCommand.setIgnoreErrors(true);
                }
            }
            
            createDirCommand.addDir(targetDir);
            if (cleanDirectories) {
                removeDirCommand.addDir(targetDir);
            }
        }
        // filesystem root
        else if (cleanDirectories && !isCleanRootWarningFlag){
            String warningMsg=String.format( "Ignoring request to clean the root directory '%s' when running 'copy' with 'cleanDirectories=true'. Use 'delete' instead.", targetDir );
            logger.warn( warningMsg );
            isCleanRootWarningFlag=true;
        }

        String targetFileNameToUse=targetFileName;
        if (targetFileNameToUse==null)
            targetFileNameToUse=input.getName();

        Element setFileElement = createSetFileElement( getTargetFile(targetFileNameToUse, targetDir) );
        fileMapper.add( setFileElement, input);

    }

    
    private String getTargetFile( String sourceFile, String targetDir){
        if (!targetDir.endsWith("/"))
            targetDir+="/";
        
        // figure out the file name
        File fileF = new File( sourceFile );
        String fileName = fileF.getName();
        
        String targetFile = targetDir+fileName;
        return targetFile;
    }
    
    private Element createSetFileElement( String targetFile ){
        Element setFileElement = new Element("set-file", DPConst.DP_NS);
        setFileElement.setAttribute("name", targetFile);
        getRequestElement().addContent(setFileElement);
        
        return setFileElement;
    }
    
    public boolean hasFiles() {
        return getRequestElement().getChildren().size()> 0;
    }
   
    
    public String toString( ){
        StringBuilder buf=  new StringBuilder();
        if ( createDirCommand!=null)
            buf.append("Composite request\n"+createDirCommand.toString()+"\n");
        
        buf.append(super.toString());
            
        
        return buf.toString();
    }
    
    
    public int getNumberOfFiles() {
        List<Element> files = getRequestElement().getChildren("set-file", DPConst.DP_JDOM_NS);
        return files.size();
    }
    
    public List<String> getRemoteFilePaths() {
        List<Element> files = getRequestElement().getChildren("set-file", DPConst.DP_JDOM_NS);
        List<String> filePaths=new ArrayList<String>();
        for(Element fileElt:files){
            filePaths.add(fileElt.getAttributeValue("name"));
        }
        return filePaths;
    }
    
    public String getRemoteFilePath() {
        List<Element> files = getRequestElement().getChildren("set-file", DPConst.DP_JDOM_NS);
        List<String> filePaths=new ArrayList<String>();
        for(Element fileElt:files){
            filePaths.add(fileElt.getAttributeValue("name"));
        }
        
        if (filePaths.size()<1)
            new IllegalStateException("CopyFilesCommand does not have any files");
        if (filePaths.size()>1)
            new IllegalStateException("CopyFilesCommand contains multiple remote files, which one to return?");
        
        return filePaths.get( 0 );
    }

    public ResultResponse execute( Device device ) {
        if ( ! hasFiles() )
            throw new DPBuddyException("You must provide at least one file to upload");
        
        setStreamElementMapper(fileMapper);
        
        if (removeDirCommand!=null) {
            removeDirCommand.execute(device);
        }
        if (createDirCommand!=null) {
            createDirCommand.execute( device );
        }


        if (!isQuiet)
            logger.info("File location(s) on the device:\n"+StringUtils.join(getRemoteFilePaths(), ", "));
        
        // we use 2004 schema because it allows for multiple files
        Document responseDoc=device.executeRequest(this, true);
        ResultResponse dpResponse=new ResultResponse(responseDoc);
        dpResponse.validate();
        
        logger.info(String.format("Copied %d file(s) to the device", getNumberOfFiles()));
        
        return dpResponse;
    }
    
    
}
