package com.myarch.dpbuddy.filemanagement;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang3.time.DateParser;
import org.apache.commons.lang3.time.FastDateFormat;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Element;

public class DPFileOrDir {
    private Log logger = LogFactory.getLog(getClass());
    
    public final static String DIR_ELEMENT_NAME="directory";
    public final static String LOCATION_ELEMENT_NAME="location";
    public final static String FILE_ELEMENT_NAME="file";
    public final static String MODIFIED_ELEMENT_NAME="modified";

    private Element fileOrDirElt;

    /**
     * @param fileOrDirElt
     */
    public DPFileOrDir(Element fileOrDirElt) {
        super();
        validateFileOrDir( fileOrDirElt );
        this.fileOrDirElt=fileOrDirElt;

    }

    public static boolean isValidFileDirElement( Element fileOrDirElt ) {
        String eltName=fileOrDirElt.getName();
        return (eltName.equals( FILE_ELEMENT_NAME ) || eltName.equals( DIR_ELEMENT_NAME ) || eltName
                        .equals( LOCATION_ELEMENT_NAME ));

    }

    private void validateFileOrDir( Element fileOrDirElt ) {
        if (!isValidFileDirElement( fileOrDirElt ))
            throw new IllegalArgumentException(
                            "Element must be file, directory or location, you provided "
                                + fileOrDirElt.getName() );
    }
    
    
    private static final DateParser DATE_PARSER = FastDateFormat.getInstance("yyyy-MM-dd HH:mm:ss");
    
    public Date getModifiedTimestamp() {
        Date modifiedTimestamp=null;
        //<modified>2013-05-26 17:45:31</modified>
        Element modifiedElt=fileOrDirElt.getChild( MODIFIED_ELEMENT_NAME );
        if (modifiedElt!=null) {
            String dateStr=modifiedElt.getTextTrim();
            try {
                modifiedTimestamp=DATE_PARSER.parse( dateStr );
            }
            catch (ParseException e) {
                logger.warn( String.format("Unable to parse the timestamp %s of the file %s", dateStr, getPath() ));
            }
        }
        
        return modifiedTimestamp;
    }

    /**
     * Returns full path to the resource, including filestore name and the path
     * itself.
     * 
     * @return
     */
    public String getPath() {
        if (isDirectory())
            return getRawName();
        else
            return getFilePath();
    }

    private String getFilePath() {
        String localName=getRawName();

        DPFileOrDir parentDir=getParentDir();

        Element containingDir=fileOrDirElt.getParentElement();
        if (containingDir == null)
            throw new IllegalStateException( String.format( "File '%s' is not nested inside directory",
                            localName ) );
        return parentDir.getPath() + "/" + localName;
    }

    public DPFileOrDir getParentDir() {
        Element parentDirElt=fileOrDirElt.getParentElement();

        DPFileOrDir parentDir=null;

        if (parentDirElt == null && !isDirectory())
            throw new IllegalStateException( String.format( "File '%s' is not nested inside directory",
                            getRawName() ) );

        if (parentDirElt != null)
            parentDir=new DPFileOrDir( parentDirElt );

        return parentDir;
    }

    public List<DPFileOrDir> listFiles() {
        if (!isDirectory())
            throw new IllegalStateException( String.format(
                            "Attemt to get nested files from a non-directory '%s'", getRawName() ) );

        List<DPFileOrDir> filesAndDirs=new ArrayList<DPFileOrDir>();
        List<Element> fileElts=fileOrDirElt.getChildren( FILE_ELEMENT_NAME );
        for (Element fileElt : fileElts) {
            DPFileOrDir file=new DPFileOrDir( fileElt );
            filesAndDirs.add( file );
        }

        return filesAndDirs;
    }

    /**
     * 
     * @return True if directory. Filesystem root (local:/) is considered a
     *         directory.
     */
    public boolean isDirectory() {
        String dirAttrVal=fileOrDirElt.getAttributeValue( "list" );
        return (dirAttrVal != null && dirAttrVal.equals( "true" ));
    }

    public boolean isFileSystemRoot() {
        return isDirectory() && "location".equals( fileOrDirElt.getName() );
    }

    public boolean isFile() {
        return "file".equals( fileOrDirElt.getName() );
    }
    
    private String getRawName() {
        return fileOrDirElt.getAttributeValue( "name" );
    }

    public String toString() {
        return getPath();
    }
}
