package com.myarch.dpbuddy.filemanagement;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.Device;
import com.myarch.util.FilePathUtils;

public class DownloadMultipleFilesCommand implements DPCommand<List<FileResponse>> {

    private Log logger=LogFactory.getLog( getClass() );

    private String pattern=null;
    private File toDir;
    private boolean isFlatten=false;
    private boolean isDryRun=false;
    private boolean isMatchRequired=true;
    private boolean isCleanToDir=false;
    private boolean isPreserveTimestamp=true;
    private boolean isDownloadIfNewer=false;

    public DownloadMultipleFilesCommand(String pattern, File toDir, boolean isFlatten, 
                    boolean isPreserveTimestamp, boolean isDownloadIfNewer) {
        this.pattern=pattern;
        this.toDir=toDir;
        this.isFlatten=isFlatten;
        this.isPreserveTimestamp=isPreserveTimestamp;
        this.isDownloadIfNewer=isDownloadIfNewer;
    }

    public boolean isDryRun() {
        return isDryRun;
    }

    public void setDryRun( boolean isDryRun ) {
        this.isDryRun=isDryRun;
    }

    public void setMatchRequired( boolean isMatchRequired ) {
        this.isMatchRequired=isMatchRequired;
    }

    public void setCleanToDir( boolean cleanDirectory ) {
        this.isCleanToDir=cleanDirectory;
    }

    public List<FileResponse> execute( Device device ) {

        logger.info( "Obtaining the list of files to download using pattern '" + pattern + "'" );

        boolean filesOnly=true;
        List<DPFileOrDir> filesToDownload=device.fetchMatchingFileInfo( pattern, filesOnly );

        if (filesToDownload.size() > 0) {
            if (isDryRun)
                logger.info( "Dry-run mode is set, the files won't actually be downloaded. The following files would be downloaded:" );

            if (isCleanToDir && !isDryRun) {
                try {
                    FileUtils.deleteDirectory( toDir );
                }
                catch (IOException e) {
                    throw new DPBuddyException( "Unable to delete directory '%s' prior to downloading files", e,
                                    toDir );
                }
            }
        }

        List<FileResponse> responses=new ArrayList<FileResponse>();
        for (DPFileOrDir fileInfo : filesToDownload) {
            File saveToDir=createSaveToDir( fileInfo );
            if (isDryRun) {
                logger.info( fileInfo.getPath() );
            }
            else {
                FileResponse downloadResponse=downloadFile( device, fileInfo, saveToDir );
                if (downloadResponse != null) {
                    responses.add( downloadResponse );
                }
            }
        }

        if (filesToDownload.size() == 0) {
            String noFilesMessage=String.format( "Didn't find any files to download using pattern '%s'", pattern );
            if (isMatchRequired)
                throw new DPBuddyException( noFilesMessage );
            else
                logger.info( noFilesMessage );
        }

        return responses;
    }

    private File createSaveToDir( DPFileOrDir dpFile ) {
        File dirToDownload=toDir;

        if (!isFlatten) {
            String dpPath=dpFile.getPath();
            dpPath=FilenameUtils.getFullPath( FilePathUtils.removePrefix( dpPath ) );
            String localPath=toDir.getPath() + dpPath;
            dirToDownload=new File( localPath );
        }

        if (!dirToDownload.exists() && !isDryRun) {
            try {
                logger.debug( "Creating directory " + dirToDownload.getPath() + " to save the file to" );
                FileUtils.forceMkdir( dirToDownload );
            }
            catch (IOException e) {
                throw new DPBuddyException( "Unable to create directory '%s' to save files to", e, dirToDownload );
            }
        }

        return dirToDownload;
    }

    private FileResponse downloadFile( Device device, DPFileOrDir dpFile, File dirToSaveTo ) {
        DownloadFileCommand downloadFileCmd=new DownloadFileCommand();
        String destFileName=FilenameUtils.concat( dirToSaveTo.getAbsolutePath(),
                        FilenameUtils.getName( dpFile.getPath() ) );
        File destFile=new File( destFileName );

        Date dpModifiedTimestamp=dpFile.getModifiedTimestamp();
        Date destFileModifiedTimestamp=new Date( destFile.lastModified() );
        FileResponse fileDownloadResponse=null;
        if (!isDownloadIfNewer || !destFile.exists() || dpModifiedTimestamp.after( destFileModifiedTimestamp )) {
            downloadFileCmd.setToFile( destFileName );
            downloadFileCmd.setFromFile( dpFile.getPath() );

            fileDownloadResponse=downloadFileCmd.execute( device );

            if (isPreserveTimestamp) {
                destFile.setLastModified( dpModifiedTimestamp.getTime() );
            }
        }
        else if (isDownloadIfNewer) {
            logger.info( String.format( "File '%s' was not downloaded because it didn't change", dpFile.getPath() ) );
        }
        return fileDownloadResponse;
    }

}
