package com.myarch.dpbuddy.filemanagement.ant;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.tools.ant.DirectoryScanner;
import org.kohsuke.args4j.Option;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.audit.AuditEvent;
import com.myarch.dpbuddy.audit.AuditEventFile;
import com.myarch.dpbuddy.cache.FlushXMLCacheAntTask;
import com.myarch.dpbuddy.cache.XMLCacheType;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.filemanagement.CopyFilesCommand;
import com.myarch.dpbuddy.inpututil.Input;
import com.myarch.dpbuddy.inpututil.InputFileWrapper;

public class CopyFilesTask extends BaseDPBuddyTask {

    private boolean isCleanDirectories=false;
    private boolean isFlatten=false;
    private String toDir;

    private String includes;
    private String excludes;
    private File dir;

    private boolean isTransformOnly=false;

    private boolean isFlushCache=false;

    
    private String domainPatternsString;

    @Option(name="-domainPatterns", aliases="-domains",
                    usage="Comma-delimited list of regular expression patterns defining what domains to apply the command to. Use '.*' for all domains except the default. Use '.*,default' to include the default.")
    public void setDomainPatterns( String domainPatterns ) {
        this.domainPatternsString=domainPatterns;
    }
    
    @Override
    protected String getDomainPatterns() {
        return domainPatternsString;
    }      
    
    private List<TransformableFileSetWithDestinationDir> filesetList=new ArrayList<TransformableFileSetWithDestinationDir>();

    public void addDPFileset( TransformableFileSetWithDestinationDir fileset ) {
        filesetList.add( fileset );
    }

    @Option(name="-cleanDirectories", handler=BoolOptionHandler.class,
                    usage="Delete target directories on the device before uploading files.")
    public void setCleanDirectories( boolean cleanDirectories ) {
        this.isCleanDirectories=cleanDirectories;
    }

    @Option(
                    name="-flatten",
                    handler=BoolOptionHandler.class,
                    usage="If true, do not create any subdirectories; create only root directories on the device.")
    public void setFlatten( boolean flatten ) {
        this.isFlatten=flatten;
    }

    @Option(name="-toDir", usage="The target directory for uploaded files on the device.")
    public void setToDir( String toDir ) {
        this.toDir=toDir;
    }

    @Option(
                    name="-includes",
                    usage="Comma- or space-separated list of patterns of files that must be included in Ant pattern format, e.g, **/*.*")
    public void setIncludes( String includes ) {
        this.includes=includes;
    }

    @Option(
                    name="-excludes",
                    usage="Comma- or space-separated list of patterns of files that must be excluded in Ant pattern format, e.g, **/*.*")
    public void setExcludes( String excludes ) {
        this.excludes=excludes;
    }

    @Option(
                    name="-dir",
                    usage="The root of the directory tree of the local files; use it conjunction with -includes option.")
    public void setDir( File dir ) {
        this.dir=dir;
    }

    @Option(
                    name="-flushCache",
                    handler=BoolOptionHandler.class,
                    usage="Flush the document and stylesheet caches for all XML managers after all files are copied.")
    public void setFlushCache( boolean isFlushCache ) {
        this.isFlushCache=isFlushCache;
    }

    public void setTransformOnly( boolean transformOnly ) {
        this.isTransformOnly=transformOnly;
    }

    @Override
    public void executeDPTask() {
        if (isTransformOnly) {
            logger.info( "Executing 'copy' in 'transform-only' mode; files will be transformed, but the command won't be executed" );
        }

        CopyFilesCommand copyFilesCommand=new CopyFilesCommand();
        copyFilesCommand.setCleanDirectories( isCleanDirectories );

        AuditEvent copyAuditEvent=createAuditEvent( "copy" );
        if (( includes != null || excludes != null ) && dir == null)
            throw new DPBuddyException(
                            "You must specify 'dir' attribute when you provide 'include' or 'exclude'" );

        if (dir != null) {
            filesetList.add( createFileSetFromAttrs() );
        }

        for (TransformableFileSetWithDestinationDir fileset : filesetList) {
            DirectoryScanner dirScanner=fileset.getDirectoryScanner();

            String targetDir=fileset.getToDir();
            // if not specified at the fileset level, get it from the task level
            if (targetDir == null)
                targetDir=toDir;
            // if not specified at the fileset or task level, use the base dir
            // of the fileset
            if (targetDir == null)
                throw new DPBuddyException( "You must provide 'toDir' at the task or the fileset level" );

            String[] files=dirScanner.getIncludedFiles();
            if (files.length == 0 && fileset.isMatchRequired()) {
                throw new DPBuddyException(
                                "Could not find any files matching the fileset with the base dir '%s'",
                                fileset.getDirectoryScanner().getBasedir() );
            }

            for (String file : files) {
                String filePath=dirScanner.getBasedir() + System.getProperty( "file.separator" ) + file;

                logger.debug( "Adding this file to the upload request: " + filePath );
                String fileTargetDir=targetDir;
                if (!isFlatten) {
                    File fileObj=new File( file );
                    String dir=fileObj.getParent();
                    if (dir != null) {
                        dir=dir.replace( '\\', '/' );
                        fileTargetDir+="/" + dir;
                    }
                }
                copyAuditEvent.addFile( new AuditEventFile( new File( filePath ), fileTargetDir+"/"+file ) );

                logger.debug( "Target directory on the device: " + fileTargetDir );

                Input inputToUpload;
                File fileToUpload=new File( filePath );
                if (fileset.getTransform() != null) {
                    inputToUpload=fileTransformer.transformXMLFile( fileToUpload,
                                    fileset.getTransform(), getDevice().getDPConnection() );
                }
                else {
                    inputToUpload=new InputFileWrapper( fileToUpload );
                }
                copyFilesCommand.addFile( inputToUpload, fileTargetDir );
            }
        }

        if (isTransformOnly)
            return;

        if (copyFilesCommand.hasFiles()) {
            try {
                executeRequest( copyFilesCommand );
            }
            catch (DPBuddyException e) {
                e.setRollback( true );
                throw e;
            }

            if (isFlushCache)
                flushCache();
            if (auditEventHandler != null) {
                auditEventHandler.handleAuditEvent( copyAuditEvent, getDevice() );
            }
        }
        else
            logger.info( "No files matched any of the fileset(s); nohting will be copied to the device" );
    }

    private void flushCache() {
        FlushXMLCacheAntTask flushTask=new FlushXMLCacheAntTask();
        flushTask.setProject( getProject() );
        flushTask.performFlush( getDevice(), XMLCacheType.ALL );
    }

    private TransformableFileSetWithDestinationDir createFileSetFromAttrs() {
        File rootDir=new File( "./" );
        if (dir != null)
            rootDir=dir;
        TransformableFileSetWithDestinationDir filesetFromAttr=new TransformableFileSetWithDestinationDir();
        filesetFromAttr.setDir( rootDir );
        if (includes != null) {
            filesetFromAttr.setIncludes( includes );
        }
        if (excludes != null) {
            filesetFromAttr.setExcludes( excludes );
        }
        filesetFromAttr.setProject( getProject() );

        return filesetFromAttr;
    }

}
