package com.myarch.dpbuddy.filemanagement.ant;

import java.io.File;

import org.apache.tools.ant.BuildException;
import org.kohsuke.args4j.Option;

import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.cli.DPBuddyCliExecutor;
import com.myarch.dpbuddy.cli.DPBuddyMain;
import com.myarch.dpbuddy.filemanagement.DownloadMultipleFilesCommand;

/**
 * Downloads multiple files from DataPower based on regexp pattern.
 */
public class DownloadMultipleFilesTask extends BaseDPBuddyTask {

    private String pattern;
    private File toDir;
    private boolean cleanToDir=false;
    private boolean isMatchRequired=true;
    private boolean isFlatten=false;
    private boolean isDryRun=false;
    private boolean isPreserveTimestamp=true;
    private boolean isDownloadIfNewer=false;
    private boolean isCreateHostDir=false;

    @Option(name="-include", required=true,
                    usage="Regular expression pattern defining what files to download. The match is done against the entire path of a file, including the filesystem (e.g., local:/.* ). Note that the filesystem is separated by a single slash, i.e., the pattern local://.*xsd will not match anything.")
    public void setInclude( String pattern ) {
        this.pattern=pattern;
    }

    @Option(name="-toDir", required=true, usage="Local directory where to save the files. ")
    public void setToDir( File toDir ) {
        this.toDir=toDir;
    }

    // used from LS plugin
    public void setToDirStr( String toDir ) {
        this.toDir=new File( toDir );
    }

    @Option(name="-cleanToDir", handler=BoolOptionHandler.class,
                    usage="Remove all files and subdirectories in 'toDir' prior to downloading the files")
    public void setCleanToDir( boolean cleanDirectory ) {
        this.cleanToDir=cleanDirectory;
    }

    @Option(name="-matchRequired", handler=BoolOptionHandler.class,
                    usage="Fail if no files matched the 'include' pattern.")
    public void setMatchRequired( boolean isMatchRequired ) {
        this.isMatchRequired=isMatchRequired;
    }

    @Option(name="-dryRun", handler=BoolOptionHandler.class,
                    usage="Print the files that will be downloaded but don't actually download anything.")
    public void setDryRun( boolean isDryRun ) {
        this.isDryRun=isDryRun;
    }

    @Option(name="-flatten", handler=BoolOptionHandler.class,
                    usage="Download all files into 'toDir' instead of creating the directory tree replicating remote directory structure.")
    public void setFlatten( boolean isFlatten ) {
        this.isFlatten=isFlatten;
    }

    @Option(name="-preserveTimestamp", handler=BoolOptionHandler.class, usage="Preserve file timestamp from DataPower.")
    public void setPreserveTimestamp( boolean isPreserveTimestamp ) {
        this.isPreserveTimestamp=isPreserveTimestamp;
    }

    @Option(name="-createHostDir", handler=BoolOptionHandler.class,
                    usage="Create subdirectory under 'toDir' with the name matching DataPower host name.")
    public void setCreateHostDir( boolean isCreateHostDir ) {
        this.isCreateHostDir=isCreateHostDir;
    }

    @Option(name="-downloadIfChanged", handler=BoolOptionHandler.class,
                    usage="Download only a file is newer on DataPower")
    public void setDownloadIfChanged( boolean isDownloadIfNewer ) {
        this.isDownloadIfNewer=isDownloadIfNewer;
    }

    @Override
    public void executeDPTask() throws BuildException {

        validator.validateRequired( pattern, "pattern", "Regexp pattern defining what files to download" );
        validator.validateRequired( toDir, "to", "Local directory to download files to" );
        validator.throwIfErrors();

        DownloadMultipleFilesCommand downloadCommand=new DownloadMultipleFilesCommand( pattern, formToDirName(),
                        isFlatten, isPreserveTimestamp, isDownloadIfNewer );
        downloadCommand.setDryRun( isDryRun );
        downloadCommand.setMatchRequired( isMatchRequired );
        downloadCommand.setCleanToDir( cleanToDir );

        downloadCommand.execute( getDevice() );
    }

    public static final String HOST_SEPARATOR="__";

    private File formToDirName() {
        File fullToDir=toDir;
        if (isCreateHostDir) {
            String host=getDevice().getDPConnection().getHost();
            fullToDir=new File( fullToDir, host );
        }
        return fullToDir;
    }

    public static DownloadMultipleFilesTask instantiate( String configFilePath ) {
        DPBuddyCliExecutor commandFactory=new DPBuddyCliExecutor( DPBuddyMain.getCliCmdDefs() );
        BaseDPBuddyTask task=commandFactory.instantiateTask( "download" );
        DownloadMultipleFilesTask downloadTask=(DownloadMultipleFilesTask) task;

        if (configFilePath != null) {
            downloadTask.setConfigFile( configFilePath );
        }
        commandFactory.setProjectProperties( downloadTask );

        return downloadTask;
    }

}
