package com.myarch.dpbuddy.firmware;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.action.ActionCommand;
import com.myarch.dpbuddy.filemanagement.RemoteCopyCommand;
import com.myarch.dpbuddy.restart.DPPoller;
import com.myarch.dpbuddy.status.StatusCommand;
import com.myarch.dpbuddy.status.StatusResponse;
import com.myarch.util.ThreadUtils;

public class FirmwareCommand  {
    private final Log logger=LogFactory.getLog( getClass() );

    // reboot, reload, halt
    public final static String DEFAULT_RESTART_MODE="reload";
    public final static int DEFAULT_TIMEOUT=300;
    // ms
    public final static int DEFAULT_SHUTDOWN_WAIT_TIME=20000;

    private int shutdownWaitTime=DEFAULT_SHUTDOWN_WAIT_TIME;
    public final static String IMAGE_FILESTORE="image";

    public ResultResponse execute( Device device, String firmwareImageSourceUrl, String firmwareFileOnDP,
                    int timeout, String xmlManagerForHttpsCopy ) {

        Device defaultDomainDevice=device.switchDefaultDomain();

        String imageUrlOnDp;

        if (firmwareImageSourceUrl != null) {
            imageUrlOnDp=constructTargetUrl( firmwareImageSourceUrl, firmwareFileOnDP );
            remoteCopyImage( defaultDomainDevice, firmwareImageSourceUrl, imageUrlOnDp,
                            xmlManagerForHttpsCopy );
        }
        else if (StringUtils.isNotBlank( firmwareFileOnDP )) {
            imageUrlOnDp=Device.prependFileStore( IMAGE_FILESTORE, firmwareFileOnDP );
        }
        else {
            throw new IllegalArgumentException(
                            "Firmware file was not provided or can't be inferred from the source URL" );
        }

        // Validation fails if the format is not :///
        imageUrlOnDp=StringUtils.replace( imageUrlOnDp, ":/", ":///" );

        ActionCommand applyPatchAction=createAction( imageUrlOnDp );
        logger.info( String.format( "Applying firmware image '%s' ...", imageUrlOnDp ) );

        ResultResponse response=applyPatchAction.execute( defaultDomainDevice );
        response.validate();

        if (timeout < 0) {
            logger.info( "Negative timeout was provided, so we won't be waiting for DataPower to come back up" );
        }
        else {

            long startTime=System.nanoTime();

            logger.info( String.format( "Waiting for %d seconds for the reboot process to start ...",
                            DEFAULT_SHUTDOWN_WAIT_TIME / 1000 ) );
            ThreadUtils.delay( shutdownWaitTime );

            DPPoller poller=new DPPoller();
            logger.info( String.format( "Waiting for the device to come back up. "
                + "This will take some time (up to %d seconds timeout) ...", timeout ) );

            poller.pollUntilSuccess( defaultDomainDevice, timeout );
            long elapsedTime=System.nanoTime() - startTime;
            logger.debug( "Received IOException from the device in " + elapsedTime / 1000 / 1000
                + " ms" );

            logger.info( String.format(
                            "Restart has been completed in %d seconds, DataPower is up and running.",
                            elapsedTime / 1000 / 1000 / 1000 ) );

            StatusResponse statusResponse=retrieveVersionInfo( device );
            logger.info( "Firmware version reported by DataPower:" );
            logger.info( statusResponse.toStatusString() );
        }

        return response;
    }

    private StatusResponse retrieveVersionInfo( Device device ) {
        StatusCommand statusCommand=StatusCommand.VERSION_STATUS_COMMAND;
        return statusCommand.execute( device );
    }

    private ActionCommand createAction( String firmwareFileOnDP ) {
        ActionCommand actionCommand=new ActionCommand( "ApplyPatch" );
        actionCommand.addActionChildElement( "AcceptLicense", "true" );
        actionCommand.addActionChildElement( "File", firmwareFileOnDP );

        return actionCommand;
    }

    private void remoteCopyImage( Device device, String fromUrl, String dpFile, String xmlManager ) {
        RemoteCopyCommand remoteCopy=new RemoteCopyCommand();
        boolean force=true;
        logger.info("We need to copy the firmware image to DataPower first. This will take a few minutes");
        remoteCopy.execute( device, fromUrl, dpFile, force, xmlManager );
    }

    private String constructTargetUrl( String fromUrl, String dpFile ) {
        return RemoteCopyCommand.constructTargetUrl( fromUrl, IMAGE_FILESTORE, dpFile );
    }

}
