package com.myarch.dpbuddy.license;

import java.io.File;
import java.util.Date;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import javax.security.auth.x500.X500Principal;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.util.FilePathUtils;
import com.myarch.util.FilePreferences;

import de.schlichtherle.license.CipherParam;
import de.schlichtherle.license.DefaultCipherParam;
import de.schlichtherle.license.DefaultKeyStoreParam;
import de.schlichtherle.license.KeyStoreParam;
import de.schlichtherle.license.LicenseContent;
import de.schlichtherle.license.LicenseContentException;
import de.schlichtherle.util.ObfuscatedString;

public class DPBuddyLicenseManager extends DPBuddyTrialLicenseManager {
    private final Log logger=LogFactory.getLog( getClass() );

    public static final String LICENSE_FILE_NAME="dpbuddy-license";

    // this allows us to find the license in the same folder where all the jars
    // are (lib)
    public final static String DEFAULT_LICENSE_FILE_PATH=FilenameUtils.normalize( FilePathUtils
                    .getAbsoluteFilePath( "de/schlichtherle/license" )
        + "/../../../../"
        + LICENSE_FILE_NAME );

    private boolean trialLicenseWaringIssued=false;

    public DPBuddyLicenseManager() {
        super( new DPBuddyLicenseParam() );
    }

    /**
     * Validate license found in the default location, if it does not exist,
     * create a new trial license
     * 
     * @return validated license or the newly created trial one
     */
    public License validateDefaultLocationOrTrial() {

        LicenseContent licenseContent=null;

        File licenseFile=new File( DEFAULT_LICENSE_FILE_PATH );
        logger.debug( "Attempting to load a license file from the default location at "
            + DEFAULT_LICENSE_FILE_PATH );

        if (!licenseFile.exists()) {
            logger.debug( "License file was not found. Checking if there is an installed trial license" );
            licenseContent=validateTrialIfExists();
            // null means license was not found. In all other cases an exception
            // will be thrown
            if (licenseContent == null) {
                logger.debug( "There is no existing trial license. Will issue a new one." );

                licenseContent=issueTrialLicense();
                logger.debug( "Issued a new trial license with the expiration date "
                    + licenseContent.getNotAfter() );
            }
            else
                logger.debug( "Found a trial license with the expiration date "
                    + licenseContent.getNotAfter() );
        }
        else {
            return validateLicenseFromFile( licenseFile );
        }

        return new License( licenseContent );
    }

    public License validateLicenseFromFile( File licenseFile ) {
        LicenseContent licenseContent=null;
        if (!licenseFile.exists())
            throw new LicenseException( "License file '%s' was not found", licenseFile.getAbsolutePath() );

        try {
            logger.debug( "Loading the license key from " + licenseFile );
            licenseContent=install( licenseFile );
        }
        catch (LicenseException le) {
            throw le;
        }
        catch (Exception e) {
            throw new LicenseException( "Error validating or installing DataPowerBuddy license", e );
        }
        logger.debug( "Successfully validated the license key file " + licenseFile );

        return new License( licenseContent );
    }

    public final static String LICENSE_CONTACT_MESSAGE="Please contact us at http://myarch.com/dpbuddy-license or license@myarch.com to renew your license.";
    @Override
    protected synchronized void validate( final LicenseContent license ) throws LicenseContentException {
        
        logger.debug( "Validating the following license:\n"+createLicenseDetailsString(license) );

        if (license.getNotAfter().compareTo( new Date() ) < 0)
            throw new LicenseException(
                            "We are sorry but your license issued on %tc expired on %tc. "+ LICENSE_CONTACT_MESSAGE,
                            license.getIssued(), license.getNotAfter() );
        if (isTrial( license ) && !trialLicenseWaringIssued) {
            logger.info( String.format( "Note: you are using a trial DPBuddy license expiring on %tF",
                            license.getNotAfter() ) );
            trialLicenseWaringIssued=true;
        }
        // warn if license is expired in less than X days
        if (license.getNotAfter().getTime() - System.currentTimeMillis() < 7 * 24 * 60 * 60 * 1000)
            logger.warn( String
                            .format( "Please note that your DPBuddy license will expire on %tc. "+LICENSE_CONTACT_MESSAGE,
                                            license.getNotAfter() ) );

        try {
            super.validate( license );
        }
        catch (LicenseContentException lce) {
            throw new LicenseException( "There was an issue validating your dpbuddy license file. "
                + "The license was issued on %tc", lce, license.getIssued() );
        }
    }

    private String createLicenseDetailsString(final LicenseContent license) {
        StringBuilder buf=new StringBuilder();
        
        buf.append( "Expired on: "+license.getNotAfter() );
        
        String issuedTo=null;
        X500Principal holder=license.getHolder();
        if (license.getHolder() != null) {
            issuedTo=holder.getName();
        }

        if (issuedTo!=null)
            buf.append( " Issued to: "+issuedTo );
            
        String licenseInfo=license.getInfo();
        if (licenseInfo!=null) {
            buf.append( " "+licenseInfo);
        }

        return buf.toString();
        
    }
    
    static boolean isTrial( License license ) {
        return isTrial( license.getLicenseContent() );
    }

    static boolean isTrial( LicenseContent licenseContent ) {
        return licenseContent.getInfo().equals( DPBuddyLicenseParam.TRIAL_LICENSE_INFO_STRING );
    }

    public static void clearLicense() {
        try {
            DPBuddyLicenseManager.DPBuddyLicenseParam.LICENSE_PREFERENCES_NODE.clear();
        }
        catch (BackingStoreException e) {
            throw new DPBuddyException( e );
        }
    }
    
    static class DPBuddyLicenseParam implements de.schlichtherle.license.ftp.LicenseParam {
        /**
         * The subject for the license manager and also the alias of the private
         * key entry in the keystore.
         */
        static final String SUBJECT="dpbuddy";
        /** The password for the keystore. */
        static final String KEYSTORE_STORE_PWD=new ObfuscatedString( new long[] { 0xCE82084A3768AF38L,
            0xBA5EA2E23D9EA660L, 0xD1F7003EA83B7F15L } ).toString();

        /* The password for the private key entry in the keystore. */
        static final String KEYSTORE_KEY_PWD=KEYSTORE_STORE_PWD;

        /** The password to encrypt the generated license key file. */
        static final String CIPHER_KEY_PWD=new ObfuscatedString( new long[] { 0xCFA5212567F518ECL,
            0x9F2C9695C8D19A50L, 0xE17E7AED57772A2AL } ).toString();

        static final String TRIAL_KEY_PWD=new ObfuscatedString( new long[] { 0xC617FC92C6FC818FL,
            0xD1ABFBD17BAA9882L, 0x3318E4720BF0790EL } ).toString();

        /** The resource name of your public keystore file. */
        static final String KEYSTORE_RESOURCE="/dpbuddy.ks";

        final static private String PREFERENCES_PATH=new ObfuscatedString( new long[] {
            0xC2341C513BB7101BL, 0x2564A931B3E2CDA5L, 0x9E90C7D161856DFEL, 0x1AA51334FBDA5E23L } )
                        .toString();
//        final static Preferences LICENSE_PREFERENCES_NODE=Preferences.userRoot().node( PREFERENCES_PATH );
//        final static Preferences LICENSE_PREFERENCES_NODE=new FilePreferences( null, "" );//.node( "dpbuddy" );
        final static Preferences LICENSE_PREFERENCES_NODE=new FilePreferences( null, "" ).node( PREFERENCES_PATH );

        final static int TRIAL_LICENSE_EXPIRATION_DAYS=45+2;
        final static String TRIAL_LICENSE_INFO_STRING="Trial";

        private int trialExpirationDays=TRIAL_LICENSE_EXPIRATION_DAYS;

        public String getSubject() {
            return SUBJECT;
        }

        public Preferences getPreferences() {
            return LICENSE_PREFERENCES_NODE;
        }

        public KeyStoreParam getKeyStoreParam() {
            return new DefaultKeyStoreParam( DPBuddyLicenseManager.class, KEYSTORE_RESOURCE, SUBJECT,
                            KEYSTORE_STORE_PWD, null );
        }

        public CipherParam getCipherParam() {
            return new DefaultCipherParam( CIPHER_KEY_PWD );
        }

        public KeyStoreParam getFTPKeyStoreParam() {
            return new DefaultKeyStoreParam( DPBuddyLicenseManager.class, KEYSTORE_RESOURCE, SUBJECT
                + "-trial", KEYSTORE_STORE_PWD, TRIAL_KEY_PWD );
        }

        public int getFTPDays() {
            return trialExpirationDays;
        }

        public void setTrialExpirationDays( int days ) {
            trialExpirationDays=days;
        }

        public boolean isFTPEligible() {
            return true;
        }

        public LicenseContent createFTPLicenseContent() {
            LicenseContent license=new LicenseContent();
            license.setConsumerType( "User" );
            license.setIssued( new Date() );
            license.setInfo( TRIAL_LICENSE_INFO_STRING );

            return license;
        }

        public void removeFTPEligibility() {

        }

        public void ftpGranted( LicenseContent content ) {

        }

    }
}
