package com.myarch.dpbuddy.license;

import java.util.Calendar;
import java.util.Date;

import javax.security.auth.x500.X500Principal;

import de.schlichtherle.license.IllegalPasswordException;
import de.schlichtherle.license.KeyStoreParam;
import de.schlichtherle.license.LicenseContent;
import de.schlichtherle.license.LicenseNotary;
import de.schlichtherle.license.ftp.LicenseParam;
import de.schlichtherle.util.ObfuscatedString;

public class DPBuddyTrialLicenseManager extends de.schlichtherle.license.LicenseManager {

    private static final String FTP_KEYSTORE_PARAM=new ObfuscatedString( new long[] {
        0x5AC618555EC967B9L, 0xABE6B36EE9502E44L, 0x58814EE1E5E3CF8FL } ).toString(); /*
                                                                                       * =
                                                                                       * >
                                                                                       * "ftpKeyStoreParam"
                                                                                       */

    private static final String EQUAL_KEY_STORE_PARAMS=new ObfuscatedString( new long[] {
        0x4C72769A851F67C3L, 0x40A67DFB8206B61BL, 0x412E3FE142BA378EL, 0x68B2E25FA4BE7607L,
        0x7A6AB29D2E2E5FDDL, 0x0432DC122188C63FL, 0x3AAAFA7C2F900E82L, 0xE7676924E761DBA6L } )
                    .toString(); /*
                                  * =>
                                  * "Equal key store parameters for regular and FTP licenses!"
                                  */

    // private static final String FTP_DAYS = new ObfuscatedString(new long[] {
    // 0x73F69B7ABA25F620L, 0x40BA225C709D724EL
    // }).toString(); /* => "ftpDays" */

    private LicenseNotary ftpNotary;

    private static final String CN_FTP_USER=CN
        + Resources.getString( new ObfuscatedString( new long[] { 0xA54F04397F20F4A8L,
            0xC56F605BF6AF0408L } ).toString() ); /* => "user" */

    /**
     * Creates a new License Manager. <LicenseParam> <b>Warning:</b> The manager
     * created by this constructor is <em>not</em> valid and cannot be used
     * unless {@link #setLicenseParam(LicenseParam)} is called!
     */
    protected DPBuddyTrialLicenseManager() {
    }

    /**
     * Creates a new License Manager.
     * 
     * @param param
     *            The license configuration parameters - may <em>not</em> be
     *            {@code null}.
     * @throws NullPointerException
     *             If the given parameter object does not obey the contract of
     *             its interface due to a {@code null} pointer.
     * @throws IllegalPasswordException
     *             If any password in the parameter object does not comply to
     *             the current policy.
     */
    public DPBuddyTrialLicenseManager(LicenseParam param) {
        setLicenseParam0( param );
    }

    /**
     * Sets the license configuration parameters. Calling this method resets the
     * manager as if it had been newly created. The implementation of this class
     * asserts that the given parameter object is an instance of
     * {@link LicenseParam} and applies some plausibility checks to it to ensure
     * that it adheres to the contract of the parameter interfaces.
     * 
     * @param param
     *            The license configuration parameters - may <em>not</em> be
     *            {@code null}.
     * @throws ClassCastException
     *             if param is not an instance of {@link LicenseParam}.
     * @throws NullPointerException
     *             If the given parameter object does not obey the contract of
     *             its interface due to a {@code null} pointer.
     * @throws IllegalPasswordException
     *             If any password in the parameter object does not comply to
     *             the current policy.
     */
    public synchronized void setLicenseParam( de.schlichtherle.license.LicenseParam param ) {
        setLicenseParam0( param );
    }

    private void setLicenseParam0( de.schlichtherle.license.LicenseParam param ) {
        // Check parameters to implement fail-fast behaviour.
        final LicenseParam p=(LicenseParam) param;
        final KeyStoreParam ftpKeyStoreParam=p.getFTPKeyStoreParam();
        if (ftpKeyStoreParam == null)
            throw new NullPointerException( FTP_KEYSTORE_PARAM );
        if (ftpKeyStoreParam.equals( p.getKeyStoreParam() ))
            throw new IllegalArgumentException( EQUAL_KEY_STORE_PARAMS );

        super.setLicenseParam( p );
    }

    LicenseContent issueTrialLicense() {
        // this is required in order to work under IBM JDK, see
        // http://java.net/jira/browse/TRUELICENSE-2
        // System.setProperty("line.separator","\n");

        // Init.
        final LicenseParam param=(LicenseParam) getLicenseParam();
        final LicenseNotary ftpNotary=getFTPLicenseNotary();

        // Create and install an FTP license key.
        LicenseContent content=param.createFTPLicenseContent();
        content.setNotAfter( ftpNotAfter( param.getFTPDays() ) ); // enforce
                                                                  // expire date

        // this is required in order to work under IBM JDK, see
        // http://java.net/jira/browse/TRUELICENSE-2
        String currentSystemLineSeparator=System.getProperty( "line.separator" );
        try {
            System.setProperty( "line.separator", "\n" );
            content=install( create( content, ftpNotary ), ftpNotary );
        }
        catch (Exception e) {
            throw new LicenseException( "Error validating or installing DataPowerBuddy license", e );
        }
        finally {
            System.setProperty( "line.separator", currentSystemLineSeparator );
        }

        param.removeFTPEligibility();
        param.ftpGranted( content );

        return content; // content of FTP license
    }

    public LicenseContent validateTrialIfExists() {
        LicenseContent license=null;
        byte[] key=getLicenseKey();
        if (key != null) {
            try {
                license=super.verify( getFTPLicenseNotary() );
            }
            // for validation errors, such as expiration simply rethrow
            catch (LicenseException dpBuddyLicenseException) {
                throw dpBuddyLicenseException;
            }
            catch (Exception ex) {
                // we should never get here under normal circumstances
                throw ( new LicenseException(
                                "We found an installed license, but there was a problem decrypting it. Please make sure that 'dpbuddy.license.file' points to a valid license key file.",
                                ex ) );
            }

        }

        return license;
    }

    protected synchronized void initialize( final LicenseContent content ) {
        if (content.getHolder() == null)
            content.setHolder( new X500Principal( CN_FTP_USER ) );
        super.initialize( content );
    }

    //
    // Various stuff.
    //

    /**
     * Returns a license notary configured to use the free trial period keystore
     * parameters contained in the current license parameters - {@code null} is
     * never returned.
     */
    protected synchronized LicenseNotary getFTPLicenseNotary() {
        if (ftpNotary == null)
            ftpNotary=new LicenseNotary( ( (LicenseParam) getLicenseParam() ).getFTPKeyStoreParam() );

        return ftpNotary;
    }

    /**
     * Returns the date after which a free trial period license certificate
     * should expire, based on {@code ftpDays}.
     */
    protected Date ftpNotAfter( int ftpDays ) {
        Calendar cal=Calendar.getInstance();
        cal.setTime( midnight() );
        cal.add( Calendar.DATE, ftpDays );

        return cal.getTime();
    }
}
