package com.myarch.dpbuddy.log;

import org.apache.commons.lang3.StringUtils;

import com.google.gson.JsonObject;
import com.myarch.dpbuddy.event.EventUtils;

public class DPTextLogMessageParser {
    /* Sample messages:
    syslog:
    {"message":"[0x80000001][xsltmsg][info] xmlfirewall(web-mgmt): trans(161874)[response][192.168.1.31] gtid(161874): 200 OK\n","@version":"1","@timestamp":"2016-10-23T17:49:26.000Z","host":"192.168.1.32","priority":14,"timestamp":"Oct 23 13:49:26","logsource":"dp75","severity":6,"facility":1,"facility_label":"user-level","severity_label":"Informational"}
    
    {"message":"[dpbuddy-samples][][all][warn] domainSamples\n","@version":"1","@timestamp":"2016-10-23T18:00:40.000Z","host":"192.168.1.32","priority":12,"timestamp":"Oct 23 14:00:40","logsource":"dp75","severity":4,"facility":1,"facility_label":"user-level","severity_label":"Warning"}    
    
    Grammar:
    [domain][ID][category][level] obj.type(obj.name): trans(3333)[direction][client ip] gtid(id): message
    
            
    File:
    20170207T185459.916Z [0x81000019][auth][info] : [172.16.253.1]: User 'dpbuddy' logged out from 'default'.
    20170207T193752.590Z [dev][0x81000234][cli][info] : tid(140704): Logging event successfully scheduled.
    file vs syslog diffs:
    in the file, if class/object is missing, there is always a " : ", in syslog, it's a space 
    "tid" vs "trans", timestamp, colon
    
    
    */

    public JsonObject parse( String dpLogMessage ) {
        dpLogMessage=StringUtils.trimToEmpty( dpLogMessage );
        JsonObject json=new JsonObject();
        JsonObject dpJson=new JsonObject();
        json.add( "dp", dpJson );

        LogMessageToken token=null;
        String curMessageSegment=dpLogMessage;
        int fieldIndex=0;
        while (true) {
            token=getNextToken( curMessageSegment );

            if (token == null) {
                break;
            }

            if (token.type() == TokenType.FIELD) {
                boolean wasHandled=handleField( fieldIndex, json, token );
                if (!wasHandled) {
                    handleMessage( json, curMessageSegment );
                    break;
                }
                ++fieldIndex;
            }

            else if (token.type() == TokenType.FUNCTION_LIKE) {
                FunctionLikeInfo functionLike=token.functionLike();

                if (isTransId( functionLike )) {
                    handleTranId( json, LogEntry.TRAN_ID_FIELD, functionLike );
                    // stop, don't try to parse the rest
                    handleMessage( json, moveToStartOfNextToken( curMessageSegment, token.endPos() ) );
                    break;
                }
                else {
                    handleClassObject( json, functionLike );
                }
            }

            else if (token.type() == TokenType.TEXT) {
                handleMessage( json, token.text() );
                // TODO: stop for now, parse the type(name) in the future
                break;
            }

            curMessageSegment=moveToStartOfNextToken( curMessageSegment, token.endPos() );
        }

        return json;

    }

    private String moveToStartOfNextToken( String curMessageSegment, int curTokenEndPos ) {
        curMessageSegment=curMessageSegment.substring( curTokenEndPos );
        curMessageSegment=StringUtils.stripStart( curMessageSegment, " :" );

        return curMessageSegment;
    }

    // true -- field was handled successfully
    private boolean handleField( int fieldIndex, JsonObject eventJson, LogMessageToken fieldToken ) {
        String fieldVal=fieldToken.text();
        String label=null;

        JsonObject json=eventJson;

        // first field is either message id or domain
        if (fieldIndex == 0) {
            if (fieldVal.startsWith( "0x" ) || StringUtils.isBlank( fieldVal )) {
                label="message_id";
            }
            else {
                label=EventUtils.DOMAIN_FIELD;
                json=EventUtils.getDpJson( eventJson );
            }
        }
        else {

            if (!EventUtils.containsDomain( eventJson )) {
                // keep consistent numbers as if domain was always part of the
                // log entry
                fieldIndex++;
            }

            if (fieldIndex == 1) {
                label="message_id";
            }
            else if (fieldIndex == 2) {
                label="category";
            }
            else if (fieldIndex == 3) {
                label="level";
            }
        }
        if (label == null) {
            return false;
        }

        if (StringUtils.isNotBlank( fieldVal )) {
            json.addProperty( label, fieldVal );
        }

        return true;
    }

    private void checkNumericFun( JsonObject json, String jsonPropName, String funName, String message ) {
        FunctionLikeInfo funInfo=checkForFunctionLike( funName, message );
        if (funInfo != null && StringUtils.isNumeric( funInfo.arg() )) {
            handleTranId( json, jsonPropName, funInfo );
        }

    }

    private void handleMessage( JsonObject eventJson, String message ) {
        eventJson.addProperty( "message", StringUtils.stripStart( message, null ) );
        checkNumericFun( eventJson, LogEntry.GLOBAL_TRAN_ID_FIELD, "gtid", message );
        checkNumericFun( eventJson, LogEntry.TRAN_ID_FIELD, TRAN_ID_FUNCTION_NAME, message );
        
        AdditionalLogMessageParsing.parse( eventJson, message);
        
    }

    private void handleTranId( JsonObject eventJson, String propName, FunctionLikeInfo functionLike ) {
        long tranId=0;
        try {
            tranId=Integer.parseInt( functionLike.arg() );
        }
        catch (NumberFormatException nfe) {

        }
        eventJson.addProperty( propName, tranId );

    }

    public final static String TRAN_ID_FUNCTION_NAME="trans";

    private boolean isTransId( FunctionLikeInfo functionLike ) {
        // trans in syslog and tid in files
        return ( functionLike.name().equals( TRAN_ID_FUNCTION_NAME ) || functionLike.name().equals( "tid" ) )
            && StringUtils.isNumeric( functionLike.arg() );
    }

    // for now we only handle class/object names
    private void handleClassObject( JsonObject eventJson, FunctionLikeInfo functionLike ) {

        JsonObject dpJson=EventUtils.getDpJson( eventJson );
        dpJson.addProperty( "class", functionLike.name() );
        dpJson.addProperty( "object", functionLike.arg() );

    }

    private final static String FIELD_START="[";
    private final static String FIELD_END="]";

    private LogMessageToken getNextToken( String curMessageSegment ) {
        LogMessageToken token=null;

        if (curMessageSegment.startsWith( FIELD_START )) {
            token=parseFieldToken( curMessageSegment );
        }
        else if (curMessageSegment.length() > 0) {
            FunctionLikeInfo functionLike=parseFunctionLike( curMessageSegment );
            if (functionLike != null
                && functionLike.startPos() == StringUtils.indexOfAnyBut( curMessageSegment, " " )) {
                token=new LogMessageToken( functionLike );
            }
            // TODO: colon handling here (for disk messages)
            else {
                token=new LogMessageToken( StringUtils.stripStart( curMessageSegment, null ), TokenType.TEXT );
            }
        }

        return token;
    }

    private LogMessageToken parseFieldToken( String curMessageSegment ) {
        LogMessageToken token=new LogMessageToken();
        int endPos=curMessageSegment.indexOf( FIELD_END );
        if (endPos < 0) {
            token.setType( TokenType.TEXT );
            token.setText( curMessageSegment.trim() );
            endPos=curMessageSegment.length();
        }
        else {
            token.setType( TokenType.FIELD );
            String tokenText=curMessageSegment.substring( FIELD_START.length(), endPos );
            token.setText( tokenText );
            endPos=endPos + FIELD_END.length();
        }

        token.setEndPos( endPos );

        return token;
    }

    private FunctionLikeInfo checkForFunctionLike( String functionName, String message ) {
        FunctionLikeInfo functionLike=null;
        int startPos=message.indexOf( functionName + "(" );
        if (startPos >= 0) {
            String arg=StringUtils.substringBetween( message, functionName + "(", ")" );
            if (arg != null) {
                functionLike=new FunctionLikeInfo( functionName, arg, -1, -1 );
            }
        }

        return functionLike;
    }

    private FunctionLikeInfo parseFunctionLike( String messageSegment ) {
        String endStr=")";

        FunctionLikeInfo functionLike=null;
        int endPos=messageSegment.indexOf( endStr );
        if (endPos > 0) {
            StringAndPos argInfo=getWordsBefore( messageSegment, endPos );
            if (argInfo == null) {
                return null;
            }

            String arg=argInfo.s;
            StringAndPos nameInfo=getWordsBefore( messageSegment, argInfo.pos - 1 );

            if (nameInfo != null && StringUtils.isNotBlank( nameInfo.s )) {
                functionLike=new FunctionLikeInfo( nameInfo.s, arg, nameInfo.pos, endPos + endStr.length() );
            }
        }

        return functionLike;
    }

    /**
     * Returns substring between the end pos and the beginning of the line or ( or :
     * @param s
     * @param wordEndPos
     * @return
     */
    private StringAndPos getWordsBefore( String s, int wordEndPos ) {
        char[] chars=s.toCharArray();
        int wordStartPos=-1;
        for (int i=wordEndPos - 1; i >= 0; --i) {
            // if (i == 0 || Character.isWhitespace( chars[i] ) || chars[i] ==
            // '(') {
            if (i == 0 || chars[i] == '(' || chars[i] == ':') {
                wordStartPos=( i == 0 ? i : i + 1 );
                break;
            }
        }

        if (wordStartPos < 0 || wordStartPos == wordEndPos)
            return null;

        String word=s.substring( wordStartPos, wordEndPos );
        StringAndPos sAndPos=new StringAndPos();
        sAndPos.s=word;
        sAndPos.pos=wordStartPos;

        return sAndPos;
    }

    private static class StringAndPos {
        int pos;
        String s;

        @Override
        public String toString() {
            return "StringAndPos [pos=" + pos + ", s=" + s + "]";
        }

    }

    private static class FunctionLikeInfo {
        private String name;
        private String arg;
        private int startPos;
        private int endPos;

        FunctionLikeInfo(String name, String arg, int startPos, int endPos) {
            this.name=name;
            this.arg=arg;
            this.startPos=startPos;
            this.endPos=endPos;
        }

        public String name() {
            return name;
        }

        public String arg() {
            return arg;
        }

        public int startPos() {
            return startPos;
        }

        public int endPos() {
            return endPos;
        }

        @Override
        public String toString() {
            return "FunctionLikeInfo [name=" + name + ", arg=" + arg + ", startPos=" + startPos + ", endPos=" + endPos
                + "]";
        }

    }

    private static class LogMessageToken {
        private String tokenText;
        private TokenType type;
        private int endPos;
        private FunctionLikeInfo functionLike;

        public LogMessageToken() {

        }

        public LogMessageToken(String tokenText, TokenType type) {
            this.tokenText=tokenText;
            this.type=type;
        }

        public LogMessageToken(FunctionLikeInfo functionLike) {
            this.functionLike=functionLike;
            this.type=TokenType.FUNCTION_LIKE;
            endPos=functionLike.endPos();
        }

        public String text() {
            return tokenText;
        }

        void setText( String tokenText ) {
            this.tokenText=tokenText;
        }

        void setType( TokenType type ) {
            this.type=type;
        }

        public TokenType type() {
            return type;
        }

        void setEndPos( int pos ) {
            this.endPos=pos;
        }

        public int endPos() {
            return endPos;
        }

        public FunctionLikeInfo functionLike() {
            return functionLike;
        }

        @Override
        public String toString() {
            return "LogMessageToken [type=" + type + ", tokenText=" + tokenText + ", endPos=" + endPos + "]";
        }

    }

    private static enum TokenType {
        TEXT, FIELD, FUNCTION_LIKE;

    }

}
