package com.myarch.dpbuddy.log;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.BaseRequest;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.Device;
import com.myarch.util.StringPatternUtils;

/**
 * Retrieves logs from the device and provide the ability to tail them,
 * including tailing in a continuous fashion a la tail -f. Represents "get-log"
 * SOMA operation.
 */
public class GetLogCommand implements DPCommand<List<LogEntry>> {

    private static Log logger=LogFactory.getLog( GetLogCommand.class );

    private List<LogFilter> logFilters=new ArrayList<LogFilter>();

    private int numberOfLines=-1;
    private boolean follow=false;
    // query the device every X miliseconds
    private int followInterval=3000;

    private Set<String> domains=new LinkedHashSet<String>();

    private List<Pattern> domainPatterns=new ArrayList<Pattern>();

    private boolean failOnError=false;
    private Set<Pattern> failPatterns=new LinkedHashSet<Pattern>();
    private String logFormat=LogEntry.DEFAULT_LOG_FORMAT;

    private BaseRequest request=new BaseRequest();

    public final static String GET_LOG_ELEMENT_NAME="get-log";

    private File logFile;
    private PrintStream printStream=System.out;

    // print the log or just return log entries?
    private boolean isPrintLog=true;

    public GetLogCommand() {
        request.addRequestChild( GET_LOG_ELEMENT_NAME );
    }

    public void setLogName( String dpLogName ) {
        Element getLogElt=request.getRequestChild( GET_LOG_ELEMENT_NAME );
        getLogElt.setAttribute( "name", dpLogName );
    }

    public void setNumberOfLines( int numberOfLines ) {
        this.numberOfLines=numberOfLines;
    }

    public void setPrintLog( boolean isPrintLog ) {
        this.isPrintLog=isPrintLog;
    }

    public void setFollow( boolean follow ) {
        this.follow=follow;
    }

    public void setFollowInterval( int followInterval ) {
        this.followInterval=followInterval;
    }

    public void addDomainPatterns( String namePatterns ) {
        domainPatterns.addAll( StringPatternUtils.stringToPattnernList( namePatterns ) );
    }

    public void addDomainPatterns( List<Pattern> namePatterns ) {
        this.domainPatterns.addAll( namePatterns );
    }

    public void addDomainPattern( Pattern namePattern ) {
        domainPatterns.add( namePattern );
    }

    public void addLogFilter( LogFilter filter ) {
        logFilters.add( filter );
    }

    public boolean isFailOnError() {
        return failOnError;
    }

    public void setFailOnError( boolean failOnError ) {
        this.failOnError=failOnError;
    }

    public Set<Pattern> getFailPatterns() {
        return failPatterns;
    }

    public void addFailPatterns( List<Pattern> failPatterns ) {
        this.failPatterns.addAll( failPatterns );
    }

    public void addFailPatterns( String failPatterns ) {
        this.failPatterns.addAll( StringPatternUtils.stringToPattnernList( failPatterns ) );
    }

    public String getLogFormat() {
        return logFormat;
    }

    public void setLogFormat( String logFormat ) {
        this.logFormat=logFormat;
    }

    public void setLogFile( File logFile ) {
        this.logFile=logFile;
    }

    public void setPrintStream( PrintStream printStream ) {
        this.printStream=printStream;
    }

    /**
     * Executes getLog command, logs the received log and returns the list of
     * entries to the client.
     */
    public List<LogEntry> execute( Device device ) {

        if (follow && logFile != null) {
            throw new DPBuddyException( "You can't save log to a file when 'follow' is set to 'true'." );
        }

        if (domainPatterns.size() > 0) {
            domains=device.fetchRequiredDomains( domainPatterns );
            logger.info( "Retrieving logs from the following domains: " + domains );
        }
        else
            domains.add( device.getCurrentDomain() );

        List<LogEntry> entries=null;
        FileOutputStream fos=null;

        if (logFile != null) {
            try {
                fos=new FileOutputStream( logFile );
            }
            catch (FileNotFoundException e) {
                throw new DPBuddyException(
                                "Error creating file '%s' for saving the log file downloaded from the device",
                                logFile.getAbsolutePath() );
            }
            printStream=new PrintStream( fos );
        }
        try {
            try {

                entries=retrieveLog( device, false );
                if (isPrintLog) {
                    logTail( entries, numberOfLines );
                    if (follow) {
                        autoLog( device, entries );
                    }
                }
            }
            finally {
                if (printStream != null && printStream != System.out && printStream != System.err)
                    printStream.close();
                if (fos != null)
                    fos.close();
            }
        }
        catch (IOException e) {
            throw new DPBuddyException( "Error saving file to '%s'. Caused by:\n%s", e,
                            logFile.getAbsolutePath(), e.getMessage() );
        }
        if (logFile != null) {
            logger.info( String.format( "Saved DataPower log file has to '%s'",
                            logFile.getAbsolutePath() ) );
        }

        return entries;
    }

    private void autoLog( Device device, List<LogEntry> curEntries ) {
        List<LogEntry> oldEntries=curEntries;
        while (true) {
            try {
                Thread.sleep( followInterval );
            }
            catch (InterruptedException e) {
                logger.debug(e);
            }
            List<LogEntry> newEntries=retrieveLog( device, true );

            LogEntry lastOldEntry=null;
            if (oldEntries.size()>0) {
                lastOldEntry=oldEntries.get( oldEntries.size() - 1 );
            }
            int i=0;
            for (LogEntry newEntry : newEntries) {
                if (lastOldEntry==null || newEntry.getUTC() > lastOldEntry.getUTC()) {
                    logStartingWith( newEntries, i );
                    oldEntries=newEntries;
                    break;
                }
                ++i;
            }
        }
    }

    private List<LogEntry> retrieveLog( Device device, Boolean quiet ) {

        ArrayList<LogEntry> logEntries=new ArrayList<LogEntry>();
        for (String domain : domains) {
            Document responseDoc=device.executeRequest( domain, request, quiet );
            LogResponse response=new LogResponse( responseDoc );
            List<LogEntry> retrievedLogEntries=response.getLogEntries( logFormat );
            if (logFilters.size() > 0)
                retrievedLogEntries=filterLogEntries( retrievedLogEntries, logFilters );

            logEntries.addAll( retrievedLogEntries );
        }

        Collections.sort( logEntries, new LogEntry.TimestampComparator() );

        return logEntries;
    }

    public static List<LogEntry> filterLogEntries( List<LogEntry> logEntries, LogFilter filter ) {
        List<LogFilter> filters=new ArrayList<LogFilter>();
        filters.add( filter );
        return filterLogEntries( logEntries, filters );
    }

    public static List<LogEntry> filterLogEntries( List<LogEntry> logEntries, List<LogFilter> filters ) {
        List<LogEntry> filteredLogEntries=new ArrayList<LogEntry>();

        for (LogEntry logEntry : logEntries) {
            for (LogFilter filter : filters) {
                if (filter.matches( logEntry )) {
                    filteredLogEntries.add( logEntry );
                    break;
                }
            }
        }
        return filteredLogEntries;
    }

    /**
     * Log (e.g., write to sysout) the tail of the device log
     * 
     * @param entries
     * @param numberOfLines
     */
    private void logTail( List<LogEntry> entries, int numberOfLines ) {
        int i=0;
        for (LogEntry entry : entries) {
            if (numberOfLines < 0 || ( entries.size() - i ) <= numberOfLines) {
                logEntry( entry );
            }
            ++i;
        }
    }

    private void logStartingWith( List<LogEntry> entries, int startingI ) {
        for (int i=startingI; i < entries.size(); ++i) {
            LogEntry entry=entries.get( i );
            logEntry( entry );
        }
    }

    private void logEntry( LogEntry entry ) {
        String logMessage=entry.getLogString();
        printStream.println( logMessage );

        if (failOnError && entry.isError()) {
            throw new DPBuddyException( "Encountered an error in DataPower logs. Log entry:\n"
                + logMessage );
        }

        for (Pattern pattern : failPatterns) {
            Matcher m=pattern.matcher( logMessage );
            if (m.find( 0 )) {
                throw new DPBuddyException(
                                "Encountered error pattern '%s' in DataPower logs. Log entry:\n"
                                    + logMessage, pattern );
            }
        }

    }

}
