package com.myarch.dpbuddy.log;

import java.io.File;
import java.util.Date;
import java.util.Map;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.google.gson.JsonObject;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.event.EventUtils;
import com.myarch.eventmodel.Event;
import com.myarch.eventmodel.EventType;
import com.myarch.hocon.HoconEnvironmentUtils;
import com.myarch.hocon.HoconUtils;
import com.myarch.util.DateUtils;
import com.myarch.util.JsonUtils;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import com.typesafe.config.ConfigObject;
import com.typesafe.config.ConfigValue;
import com.typesafe.config.ConfigValueType;

public class JsonLogEventConverter {

    public final static String SRC_LOG_FORMAT_FIELD="src_log_format";
    public final static String SRC_LOG_FILE_FIELD="src_log_file";
    
    private final Log logger=LogFactory.getLog( getClass() );

    private static JsonLogEventConverter instance=new JsonLogEventConverter();

    private DPTextLogMessageParser parser=new DPTextLogMessageParser();
    private DPXmlLogMessageParser xmlParser=new DPXmlLogMessageParser();
    private Config dpbuddyConf=null;

    private JsonLogEventConverter() {
    }

    public static JsonLogEventConverter instance( String dpbuddyConfigPath ) {
        if (dpbuddyConfigPath != null) {
            File dpbuddyConfigFile=new File( dpbuddyConfigPath );
            if (!dpbuddyConfigFile.exists()) {
                instance.logger.warn( "Conf file " + dpbuddyConfigPath + " doesn't exist" );
            }
            else {
                instance.dpbuddyConf=loadHocon( dpbuddyConfigFile );
            }
        }
        return instance;
    }

    private static Config loadHocon( File confFile ) {
        Config conf=ConfigFactory.parseFileAnySyntax( confFile );
        conf=conf.resolve();

        return conf;
    }

    public JsonObject fromAnySource( Map<String, String> sourceEvent ) {
        return fromAnySource( sourceEvent, false );
    }

    public JsonObject fromAnySource( Map<String, String> sourceEvent, boolean inferHost ) {
        JsonObject resultingEvent=null;
        if (sourceEvent.containsKey( "logsource" )) {
            resultingEvent=fromSyslog( sourceEvent );
        }
        else if (sourceEvent.containsKey( "path" )) {
            resultingEvent=fromFileSyslog( sourceEvent, inferHost );
        }
        else if (sourceEvent.containsKey( "host" ) && isMessageXml( sourceEvent.get( "message" ) )) {
            resultingEvent=fromXml( sourceEvent );
        }
        else {
            throw new IllegalStateException(
                            "Unable to deterine the format of the input event. It must contain logsource or host or path fields" );
        }

        return resultingEvent;
    }

    private boolean isMessageXml( String message ) {
        return StringUtils.isNotBlank( message ) && message.startsWith( "<?xml" );
    }

    public JsonObject fromSyslog( Map<String, String> syslogEvent ) {

        Event event=createDpEvent();
        copyHostFieldsFromInput( event, syslogEvent );
        setFormatToSyslog(event);
        
        String message=syslogEvent.get( "message" );
        if (StringUtils.isNotBlank( message )) {
            JsonObject messageJson=parser.parse( message );
            JsonUtils.mergeJson( messageJson, event.json() );
        }

        String logsource=syslogEvent.get( "logsource" );
        handleSyslogLogSource( event, logsource );

        setDefaulDomainIfNotSet( event.json() );
        handleDpbuddyConfFile( event );

        return event.json();
    }

    public final static String SYSLOG_LOGSOURCE_SEPARATOR="--";

    private void handleSyslogLogSource( Event event, String logsource ) {
        if (StringUtils.isNotBlank( logsource )) {
            String[] logsourceTokens=StringUtils.splitByWholeSeparator( logsource, SYSLOG_LOGSOURCE_SEPARATOR );
            if (logsourceTokens.length > 0) {
                JsonUtils.addToArray( event.json(), "host_aliases", logsourceTokens[0] );
            }
            if (logsourceTokens.length > 1) {
                EventUtils.setDomain( event.json(), logsourceTokens[1] );
            }

        }
    }

    private void setDefaulDomainIfNotSet( JsonObject eventJson ) {
        if (!EventUtils.containsDomain( eventJson )) {
            EventUtils.setDomain( eventJson, Device.DEFAULT_DOMAIN_NAME );
        }
    }

    // add info from dpbuddy conf to the env entry: environment and analyticsInfo

    private void handleDpbuddyConfFile( Event event ) {
        if (dpbuddyConf != null) {
            String host=event.host();
            if (host != null) {
                HoconUtils.ConfObjWithKey envInfo=HoconEnvironmentUtils.findEnvironmentByHost( dpbuddyConf, host );
                if (envInfo != null) {
                    addEnvInfoToEvent(envInfo, event);
                }
            }
        }
    }

    
    private void addEnvInfoToEvent(HoconUtils.ConfObjWithKey envInfo, Event event ) {
        event.setEnvironment( envFromEnvInfo( envInfo ) );
        ConfigValue analyticsInfo=envInfo.confObj().get( "analyticsInfo" );
        if (analyticsInfo != null && analyticsInfo.valueType() == ConfigValueType.OBJECT) {
            HoconUtils.addStringValuesToJson( (ConfigObject) analyticsInfo, event.json() );
        }
    }
    
    private String envFromEnvInfo( HoconUtils.ConfObjWithKey envInfo ) {
        String envName=HoconUtils.get( envInfo.confObj(), "environment" );
        if (envName == null) {
            envName=envInfo.key();
        }
        return envName;
    }

    // syslog format in file
    public JsonObject fromFileSyslog( Map<String, String> syslogEvent, boolean inferHostFromPath ) {

        Event event=createDpEvent();
        setFormatToSyslog(event);

        String path=syslogEvent.get( "path" );
        if (path != null) {
            String normalizedPath=FilenameUtils.normalize( path );
            event.json().addProperty( "path", normalizedPath );
            event.json().addProperty( SRC_LOG_FILE_FIELD, normalizedPath);
            if (inferHostFromPath) {
                populateHostDomainFromPath( event, normalizedPath );
            }
        }
        
        String message=syslogEvent.get( "message" );
        message=extractAndSetDate( message, event );
        if (StringUtils.isNotBlank( message )) {
            JsonObject messageJson=parser.parse( message );
            JsonUtils.mergeJson( messageJson, event.json() );
        }

        setDefaulDomainIfNotSet( event.json() );
        handleDpbuddyConfFile( event );

        return event.json();
    }

    private void setFormatToSyslog(Event event) {
        event.json().addProperty( SRC_LOG_FORMAT_FIELD, "syslog");
    }
    
    private void populateHostDomainFromPath( Event event, String path ) {

        String host=null;
        String domain=null;

        File logFile=new File( path );
        File parent=logFile.getParentFile();
        if (parent != null) {
            if (StringUtils.startsWithIgnoreCase( logFile.getName(), Device.DEFAULT_DOMAIN_NAME )) {
                host=parent.getName();
            }
            else {
                domain=parent.getName();
                File parentParent=parent.getParentFile();
                if (parentParent != null) {
                    host=parentParent.getName();
                }

            }
        }

        if (host != null) {
            event.setHost( host );
        }
        if (domain != null && !EventUtils.containsDomain( event.json() )) {
            EventUtils.setDomain( event.json(), domain );
        }

    }

    private String extractAndSetDate( String message, Event event ) {
        String restOfMessage=message;
        String[] dateAndTheRestOfMessage=StringUtils.splitByWholeSeparator( message, " ", 2 );
        if (dateAndTheRestOfMessage.length == 2) {
            String dateStr=dateAndTheRestOfMessage[0];
            restOfMessage=dateAndTheRestOfMessage[1];
            Date timestamp=DateUtils.parseZuluDateTime( dateStr );
            if (timestamp != null) {
                event.setTimestamp( timestamp );
            }
            else {
                String file=event.get( "path" );
                String msg="Unable to parse date field from value: '" + dateStr + "'";
                if (file != null) {
                    msg+=" File: " + file;
                }
                msg+=", full text:\n"+message;
                logger.warn( msg );
            }

        }

        return restOfMessage;
    }

    public JsonObject fromXml( Map<String, String> inputEvent ) {
        Event event=createDpEvent();
        copyHostFieldsFromInput( event, inputEvent );
        event.json().addProperty( SRC_LOG_FORMAT_FIELD, "xml");
        
        String message=inputEvent.get( "message" );
        if (StringUtils.isNotBlank( message )) {
            JsonObject messageJson=xmlParser.parse( message );
            JsonUtils.mergeJson( messageJson, event.json() );
        }

        handleDpbuddyConfFile( event );

        return event.json();
    }

    private Event createDpEvent() {
        Event event=new Event( "DataPower", EventType.LOG_ENTRY );
        event.setType( "dp.log" );
        JsonObject dpJson=new JsonObject();
        event.json().add( "dp", dpJson );

        return event;
    }

    private void copyHostFieldsFromInput( Event event, Map<String, String> lsEvent ) {
        setHost( event, lsEvent );
    }

    private void setHost( Event event, Map<String, String> lsEvent ) {
        String host=lsEvent.get( "host" );
        if (StringUtils.isNotBlank( host )) {
            event.setHost( host );
        }
    }
}
