package com.myarch.dpbuddy.log;

import java.text.MessageFormat;
import java.util.Comparator;
import java.util.Date;

import org.apache.commons.lang3.StringUtils;
import org.jdom2.Element;

import com.google.gson.JsonObject;
import com.myarch.dpbuddy.Device;
import com.myarch.util.JsonUtils;

public class LogEntry {

    public final static String TRAN_ID_FIELD="tran_id";
    public final static String GLOBAL_TRAN_ID_FIELD="global_tran_id";
    
    
    private long utc;
    private long tid;

    private Element logEntryElt;

    public LogEntry(Element logEntryElt, String format) {
        this.logEntryElt=logEntryElt;
        if (format != null)
            setFormat( format );
        populate();
    }

    public LogEntry(Element logEntryElt) {
        this( logEntryElt, null );
    }

    private void populate() {
        String timeS=logEntryElt.getChild( "time" ).getAttributeValue( "utc" );
        utc=Long.parseLong( timeS );

        String tidS=logEntryElt.getChildText( "transaction" );
        tid=Long.parseLong( tidS );

    }

    public long getUTC() {
        return utc;
    }

    public String getDomain() {
        return logEntryElt.getAttributeValue( "domain" );
    }

    public String getCategory() {
        return logEntryElt.getChildText( "type" );
    }

    public String getMessage() {
        return logEntryElt.getChildText( "message" );
    }

    public String getObject() {
        return logEntryElt.getChildText( "object" );
    }

    public String getDPClass() {
        return logEntryElt.getChildText( "class" );
    }

    public String getFormattedObject() {
        String object=getObject();
        String clazz=getDPClass();

        StringBuilder buf=new StringBuilder( "" );
        if (StringUtils.isNotBlank( clazz ))
            buf.append( clazz );

        if (StringUtils.isNotBlank( object )) {
            if (buf.length() > 0)
                buf.append( ' ' );

            buf.append( "(" + object + ")" );
        }
        if (buf.length() > 0)
            buf.append( ": " );

        return buf.toString();
    }

    public Date getTimestamp() {
        return new Date( getUTC() );
    }

    public String getLevel() {
        return logEntryElt.getChildText( "level" );
    }

    public DPLogLevel getDPLogLevel() {
        return DPLogLevel.fromString( getLevel() );
    }

    public String getAbbreviatedLevel() {

        return getDPLogLevel().getAbbreviatedString();

    }

    public boolean isError() {
        return getDPLogLevel().isError();
    }

    public String getLogString() {
        return format();
    }

    public String toString() {
        return getLogString();
    }

    public JsonObject toJson() {
        JsonObject json=new JsonObject();
        JsonObject dpJson=new JsonObject();
        json.add( "dp", dpJson );

        json.addProperty( "message", StringUtils.strip( getMessage() ) );
        String messageId=logEntryElt.getChildText( "code" );
        JsonUtils.addPropertyIfNotBlank( json, "message_id", messageId );

        String timestamp=JsonUtils.dateToIsoStr( getTimestamp() );
        json.addProperty( "@timestamp", timestamp );
        json.addProperty( "timestamp", timestamp );

        json.addProperty( "level", getLevel() );
        // json.addProperty( "level_id", getLevel() );

        JsonUtils.addPropertyIfNotBlank( json, "category", getCategory() );

        addTransToJson( json );
        JsonUtils.addPropertyIfNotBlank( json, "tran_type", logEntryElt.getChildText( "transaction-type" ) );
        JsonUtils.addPropertyIfNotBlank( json, "client_host", logEntryElt.getChildText( "client" ) );

        addDomain( dpJson );
        addClassObject( dpJson );

        return json;
    }

    private void addTransToJson( JsonObject json ) {
        JsonUtils.addPropertyIfNotBlankGuessType( json, TRAN_ID_FIELD, logEntryElt.getChildText( "transaction" ) );
        JsonUtils.addPropertyIfNotBlankGuessType( json, GLOBAL_TRAN_ID_FIELD, logEntryElt.getChildText( "gtid" ) );
    }

    private void addDomain( JsonObject json ) {
        String domain=getDomain();
        if (StringUtils.isBlank( domain )) {
            domain=Device.DEFAULT_DOMAIN_NAME;
        }
        json.addProperty( "domain", domain );
    }

    private void addClassObject( JsonObject json ) {
        JsonUtils.addPropertyIfNotBlank( json, "class", getDPClass() );
        JsonUtils.addPropertyIfNotBlank( json, "object", getObject() );
    }

    // ISO8601 date format is used
    // alternative format with tid
    // public final static String DEFAULT_LOG_FORMAT="{1,date,yyyy-MM-dd
    // HH:mm:ss} | {2} | {4,number,#} | {0}{3}";
    public final static String DEFAULT_LOG_FORMAT="{1,date,yyyy-MM-dd HH:mm:ss} |{2}|{6}|{5} {3}{0}";
    // with domain:
    // public final static String DEFAULT_LOG_FORMAT="{5} | {1,date,yyyy-MM-dd
    // HH:mm:ss} |{2}| {0}{3}";

    private String format=DEFAULT_LOG_FORMAT;

    public void setFormat( String format ) {
        this.format=format;
    }

    /**
     * 0-message 1-timestamp 2-level (abbreviated) 3-formatted object
     * 4-transaction 5- domain 6 - category
     */
    private String format() {
        String domainDisplay="";
        if (StringUtils.isNotBlank( getDomain() )) {
            domainDisplay=getDomain() + "|";
        }

        return MessageFormat.format( format, getMessage(), getTimestamp(), getAbbreviatedLevel(), getFormattedObject(),
                        tid, domainDisplay, getCategory() );

    }

    public static class TimestampComparator implements Comparator<LogEntry> {

        public int compare( LogEntry entry1, LogEntry entry2 ) {
            if (entry1.getUTC() < entry2.getUTC())
                return -1;
            if (entry1.getUTC() == entry2.getUTC())
                return 0;
            return 1;
        }

    }

}
