package com.myarch.dpbuddy.logtarget;

import java.util.List;
import java.util.Set;

import org.apache.tools.ant.BuildException;
import org.jdom2.Element;
import org.kohsuke.args4j.Option;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.audit.AuditEvent;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.config.ConfigCommand;
import com.myarch.dpbuddy.config.ConfigCommand.ConfigOperation;
import com.myarch.dpbuddy.config.DPConfiguration;
import com.myarch.dpbuddy.status.ObjectStatus;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;

public class UpdateLogTargetLevelTask extends BaseDPBuddyTask {
    
    public UpdateLogTargetLevelTask() {
        super();
        // Enforce schema validation
        isValidateAttr=true;
    }

    private String namePattern;
    private String classPattern="LogTarget";

    @Option(name="-namePattern", aliases="-name", required=true, usage="Regular expression defining the names of the log targets to modify")
    public void setNamePattern( String namePattern ) {
        this.namePattern=namePattern;
    }

    // alias for Ant
    public void setName( String namePattern ) {
        this.namePattern=namePattern;
    }

    private String hoconStr;
    @Option(name="-levelInfo", required=true, usage="List of category:level in HOCON format, e.g., all:notice, mgmt:info")
    public void setLevelInfo( String hoconStr ) {
        this.hoconStr=hoconStr;

    }
    
    
    private boolean isDryRun=false;
    @Option(name="-dryRun", handler=BoolOptionHandler.class,
                    usage="Do not update anything, just print the matched objects.")
    public void setDryRun( boolean isDryRun ) {
        this.isDryRun=isDryRun;
    }

    
    @Override
    public void executeDPTask() throws BuildException {
        
        if (namePattern == null ) {
            throw new BuildException( "You must provide namePattern" );
        }
        
        if (isDryRun) {
            logger.info( "Executing in 'dry-run' mode, no changes will be made" );
        }
        
        Config modifications=ConfigFactory.parseString( hoconStr );
        
        DPObject pattern=new DPObject( classPattern, namePattern );
        logger.info( "Querying the device to determine what objects to modify" );
        Set<ObjectStatus> objectsToModify=getDevice().fetchObjects( pattern, true );

        logger.info( String.format( "The following objects will be updated:\n%s",
                        DPObject.objectListToString( objectsToModify ) ) );

        ConfigOperation op=ConfigOperation.MODIFY;
        ConfigCommand configCommand=new ConfigCommand( op );
        
        

        for (ObjectStatus dpObj : objectsToModify) {
            DPConfiguration dpConfiguration=buildDpConfiguration( dpObj.getClazz(), dpObj.getName(), modifications );
            configCommand.addConfiguration( dpConfiguration );
        }

        if (configCommand.isEmpty())
            throw new BuildException( "Invalid config" );

        if (logger.isDebugEnabled() && isDryRun) {
            logger.debug( "Updating log level using the following configuration:\n"+configCommand.toXmlString() );
        }        
        
        
        if (!isDryRun) {
            executeRequest( configCommand );
            AuditEvent auditEvent=createAuditEvent( "updateLevel" );
            auditEvent.setObjects( objectsToModify );
            if (auditEventHandler != null) {
                auditEventHandler.handleAuditEvent( auditEvent, getDevice() );
            }

            saveDomainConfigIfEnabled();
        }
    }

    // hocon: category: level, category: level
    private DPConfiguration buildDpConfiguration( String clazz, String objectName, Config catLevelConfig ) {

        Element classNameElt=new Element( clazz );
        classNameElt.setAttribute( "name", objectName );

        List<LogTargetEvent> logTargetEvents=LogTargetEvent.fromConfig(catLevelConfig);
        if (logTargetEvents.isEmpty()) {
            throw new DPBuddyException("You must provide at least one category:level pair");
        }
        
        for(LogTargetEvent logTargetEvent:logTargetEvents){
            classNameElt.addContent( logTargetEvent.toXmlElement() );
        }

        DPConfiguration dpConfiguration=new DPConfiguration();
        dpConfiguration.addElement( classNameElt );
        
        return dpConfiguration;
    }
    

}
