package com.myarch.dpbuddy.quiesce;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.status.StateChecker;
import com.myarch.util.StringPatternUtils;

public class QuiesceUnquiesceDomainCommand implements DPCommand<ResultResponse> {

    private final Log logger=LogFactory.getLog( getClass() );

    public final static QuiesceUnquiesceDomainCommand QUIESCE_ALL_DOMAINS=new QuiesceUnquiesceDomainCommand(QuiesceAction.QUIESCE, ".*", null);
    
    private QuiesceAction quiesceAction;
    // timeout in seconds
    private int timeout=-1;

    private List<Pattern> domainPatterns=new ArrayList<Pattern>();

    public QuiesceUnquiesceDomainCommand(QuiesceAction quiesceAction, String domainPatternsStr, Integer timeout) {
        this.quiesceAction=quiesceAction;
        this.timeout=quiesceAction.validateTimeout( timeout );
        if (StringUtils.isNotBlank( domainPatternsStr )) {
            domainPatterns.addAll( StringPatternUtils.stringToPattnernList( domainPatternsStr ) );
        }
    }

    public ResultResponse execute( Device device ) {
        ResultResponse response=null;
        if (domainPatterns.size() > 0) {
            logger.info( "Querying the device to determine domains to quiesce ..." );
            boolean isDomainMatchRequired=true;
            Set<String> domains=device.fetchDomains( domainPatterns, isDomainMatchRequired );
            for (String domain : domains) {
                response=executeSingleDomain( device, domain );
            }
        }
        else {
            response=executeSingleDomain( device, device.getCurrentDomain() );
        }
        
        return response;
    }

    /**
     * Returns the response from the domain quiesce/unquiesce action
     */
    public ResultResponse executeSingleDomain( Device device, String domain ) {
        logger.info( String.format( "The following domain will be %s: %s", quiesceAction.getVerbMessage(), domain ) );
        ResultResponse response=null;
        DPObject domainObj=new DPObject( null, domain );

        String actionName="Domain" + quiesceAction.getActionName();
        QuiesceUnquiesceRequest quiesceSingleCommand=new QuiesceUnquiesceRequest( actionName );
        int timeoutForDp=timeout;
        if (timeoutForDp < 0)
            timeoutForDp=quiesceAction.getMinTimeout();
        if (quiesceAction.isPassTimeoutToDP())
            quiesceSingleCommand.addObject( domainObj, timeoutForDp );
        else
            quiesceSingleCommand.addObject( domainObj );

        response=quiesceSingleCommand.execute( device.switchDefaultDomain() );

        if (StringUtils.equalsIgnoreCase( domain, Device.DEFAULT_DOMAIN_NAME)) {
            logger.info( String.format(
                            "'default' domain is always 'up', so we won't be checking for completion of the %s action",
                            quiesceAction ) );
        }
        else if (timeout < 0) {
            logger.info( String.format(
                            "Negative timeout was provided, so we won't be checking for completion of the %s action",
                            quiesceAction ) );
        }
        else {
            logger.info( String.format(
                            "Polling the device to check that the domain %s is in %s state. "
                                + "This might take a few seconds (up to the %d seconds timeout).",
                            domain, quiesceAction.getTargetOpState(), timeout ) );
            StateChecker stateChecker=new StateChecker( device.switchDefaultDomain() );

            stateChecker.pollStateUntilTimeout( new DPObject( "Domain", domain ), quiesceAction.getTargetOpState(),
                            timeout * 1000 );

            logger.info( String.format( "Domain '%s' has been %s successfully", domain,
                            quiesceAction.getVerbMessage() ) );
        }
        return response;
    }

}
