package com.myarch.dpbuddy.quiesce;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ResultResponse;
import com.myarch.dpbuddy.status.ObjectStatus;
import com.myarch.dpbuddy.status.StateChecker;

public class QuiesceUnquiesceObjectsCommand implements DPCommand<ResultResponse> {

    private final Log logger=LogFactory.getLog( getClass() );

    private QuiesceAction quiesceAction;
    // timeout in seconds
    private int timeout=-1;

    private List<DPObject> objectPatterns=new ArrayList<>();
    private List<DPObject> excludePatterns=new ArrayList<DPObject>();
    
    private Set<ObjectStatus> objectsToQuiesceOrUnquiesce;
    private boolean ignoreDisabled=true;

    public QuiesceUnquiesceObjectsCommand(QuiesceAction quiesceAction, Integer timeout) {
        this.quiesceAction=quiesceAction;
        this.timeout=quiesceAction.validateTimeout( timeout );
    }

    public void setIgnoreDisabled( boolean ignoreDisabled ) {
        this.ignoreDisabled=ignoreDisabled;
    }

    public void addObjectPattern( DPObject dpObject ) {
        objectPatterns.add( dpObject );
    }

    public void addObjectPatterns( List<DPObject> dpObjects ) {
        objectPatterns.addAll( dpObjects );
    }

    public void setExcludePatterns( List<DPObject> dpObjects ) {
        excludePatterns=dpObjects;
    }

    /**
     * returns the response from the last quiesce/unquiesce action
     */
    public ResultResponse execute( Device device ) {
        ResultResponse response=null;

        if (objectPatterns.size() == 0) {
            throw new DPBuddyException( "You must provide at least one object or object pattern to quiesce/unquiesce" );
        }

        // determine objects
        logger.info( "Querying the device to determine what objects to " + quiesceAction );
        objectsToQuiesceOrUnquiesce=device.fetchExistingObjects( objectPatterns, excludePatterns );

        if (objectsToQuiesceOrUnquiesce.isEmpty()) {
            throw new DPBuddyException( "Object patterns that you provided (%s) did not match any DataPower objects",
                            DPObject.objectListToString( objectPatterns ) );
        }

        if (ignoreDisabled) {
            objectsToQuiesceOrUnquiesce=filterOutDisabledObjects( objectsToQuiesceOrUnquiesce );
            if (objectsToQuiesceOrUnquiesce.isEmpty()) {
                logger.info( "All matching objects are disabled, nothing to do." );
                return response;
            }
        }

        logger.info( String.format( "The following objects will be %s:\n%s", quiesceAction.getVerbMessage(),
                        DPObject.objectListToString( objectsToQuiesceOrUnquiesce ) ) );

        String actionName="Service" + quiesceAction.getActionName();
        QuiesceUnquiesceRequest quiesceSingleCommand=new QuiesceUnquiesceRequest( actionName );
        int timeoutForDp=timeout;
        if (timeoutForDp < 0)
            timeoutForDp=quiesceAction.getMinTimeout();
        for (DPObject dpObject : objectsToQuiesceOrUnquiesce) {

            if (quiesceAction.isPassTimeoutToDP())
                quiesceSingleCommand.addObject( dpObject, timeoutForDp );
            else
                quiesceSingleCommand.addObject( dpObject );

        }

        response=quiesceSingleCommand.execute( device );

        if (timeout < 0) {
            logger.info( String.format(
                            "Negative timeout was provided, so we won't be checking for completion of the %s action",
                            quiesceAction ) );
        }
        else {
            logger.info( String.format(
                            "Polling the device to check that all objects are in %s state. "
                                + "This might take a few seconds (up to the %d seconds timeout).",
                            quiesceAction.getTargetOpState(), timeout ) );
            StateChecker stateChecker=new StateChecker( device );

            stateChecker.pollStateUntilTimeout( objectsToQuiesceOrUnquiesce, quiesceAction.getTargetOpState(),
                            timeout * 1000 );

            logger.info( String.format( "The following objects have been %s successfully:\n%s",
                            quiesceAction.getVerbMessage(),
                            DPObject.objectListToString( objectsToQuiesceOrUnquiesce ) ) );
        }

        return response;
    }

    private Set<ObjectStatus> filterOutDisabledObjects( Set<ObjectStatus> allObjects ) {
        Set<ObjectStatus> enabledObjects=new HashSet<>();
        Set<ObjectStatus> disabledObjects=new HashSet<>();
        for (ObjectStatus status : allObjects) {
            if (status.isEnabled() && !status.isDeleted()) {
                enabledObjects.add( status );
            }
            else {
                disabledObjects.add( status );
            }
        }

        if (disabledObjects.size() > 0) {
            logger.info( String.format( "The following objects are disabled, they will not be be %s:\n%s",
                            quiesceAction.getVerbMessage(), DPObject.objectListToString( disabledObjects ) ) );
        }

        return enabledObjects;
    }
}
