package com.myarch.dpbuddy.restart;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPConnectionException;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.action.ActionCommand;
import com.myarch.dpbuddy.quiesce.QuiesceUnquiesceDomainCommand;
import com.myarch.util.ThreadUtils;

public class RestartDPService {
    
    private final Log logger=LogFactory.getLog( getClass() );

    // reboot, reload, halt
    public final static String DEFAULT_RESTART_MODE="reload";
    public final static int DEFAULT_TIMEOUT=240;
    // ms
    public final static int DEFAULT_SHUTDOWN_WAIT_TIME=20000;

    private int shutdownWaitTime=DEFAULT_SHUTDOWN_WAIT_TIME;

    
    public void performRestart( Device device, String mode, int timeout, boolean isQuiesce ) {
        
        ActionCommand restartAction=new ActionCommand("Shutdown");
        restartAction.addActionChildElement( "Mode", mode );
        
        Device defaultDomainDevice=device.switchDefaultDomain();

        logger.info( "Restarting in the "+mode+" mode ..." );
        
        if (isQuiesce) {
            logger.info( "Quiescing all the domains before the restart ..." );
            QuiesceUnquiesceDomainCommand.QUIESCE_ALL_DOMAINS.execute( defaultDomainDevice );
        }
        
        try {
            defaultDomainDevice.executeRequestWithResultResponse( restartAction.getRequest() );
        }
        catch (DPConnectionException | DPBuddyException e) {
            logger.debug( "Ignoring the expected exception due to device being shutdown: "
                + e.toString() );
            logger.info( String.format( "The device is being restarted/rebooted using '%s' mode ...",
                            mode ) );
        }

        if (timeout < 0) {
            logger.info( "Negative timeout was provided, so we won't be waiting for DataPower to come back up" );
        }
        else {

            long startTime=System.nanoTime();

            logger.info( String.format( "Waiting for %d seconds for the restart process to begin ...",
                            DEFAULT_SHUTDOWN_WAIT_TIME / 1000 ) );
            ThreadUtils.delay( shutdownWaitTime );

            DPPoller poller=new DPPoller();
            logger.info( String.format( "Waiting for the device to come back up. "
                + "This will take some time (up to %d seconds timeout) ...", timeout ) );

            poller.pollUntilSuccess( defaultDomainDevice, timeout );
            long elapsedTime=System.nanoTime() - startTime;
            logger.debug( "Received IOException from the device in " + elapsedTime / 1000 / 1000
                + " ms" );

            logger.info( String.format(
                            "Restart has been completed in %d seconds, DataPower is up and running.",
                            elapsedTime / 1000 / 1000 / 1000 ) );
        }

    }

}
