package com.myarch.dpbuddy.status;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPCommand;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.log.DPObjectLogFilter;
import com.myarch.dpbuddy.log.GetLogCommand;
import com.myarch.dpbuddy.log.LogEntry;

public class AssertOperationalStateCommand implements DPCommand<OperationalStateResponse> {

    private final Log logger=LogFactory.getLog( getClass() );
    private List<DPObject> objectPatterns=new ArrayList<DPObject>();
    private List<DPObject> excludePatterns=new ArrayList<DPObject>();

    private OperationalState stateToAssert;

    // throw exception if not in state
    private boolean isExceptionIfNotInState=true;

    private boolean isObtainLogForFailedObjects=true;

    // private MultiMap<String, String> objectsToIgnore=new
    // MultiValueMap<String, String>();
//    private List<DPObject> objectsToIgnore=new ArrayList<DPObject>();

    public AssertOperationalStateCommand(OperationalState state) {
        stateToAssert=state;
    }

    public void setExceptionIfNotInState( boolean isExceptionIfNotInState ) {
        this.isExceptionIfNotInState=isExceptionIfNotInState;
    }

    public void addObjectPattern( DPObject dpObject ) {
        objectPatterns.add( dpObject );
    }

    public void addObjectPatterns( List<DPObject> dpObjects ) {
        objectPatterns.addAll( dpObjects );
    }

    public void setExcludePatterns( List<DPObject> dpObjects ) {
        excludePatterns=dpObjects;
    }

    /**
     * Returns all objects that are in the required state
     */
    public OperationalStateResponse execute( Device device ) {
        List<DPObject> objectsToCheck=new ArrayList<>(objectPatterns);
        if (objectsToCheck.isEmpty() ) {
            objectsToCheck.add( DPObject.ALL_OBJECTS );
        }

        logger.info( "Querying the device to get the list of objects to check ..." );
        Set<ObjectStatus> matchedObjects;

        matchedObjects=device.fetchExistingObjects( objectsToCheck, excludePatterns );

        Set<ObjectStatus> objectsNotInState=new HashSet<>();
        Set<ObjectStatus> objectsInState=new HashSet<>();
        for (ObjectStatus status : matchedObjects) {

            if (status.isEnabled() && !status.isDeleted()) {
                
                if (!status.isInState( stateToAssert ))
                    objectsNotInState.add( status );
                else
                    objectsInState.add( status );
                
            }
            else {
                logger.info( "This object is disabled, excluding it from the check: "
                    + status );
            }
        }

        if (matchedObjects.size() > 20)
            logger.info( String.format( "Checked the state of %d object(s)", matchedObjects.size() ) );
        else
            logger.info( String.format( "Checked the state of the following objects:\n%s",
                            DPObject.objectListToString( matchedObjects ) ) );

        if (objectsNotInState.size() > 0) {
            if (stateToAssert == OperationalState.UP && isObtainLogForFailedObjects)
                populateLogEntries( device, objectsNotInState );

            String errorMessage=String.format(
                            "State verification failed. The following objects are not in '%s' state:\n%s",
                            stateToAssert, formatFailedObjects( objectsNotInState ) );
            if (isExceptionIfNotInState)
                throw new DPBuddyException( errorMessage );
            else
                logger.info( errorMessage );
        }
        else {
            logger.info( "State verification passed. All objects are in " + stateToAssert + " state" );
        }

        return new OperationalStateResponse( stateToAssert, objectsInState, objectsNotInState );
    }
    
    private int numberOfLogEntriesToPrint=5;

    private String formatFailedObjects( Set<ObjectStatus> failedObjects ) {
        StringBuilder buf=new StringBuilder();
        for (ObjectStatus dpObject : failedObjects) {
            if (buf.length() > 0 && !buf.toString().endsWith( "\n" ))
                buf.append( "\n" );
            buf.append( "- " + dpObject.toClassObjectStr() );
            String errorCode=dpObject.getErrorCode();
            if (StringUtils.isNotBlank( errorCode )) {
                buf.append( " (" + errorCode + ")" );
            }
            List<LogEntry> logEntries=dpObject.getLogEntries();
            if (logEntries.size() > 0) {
                buf.append( " Log entries:\n" );
                int logEntriesCount=0;
                for (LogEntry logEntry : logEntries) {
                    if (logEntriesCount >= numberOfLogEntriesToPrint)
                        break;

                    buf.append( logEntry.getLogString() + "\n" );
                    ++logEntriesCount;
                }

            }
        }
        return buf.toString();
    }

    private void populateLogEntries( Device device, Iterable<ObjectStatus> downObjects ) {
        GetLogCommand getLog=new GetLogCommand();
        getLog.setPrintLog( false );
        List<LogEntry> logEntries=getLog.execute( device );

        for (ObjectStatus downObject : downObjects) {
            List<LogEntry> objectLogEntries=findLogEntriesForObject( logEntries, downObject );
            if (objectLogEntries.size() > 0)
                downObject.setLogEntries( objectLogEntries );
        }
    }

    private List<LogEntry> findLogEntriesForObject( List<LogEntry> logEntries, DPObject dpObject ) {
        DPObjectLogFilter filter=new DPObjectLogFilter( null, dpObject.getName() );
        filter.setErrorOnly( true );
        return GetLogCommand.filterLogEntries( logEntries, filter );
    }
}
