package com.myarch.dpbuddy.status;

import java.util.*;

import org.apache.tools.ant.BuildException;
import org.kohsuke.args4j.Option;

import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ObjectListParser;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.util.StringPatternUtils;

public class ObjectStatusAntTask extends BaseDPBuddyTask {
    // device, domain, class, name, op state, admin state, config state
    public final static int NAME_LENGTH=23;
    public final static int STATE_LENGTH=9;

    public final static String FORMAT="%-10.9s%-17.16s%-19.18s%-23.22s%-9s%-9s%-9s";
    public final static List<String> HEADERS=Arrays.asList( "Device", "Domain", "Type (class)", "Name", "Op St.",
                    "Adm St.", "Conf St." );

    public final static int TOTAL_LENGTH=10 + 18 + 19 + NAME_LENGTH + STATE_LENGTH * 3;

    private List<DPObject> includeObjects=new ArrayList<>();
    private List<DPObject> excludeObjects=new ArrayList<>();

    private String domainPatternsString;

    @Option(name="-domainPatterns", aliases="-domains",
                    usage="Comma-delimited list of regular expression patterns defining what domains to query.")
    public void setDomainPatterns( String domainPatterns ) {
        this.domainPatternsString=domainPatterns;
    }

    @Option(name="-objects",
                    usage="Comma-delimited list of object patterns to include in the report in the format 'classPatter:namePattern, classPatter:namePattern'.")
    public void setObjects( String includeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( includeObjectsList );
        includeObjects.addAll( objs );
    }

    @Option(name="-exclude",
                    usage="Comma-delimited list of object patterns to exclude in the format 'classPatter:namePattern, classPatter:namePattern'.")
    public void setExclude( String excludeObjectsList ) {
        List<DPObject> objs=ObjectListParser.parsePatternList( excludeObjectsList );
        excludeObjects.addAll( objs );
    }

    @Override
    public void executeDPTask() throws BuildException {

        Device device=getDevice();
        Set<String> domains=new HashSet<>();
        if (domainPatternsString != null) {
            domains=device.fetchRequiredDomains( StringPatternUtils.stringToPattnernList( domainPatternsString ) );
        }
        else {
            domains.add( device.getCurrentDomain() );
        }
        if (includeObjects.isEmpty()) {
            includeObjects.add( new DPObject( ".*", null ) );
        }

        logger.info( TabularReport.genHeader( HEADERS, FORMAT, TOTAL_LENGTH ) );

        boolean failIfNoObjects=false;
        for (String domain : domains) {

            Set<ObjectStatus> objects=device.switchDomain( domain ).fetchObjects( includeObjects, excludeObjects,
                            failIfNoObjects );
            if (!objects.isEmpty()) {
                List<List<Object>> data=statusesToRows( device.getName(), domain, objects );
                logger.info( TabularReport.genRows( data, FORMAT ) );
            }
        }
    }

    private List<List<Object>> statusesToRows( String deviceName, String domain, Set<ObjectStatus> objects ) {
        List<List<Object>> data=new ArrayList<>();
        for (ObjectStatus object : objects) {
            List<Object> row=new ArrayList<>();
            row.add( deviceName );
            row.add( domain );
            row.add( object.getClazz() );
            row.add( object.getName() );
            row.add( object.getOpState() );
            row.add( object.getAdminState() );
            row.add( object.getConfigState() );

            data.add( row );
        }

        return data;
    }
    
    @Override
    protected boolean isSuppressDomainLogging() {
        return true;
    }
    
}
