package com.myarch.dpbuddy.status;

import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Document;
import org.jdom2.Element;

import com.myarch.dpbuddy.BaseResponse;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.DPObjectMatcher;

public class ObjectStatusResponse extends BaseResponse {
    private Log logger=LogFactory.getLog( getClass() );
    private DPObjectMatcher objectMatcher=new DPObjectMatcher();

    private StatusResponse statusResponse;

    public ObjectStatusResponse(Document doc) {
        super( doc );
        statusResponse=new StatusResponse( "ObjectStatus", doc );
    }

    /**
     * Returns a list of object statuses for the list of filters containing
     * regexps. Each filter can have class or object name missing. Both class
     * and object names could be regexps.
     * 
     * @param filters
     * @return list of object statuses
     */
    public Set<ObjectStatus> getObjectStatuses( DPObject... filters ) {
        Set<ObjectStatus> objects=getObjectStatuses( Arrays.asList( filters ), Collections.<DPObject> emptyList(),
                        false );
        if (objects.size() == 0 && filters.length == 1) {
            DPObject filter=filters[0];
            String message="Could not find a DataPower object matching ";
            if (filter.getClazz() != null) {
                message+="class pattern " + filter.getClazz() + " ";
            }
            if (filter.getName() != null) {
                if (filter.getClazz() != null)
                    message+="and ";
                message+="object name pattern " + filter.getName();
            }
            throw new DPBuddyException( message );
        }

        return objects;
    }

    public Set<ObjectStatus> getObjectStatuses( Iterable<? extends DPObject> includeFilters,
                    Iterable<? extends DPObject> excludeFilters, boolean mustMatch ) {
        Iterable<ObjectStatus> statuses=getObjectStatuses();
        return objectMatcher.matchObjectStatuses( statuses, includeFilters, excludeFilters, mustMatch );
    }

    public boolean checkObjecExsists( DPObject filter ) {

        Set<ObjectStatus> matchedStatuses=new HashSet<ObjectStatus>();

        for (ObjectStatus status : getObjectStatuses()) {
            if (DPObject.objectMatchesFilter( status, filter )) {
                matchedStatuses.add( status );
            }
        }
        int matchedSize=matchedStatuses.size();
        if (matchedSize == 0) {
            logger.debug( "No object matched the provided filter" );
            return false;
        }
        else if (matchedSize > 1) {
            logger.debug( "Following objects matched the provided filter:\n"
                + DPObject.objectListToString( matchedStatuses ) );
            return false;
        }

        logger.debug( "Object exists in the environment: " + DPObject.objectListToString( matchedStatuses ) );
        return true;
    }

    /**
     * Returns list of all object statuses except the ones in deleted state
     * 
     * @return list object statuses
     */
    public Set<ObjectStatus> getObjectStatuses() {
        Set<ObjectStatus> objectStatuses=new HashSet<ObjectStatus>();
        List<Element> elts=statusResponse.getStatusElements();

        for (Element elt : elts) {
            ObjectStatus objStatus=new ObjectStatus( elt );
            if (!objStatus.isDeleted())
                objectStatuses.add( objStatus );
        }

        return objectStatuses;
    }

    /**
     * Returns list of all object statuses except the ones in saved state
     * 
     * @return list object statuses
     */
    public Set<ObjectStatus> getUnsavedObjectStatuses() {
        Set<ObjectStatus> objectStatuses=new HashSet<ObjectStatus>();
        List<Element> elts=statusResponse.getStatusElements();

        for (Element elt : elts) {
            ObjectStatus objStatus=new ObjectStatus( elt );
            if (!objStatus.isSaved())
                objectStatuses.add( objStatus );
        }

        return objectStatuses;
    }

}
