package com.myarch.dpbuddy.status;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.apache.tools.ant.BuildException;
import org.kohsuke.args4j.Option;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.JsonObject;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPConnectionException;
import com.myarch.dpbuddy.Device;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.dpbuddy.cli.BoolOptionHandler;
import com.myarch.dpbuddy.cli.DPBuddyCliExecutor;
import com.myarch.dpbuddy.cli.DPBuddyMain;
import com.myarch.eventmodel.Event;
import com.myarch.eventmodel.EventType;
import com.myarch.util.StringPatternUtils;

/**
 * Executes "do-status request.
 */
public class StatusAntTask extends BaseDPBuddyTask {

    public final static String STATUS_LOGGER_NAME="dp-status";
    private Logger jsonStatusLogger=LoggerFactory.getLogger( STATUS_LOGGER_NAME );

    @Option(name="-class", aliases="-type", required=true,
                    usage="Type ('class') of the DataPower object, e.g., FilesystemStatus. See xml-mgmt.xsd, StatusEnum for a complete list.")
    private String clazz;

    public void setClass( String clazz ) {
        this.clazz=clazz;
    }

    // for now, only used from JSON
    private List<String> statusTypes=new ArrayList<>();

    private boolean isJson;
    private boolean isPrintToJsonLogOnly=false;
    private String domainPatternsString;

    @Option(name="-json", handler=BoolOptionHandler.class,
                    usage="Output status in JSON format as opposed to a more human-readable name:value format. Defaults to false.")
    public void setJson( boolean isJson ) {
        this.isJson=isJson;
        setLogStartAndCompletion( !isJson );
    }

    @Option(name="-domainPatterns", aliases="-domains",
                    usage="Comma-delimited list of regular expression patterns defining what domains to query.")
    public void setDomainPatterns( String domainPatterns ) {
        this.domainPatternsString=domainPatterns;
    }

    private List<Event> jsonEvents=new ArrayList<>();

    @Override
    public void executeDPTask() throws BuildException {

        Device device=getDevice();
        if (isJson) {
            setLogStartAndCompletion( false );

            List<String> statusTypesToUse=new ArrayList<>();
            if (clazz != null) {
                statusTypesToUse.add( clazz );
            }
            statusTypesToUse.addAll( statusTypes );
            Set<String> domains=new HashSet<>();
            if (domainPatternsString != null) {
                domains=getDevice().fetchRequiredDomains(
                                StringPatternUtils.stringToPattnernList( domainPatternsString ) );
            }
            else {
                domains.add( device.getCurrentDomain() );
            }
            List<Event> events=fetchJsonEvents( device, domains, statusTypesToUse );

            printJson( events );
            jsonEvents.addAll( events );
        }
        else {
            validator.validateRequired( clazz, "class", "Status class" );
            validator.throwIfErrors();

            StatusCommand statusRequest=new StatusCommand( clazz );
            StatusResponse response=statusRequest.execute( getDevice() );
            logger.info( String.format( "Status for the class '%s':", StatusUtils.getDisplayClass( clazz ) ) );
            String statusString=response.toStatusString();
            if (statusString.length() > 0) {
                logger.info( statusString );
            }
            else {
                logger.info( "No object of the given class was found" );
            }
        }
    }

    /**
     * @return Json events produced by the execute
     */
    private List<Event> getJsonEvents() {
        return jsonEvents;
    }

    // used by logstash plugins
    public List<Event> executeAndProduceJson( List<String> statusTypes ) {
        this.statusTypes=statusTypes;
        setJson( true );
        jsonEvents.clear();
        isPrintToJsonLogOnly=true;
        execute();
        return getJsonEvents();
    }

    public static StatusAntTask instantiate( String configFilePath ) {
        DPBuddyCliExecutor commandFactory=new DPBuddyCliExecutor( DPBuddyMain.getCliCmdDefs() );
        BaseDPBuddyTask task=commandFactory.instantiateTask( "status" );
        StatusAntTask statusTask=(StatusAntTask) task;
        // FIXME
        statusTask.statusTypes.clear();
        if (configFilePath != null) {
            statusTask.setConfigFile( configFilePath );
        }
        commandFactory.setProjectProperties( statusTask );

        return statusTask;
    }

    private List<Event> fetchJsonEvents( Device device, Set<String> domains, List<String> statusTypes ) {
        List<Event> jsonEvents=new ArrayList<>();

        for (String statusType : statusTypes) {
            jsonEvents.addAll( fetchJsonEventsForType( device, domains, statusType ) );
            /*
            if (statusTypes.size()>1) {
                try {
                    Thread.sleep( 10 );
                }
                catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
            */
        }

        return jsonEvents;
    }

    private List<Event> fetchJsonEventsForType( Device device, Set<String> domains, String statusType ) {
        List<Event> jsonEvents=new ArrayList<>();
        List<String> domainsForCurType=new ArrayList<>();
        if (isMultiDomainStatus( statusType )) {
            domainsForCurType.addAll( domains );
        }
        else {
            domainsForCurType.add( device.getCurrentDomain() );
        }

        for (String domain : domainsForCurType) {
            jsonEvents.addAll( fetchJsonEventsForDomainAndType( device, domain, statusType ) );
        }
        // Before uploading the first event, update all the ones that exist to
        // make them history
        setBatchStart( jsonEvents );

        return jsonEvents;
    }

    // TODO: more elegant way -- batch_start, batch_end, batch_id
    private void setBatchStart( List<Event> jsonEvents ) {
        if (jsonEvents.size() > 0) {
            Event firstEvent=jsonEvents.get( 0 );
            firstEvent.setBatchStart( true );
        }
    }

    // TODO: should be a parameter
    public final static String DEFAULT_MULTI_DOMAIN_TYPES="ServicesStatus, ObjectStatus";

    private String multiDomainTypesStr=DEFAULT_MULTI_DOMAIN_TYPES;

    private boolean isMultiDomainStatus( String statusType ) {
        String[] types=StringUtils.splitByWholeSeparator( multiDomainTypesStr, "," );
        for (String type : types) {
            if (statusType.equalsIgnoreCase( type.trim() ))
                return true;
        }
        return false;
    }

    private List<Event> fetchJsonEventsForDomainAndType( Device device, String domain, String statusType ) {

        StatusCommand statusRequest=new StatusCommand( statusType );
        List<Event> jsonEvents=new ArrayList<>();
        JsonStatusEventCreator jsonCreator=JsonStatusEventCreator.instance();

        try {
            StatusResponse response=statusRequest.execute( device, domain );

            // assuming that the array contains only objects
            for (JsonObject jsonObj : response.toJson()) {
                Event event=jsonCreator.createStatusEvent( getDevice(), domain, propsFromEnvContext(), statusType,
                                jsonObj );
                jsonEvents.add( event );
            }

        }
        catch (DPConnectionException | DPBuddyException e) {
            logger.error( e );

            EventType eventType=EventType.ERROR;
            if (e instanceof DPConnectionException) {
                eventType=EventType.CONNECTION_ERROR;
            }

            Event errorEvent=jsonCreator.createErrorEvent( device, eventType, e );
            jsonEvents.add( errorEvent );
        }

        return jsonEvents;

    }

    private JsonObject propsFromEnvContext() {
        // TODO: accommodate tags as well
        // for now it's just the environment
        JsonObject json=new JsonObject();
        String environment=getEnvironmentLabel();
        if (environment != null) {
            json.addProperty( "environment", environment );
        }
        return json;
    }

    private void printJson( List<Event> jsonEvents ) {
        for (Event event : jsonEvents) {
            String eventStr=event.json().toString();
            // log to the separate json file
            jsonStatusLogger.info( eventStr );
            if (!isPrintToJsonLogOnly) {
                // log to the console using ant logger which will also invoke
                // logback logger
                logger.info( eventStr );
            }
        }
    }

}
