package com.myarch.dpbuddy.status;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.jdom2.Document;
import org.jdom2.Element;

import com.google.gson.JsonObject;
import com.myarch.dpbuddy.BaseResponse;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.util.JsonUtils;

public class StatusResponse extends BaseResponse {

    // private static Log logger = LogFactory.getLog( StatusResponse.class );

    private String requestClass;

    private StatusResponse(Document doc) {
        super( doc );
    }

    public StatusResponse(String clazz, Document doc) {
        this( doc );
        this.requestClass=clazz;
        validate();
    }

    public String getStatusClass() {
        return requestClass;
    }

    private void validate() {
        // make sure that we have at lease one status
        getRequiredResponseElements( "dp:status" );
    }

    public List<String> getStatusValues( String statusEltName ) {
        List<String> statusValues=new ArrayList<String>();

        List<Element> statusElts=getStatusElements();
        for (Element statusElt : statusElts) {
            List<Element> statusChildren=statusElt.getChildren( statusEltName );
            for (Element statusChildElt : statusChildren) {
                statusValues.add( statusChildElt.getTextTrim() );
            }
        }

        return statusValues;
    }

    /**
     * 
     * @return element names and values of the status entry in a map. Raises an
     *         error if there is more than one status elements
     */
    public Map<String, String> getSingleStatusValueMap() {
        List<Element> statusElts=getStatusElements();
        if (statusElts.size() > 1)
            throw new IllegalStateException( "Found multiple status elements, expected only one" );
        Element statusElt=statusElts.get( 0 );
        return statusElementToMap( statusElt );
    }

    public List<Element> getStatusElements() {
        List<Element> statusElements=new ArrayList<Element>();

        List<Element> statusRoots=getRequiredResponseElements( "dp:status" );
        if (statusRoots.size() > 1)
            throw new DPBuddyException( "Encountered multiple 'status' elements" );

        if (statusRoots.size() == 0)
            return statusElements;
        Element statusRoot=statusRoots.get( 0 );

        return (List<Element>) statusRoot.getChildren();
    }

    public String getStatusValue( String identifyingElementName, String identifyingValue,
                    String valueElementName ) {
        Element elt=getRequiredStatusWhereElement( identifyingElementName, identifyingValue );
        Element valueElement=elt.getChild( valueElementName );
        if (valueElement != null)
            return valueElement.getText();
        return null;
    }

    public Element getRequiredStatusWhereElement( String elementName, String elementVal ) {
        Element elt=getStatusWhereElement( elementName, elementVal );
        if (elt == null)
            throw new DPBuddyException( "Status for %s=%s was not found", elementName, elementVal );
        return elt;
    }

    public Element getStatusWhereElement( String elementName, String elementVal ) {
        List<Element> statusElts=getStatusElements();
        for (Element statusElt : statusElts) {
            Element identifyingElement=statusElt.getChild( elementName );
            if (identifyingElement != null && identifyingElement.getText() != null
                && identifyingElement.getTextTrim().equalsIgnoreCase( elementVal ))
                return statusElt;
        }

        return null;
    }

    public final static String STATUS_ATTR_FORMAT="%s: %s";

    public String toStatusString() {
        // name: value, name: value
        List<Element> statusElts=getStatusElements();
        StringBuilder buf=new StringBuilder();
        // we can get statuses of multiple objects. 
        // each status item is displayed as name:value
        for (Element statusElt : statusElts) {
            buf.append( statusElementToString( statusElt ) );
            buf.append( "\n" );
        }

        return buf.toString();
    }
    
    public List<JsonObject> toJson() {
        List<JsonObject> jsonArr=new ArrayList<>();
        List<Element> statusElts=getStatusElements();
        for (Element statusElt : statusElts) {
            JsonObject statusEltJson=statusElementToJson( statusElt );
            jsonArr.add( statusEltJson );
        }

        return jsonArr;
    }

    private JsonObject statusElementToJson(Element statusElt) {
        JsonObject jsonObj=new JsonObject();
        for (Element statusChildElt : statusElt.getChildren()) {
            JsonUtils.addPropertyGuessType( jsonObj, statusChildElt.getName(), statusChildElt.getTextTrim() );
        }
        
        return jsonObj;
    }
    
    public static Map<String, String> statusElementToMap( Element statusElt ) {
        Map<String, String> statusValMap=new HashMap<String, String>();
        for (Element statusChildElt : statusElt.getChildren()) {
            statusValMap.put( statusChildElt.getName(), statusChildElt.getTextTrim() );
        }

        return statusValMap;
    }

    public static String statusElementToString( Element statusElt ) {
        StringBuilder lineBuf=new StringBuilder();
        for (Element statusChildElt : statusElt.getChildren()) {

            String statusValue=statusChildElt.getTextTrim();
            if (StringUtils.isNotBlank( statusValue )) {
                if (lineBuf.length() > 0)
                    lineBuf.append( ", " );
                lineBuf.append( String.format( STATUS_ATTR_FORMAT, statusChildElt.getName(),
                                statusValue ) );
            }
        }

        return lineBuf.toString();
    }

}
