package com.myarch.dpbuddy.xmltransform;

import java.io.File;
import java.util.List;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.types.FileSet;

import com.myarch.antutil.AntUtils;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPConnection;
import com.myarch.dpbuddy.ExportImportFileType;
import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.dpbuddy.xmlutil.XMLInput;
import com.myarch.dpbuddy.xmlutil.XMLParsingUtils;
import com.myarch.propmanagement.ExpressionEvaluator;

public class FileTransformer {
    
    private final Log logger = LogFactory.getLog(getClass());
    
    private ExpressionEvaluator expressionEvaluator;
    private Project antProject;
    
    /**
     * @param expressionEvaluator
     */
    public FileTransformer(Project antProject, ExpressionEvaluator expressionEvaluator) {
        super();
        this.expressionEvaluator = expressionEvaluator;
        this.antProject=antProject;
    }

    public void transformAndPersistPatternSets(File baseDir, List<TransformablePatternSet> tranformablePatternSets) {
        for( TransformablePatternSet tranformablePatternSet: tranformablePatternSets) {
            transformAndPersistPatternSet(baseDir, tranformablePatternSet);
        }
    }
    
    private void transformAndPersistPatternSet(File baseDir, TransformablePatternSet tranformablePatternSet) {
        
        logger.debug(String.format("Looking for files to update in %s", baseDir));
        
        FileSet fileSet=AntUtils.createFileSetFromPatternSet(antProject, baseDir, tranformablePatternSet);
        
        fileSet.setDir(baseDir);
        DirectoryScanner ds = fileSet.getDirectoryScanner();
        
        String [] files = ds.getIncludedFiles();
        
        int transformedFileCount=0;
        for( String file:files ) {
            String filePath = FilenameUtils.concat(baseDir.getAbsolutePath(), file);
            // exclude zips
            if (ExportImportFileType.isZipExtension(new File(filePath))) {
                logger.debug(String.format("Ignoring file '%s' because it's a zip file",filePath ));
            }
            else {
                transformAndPersistFile(new File(filePath), tranformablePatternSet.getTransform() );
                ++transformedFileCount;
            }
        }
        
        if (transformedFileCount==0)
            throw new DPBuddyException("Provided %s didn't match any of the files. At lease one file must match.", tranformablePatternSet.toString());
    }
    
    private void transformAndPersistFile(File file, TransformActionContainer transformActionContainer) {
        XMLInput xmlInput=transformXMLFile(file, transformActionContainer);
        saveFile(file, xmlInput);
    }
    
    public XMLInput transformXMLFile(File file, TransformActionContainer transformActionContainer) {
        
        logger.info(String.format("Transforming file '%s' ", file.getAbsolutePath()) );
        
        XMLInput xmlInput=transformXMLInput(new XMLInput(file), transformActionContainer);
        return xmlInput;
    }
    
    public XMLInput transformXMLInput(XMLInput xmlInput, TransformActionContainer transformActionContainer) {
        
        xmlInput.transform(transformActionContainer);
        xmlInput.resolveVars(expressionEvaluator);
        
        String xmlString=JDomUtils.docToCutBase64String( xmlInput.getDocument() );
        logger.debug("Transformed XML file after all transformations and property substitution:\n"+xmlString);
        return xmlInput;
    }    
    
    
    public XMLInput transformXMLFile(File file, 
                    TransformActionContainer transform, DPConnection dpConnection){
        transform.setDPConnection(dpConnection);
        return transformXMLFile(file, transform);      
    }
    
    public XMLInput transformXMLFileIfTransformProvided(File file, 
                    TransformActionContainer transform, DPConnection dpConnection){
        XMLInput resultingFile; 
        if (transform!=null) {
            transform.setDPConnection(dpConnection);
            resultingFile=transformXMLFile(file, transform);
        }
        else {
            resultingFile=new XMLInput(file);
        }
        
        return resultingFile;
    }
    
    
    public void saveFile(File file, XMLInput xml) {
        XMLParsingUtils.persistXML(file, xml.getDocument());
    }
    
}
