package com.myarch.dpbuddy.xmltransform;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.tools.ant.types.DataType;
import org.jdom2.Document;
import org.jdom2.Namespace;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPConnection;

public class TransformActionContainer extends DataType {

    private Log logger=LogFactory.getLog( getClass() );

    private List<Namespace> namespaces=new ArrayList<Namespace>();
    // recursive containers for references
    private List<TransformActionContainer> transformContainers=new ArrayList<TransformActionContainer>();

    private List<TransformAction> transformActions=new ArrayList<TransformAction>();
    private DPConnection connectionFilter;

    private String ifExpression;

    private boolean isVerbose=false;
    private TransformLogger transformLogger=new TransformLogger( logger );

    public void setVerbose( boolean isVerbose ) {
        // for references
        this.isVerbose=isVerbose;
        transformLogger.setVerbose( isVerbose );
    }

    boolean isVerbose() {
        return isVerbose;
    }

    // The following method are needed for Ant introspection:

    public void addConfiguredNamespace( NamespaceConfig namespaceConfig ) {
        namespaces.add( namespaceConfig.getNamespace() );
    }

    public void insertFirst( TransformAction action ) {
        transformActions.add( 0, action );
    }

    public void addAdd( AddUpdateTransformAction action ) {
        action.setRemoveBeforeAdd( false );
        addAction( action );
    }

    public void addDelete( DeleteTransformAction action ) {
        addAction( action );
    }

    public void addUpdate( AddUpdateTransformAction action ) {
        action.setRemoveBeforeAdd( true );
        addAction( action );
    }

    public void addUpdateValues( UpdateValuesTransformAction action ) {
        addAction( action );
    }

    public void addSetText( SetTextTransformAction action ) {
        addAction( action );
    }

    public void addReplaceText( ReplaceTextTransformAction action ) {
        addAction( action );
    }

    public void addDPInclude( IncludeExcludeDPObjectAction action ) {
        addAction( action );
    }

    public void addDPExclude( IncludeExcludeDPObjectAction action ) {
        action.setExclude( true );
        addAction( action );
    }

    public void addTransform( TransformActionContainer container ) {
        transformContainers.add( container );
    }

    private void addAction( TransformAction action ) {
        transformActions.add( action );
    }

    public void setIf( String ifExpression ) {
        this.ifExpression=ifExpression;
    }

    public void setDPConnection( DPConnection connectionFilter ) {
        this.connectionFilter=connectionFilter;
    }

    public Document transform( Document doc ) {
        return transform( doc, connectionFilter );
    }

    public List<Namespace> getNamespaces() {
        List<Namespace> allNamespaces=new ArrayList<Namespace>();
        if (isReference()) {
            TransformActionContainer referencedContainer=( (TransformActionContainer) getCheckedRef() );
            allNamespaces.addAll( referencedContainer.getNamespaces() );
        }

        for (TransformActionContainer container : transformContainers) {
            allNamespaces.addAll( container.getNamespaces() );
        }

        allNamespaces.addAll( namespaces );
        return allNamespaces;
    }

    public List<TransformAction> getTransformActions() {
        List<TransformAction> allTransformActions=new ArrayList<TransformAction>();
        if (isReference()) {
            TransformActionContainer referencedContainer=( (TransformActionContainer) getCheckedRef() );
            if (referencedContainer.isVerbose) {
                setVerbose( true );
            }
            allTransformActions.addAll( referencedContainer.getTransformActions() );
        }
        // actions from nested containers are added before native actions
        for (TransformActionContainer container : transformContainers) {
            allTransformActions.addAll( container.getTransformActions() );
        }

        allTransformActions.addAll( transformActions );

        return allTransformActions;
    }

    private Document transform( Document doc, DPConnection dpConnection ) {
        transformLogger.setProject( getProject() );

        Document transformedDoc=doc;

        GroovyExpressionEvaluator expressionEvaluator=new GroovyExpressionEvaluator( getProject(), dpConnection );
        if (StringUtils.isNotBlank( ifExpression )) {
            boolean isTransformContainerApplicable=expressionEvaluator.evaluateBoolean( ifExpression );
            if (!isTransformContainerApplicable) {
                transformLogger.log( "Transformations were not applied because '%s' expression evaluated to 'false'",
                                ifExpression );
                return doc;
            }
        }

        for (TransformAction action : getTransformActions()) {
            boolean isActionApplicable=true;

            if (action.getIfExpression() != null) {
                isActionApplicable=expressionEvaluator.evaluateBoolean( action.getIfExpression() );
            }

            if (isActionApplicable) {
                transformLogger.log( "Applying transformation %s", action.toDisplayString() );
                if (!action.isVerboseSet()) {
                    action.setVerbose( transformLogger.isVerbose() );
                }
                action.setNamespaces( getNamespaces() );
                transformedDoc=action.transform( transformedDoc );
            }
            else {
                transformLogger.log(
                                "Transformation %s was not applied because its IF expression '%s' evaluated to 'false'",
                                action.toDisplayString(), action.getIf() );
            }

        }
        return transformedDoc;
    }

    public static TransformActionContainer createTransformOnlyIfNull(
                    TransformActionContainer currentTransformActionContainer ) {
        if (currentTransformActionContainer == null)
            return new TransformActionContainer();
        else
            throw new DPBuddyException( "You can specify only one nested 'transform' element" );
    }

}
