package com.myarch.dpbuddy.xmltransform;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdom2.Attribute;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.filter.ElementFilter;
import org.jdom2.util.IteratorIterable;

import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPObject;
import com.myarch.dpbuddy.xmlutil.JDomUtils;
import com.myarch.hocon.HoconUtils;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;

public class UpdateValuesTransformAction extends TransformAction {

    @SuppressWarnings("unused")
    private Log logger=LogFactory.getLog( getClass() );

    // TODO: match all by default

    // preserve for logging only
    private String updatesHoconStr;
    private String applyUpdatesToHoconStr;

    private List<DPObject> applyUpdatesTo=new ArrayList<>();
    private List<Pair<String, String>> updates;

    public UpdateValuesTransformAction() {
    }

    public UpdateValuesTransformAction(String updatesHoconStr, String applyUpdatesToHoconStr) {
        setUpdates( updatesHoconStr );
        setApplyUpdatesTo( applyUpdatesToHoconStr );
    }

    public void setUpdates( String updatesHoconStr ) {
        this.updatesHoconStr=updatesHoconStr;

        Config updatesConfig=ConfigFactory.parseString( updatesHoconStr );
        updates=HoconUtils.configToNameStringValueList( updatesConfig );
    }

    public void setApplyUpdatesTo( String applyUpdatesToHoconStr ) {
        if (StringUtils.isNotBlank( applyUpdatesToHoconStr )) {
            this.applyUpdatesToHoconStr=applyUpdatesToHoconStr;

            Config applyUpdatesToConfig=ConfigFactory.parseString( applyUpdatesToHoconStr );
            applyUpdatesTo=HoconUtils.configToDPObjectList( applyUpdatesToConfig );
        }
        else {
            // match all
            applyUpdatesTo.add( new DPObject() );
        }
    }

    public Document transform( Document doc ) {
        Document transformedDoc=doc.clone();

        for (DPObject filter : applyUpdatesTo) {
            List<Element> elementsToTransform=findObjectElements( transformedDoc, filter );
            for (Element elt : elementsToTransform) {
                logTransform( "Applying updates to '%s' ...", new DPObject( elt ).toClassObjectStr() );
                updateElementsForObject( elt, updates );
            }
        }

        return transformedDoc;
    }

    private void updateElementsForObject( Element objectParentElt, List<Pair<String, String>> updates ) {
        IteratorIterable<Element> descendantsIter=objectParentElt.getDescendants( new ElementFilter() );

        List<Element> elementsToCheckForUpdates=new ArrayList<>();
        elementsToCheckForUpdates.add( objectParentElt );
        // we can't apply changes to the elements in the iterator
        for (Element descendant : descendantsIter) {
            elementsToCheckForUpdates.add( descendant );
        }

        for (Element descendant : elementsToCheckForUpdates) {
            applyUpdatesToElementIfMatches( descendant, updates );
        }
    }

    private void applyUpdatesToElementIfMatches( Element elt, List<Pair<String, String>> updates ) {
        for (Pair<String, String> update : updates) {
            if (elt.getName().equalsIgnoreCase( update.getKey() )) {
                logTransform( "Replaced text '%s' contained within the element '%s' with '%s'", elt.getTextTrim(),
                                JDomUtils.elementToString( elt ), maskIfNeeded( elt.getName(), update.getValue() ) );
                elt.setText( update.getValue() );
            }
            else {
                for (Attribute attr : elt.getAttributes()) {
                    if (attr.getName().equalsIgnoreCase( update.getKey() )) {
                        logTransform( "Replaced the value '%s' of the attribute '%s' with '%s'", attr.getValue(),
                                        JDomUtils.attributeToString( attr ),
                                        maskIfNeeded( attr.getName(), update.getValue() ) );

                        attr.setValue( update.getValue() );
                    }
                }
            }
        }
    }

    private List<Element> findObjectElements( Document doc, DPObject filter ) {

        Element configElt=doc.getRootElement().getChild( "configuration" );
        if (configElt == null)
            throw new DPBuddyException( "Did not find 'configuration' element in the provides XML file. "
                + "Update action can only work with XML files in the DataPower configuration format" );

        List<Element> objectElts=configElt.getChildren();
        List<Element> elementsToTransform=new ArrayList<>();
        for (Element objectElt : objectElts) {
            boolean matched=elementMatches( objectElt, filter );
            if (matched) {
                elementsToTransform.add( objectElt );
            }
        }
        return elementsToTransform;
    }

    private boolean elementMatches( Element objectElt, DPObject filter ) {
        DPObject dpObject=new DPObject( objectElt );
        return DPObject.objectMatchesFilter( dpObject, filter );
    }

    @Override
    public String toDisplayString() {
        String applyToDisplay=applyUpdatesToHoconStr;
        if (applyToDisplay == null) {
            applyToDisplay="All objects";
        }
        String displayString=String.format( "%s('%s' applied to '%s')", getTransformActionName(), updatesHoconStr,
                        applyToDisplay );
        return displayString;
    }
}
