package com.myarch.dpbuddy.xmlutil;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.nio.charset.Charset;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamReader;

import org.apache.commons.io.FileUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.stax2.XMLInputFactory2;
import org.codehaus.stax2.XMLStreamReader2;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.DOMBuilder;
import org.jdom2.input.StAXStreamBuilder;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.myarch.dpbuddy.DPBuddyException;

public class XMLParsingUtils {

    private static Log logger=LogFactory.getLog( XMLParsingUtils.class );

    public static Document parseString( String xml ) {
        return parseUsingStax2( new StringReader( xml ) );
    }

    public static Document parseDOM( DocumentFragment domFragment ) {
        Document doc=new Document();
        DOMBuilder builder=new DOMBuilder();
        NodeList nodes=domFragment.getChildNodes();

        for (int i=0; i < nodes.getLength(); ++i) {
            Node node=nodes.item( i );
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                org.w3c.dom.Element domElement=(org.w3c.dom.Element) node;
                doc.addContent( (Element) builder.build( domElement ).clone() );
            }
        }
        return doc;
    }

    public static Document parseFile( File file ) {
        long startTime=System.nanoTime();
        Document doc=null;
        try {
            DocumentBuilder documentBuilder=getDomDocumentBuilder();
            org.w3c.dom.Document domDoc=documentBuilder.parse( file );
            doc=buildFromDomDoc( domDoc );
        }
        catch (Exception e) {
            raiseParsingException( e, null );
        }

        long elapsedTime=System.nanoTime() - startTime;
        logger.debug( "Parsed file in " + elapsedTime / 1000 / 1000 + " ms" );

        return doc;

    }

    private static DocumentBuilder getDomDocumentBuilder() throws ParserConfigurationException {
        DocumentBuilderFactory domFactory=DocumentBuilderFactory.newInstance();
        domFactory.setNamespaceAware( true );
        DocumentBuilder documentBuilder=domFactory.newDocumentBuilder();

        return documentBuilder;
    }

    private static Document buildFromDomDoc( org.w3c.dom.Document domDoc ) {
        DOMBuilder builder=new DOMBuilder();
        Document doc=builder.build( domDoc );
        return doc;
    }

    public static Document parse( InputStream input ) {
        long startTime=System.nanoTime();

        Document doc=null;
        try {

            DocumentBuilder documentBuilder=getDomDocumentBuilder();
            org.w3c.dom.Document domDoc=documentBuilder.parse( input );
            doc=buildFromDomDoc( domDoc );
        }
        catch (Exception e) {
            raiseParsingException( e, null );
        }
        // TODO: close input stream
        long elapsedTime=System.nanoTime() - startTime;
        logger.debug( "Parsed input stream in " + elapsedTime / 1000 / 1000 + " ms" );

        return doc;
    }

    public static Document parseFromZip( File zipFile, String fileToExtract ) {
        Document doc;
        try (ZipFile zipFileInspector=new ZipFile( zipFile )) {
            ZipEntry zipEntry=zipFileInspector.getEntry( fileToExtract );
            if (zipEntry == null) {
                throw new DPBuddyException( "File '%s' was not found inside zip file '%s'", fileToExtract,
                                zipFile.getAbsolutePath() );
            }

            doc=parse( zipFileInspector.getInputStream( zipEntry ) );
        }
        catch (IOException e) {
            throw new DPBuddyException( e );
        }

        return doc;

    }

    public static Document parseUsingStax( InputStream input ) {
        long startTime=System.nanoTime();

        XMLInputFactory xmlif=XMLInputFactory.newInstance();

        Document doc=null;
        try {
            XMLStreamReader xmlr=xmlif.createXMLStreamReader( input );
            StAXStreamBuilder builder=new StAXStreamBuilder();
            doc=builder.build( xmlr );
        }
        catch (Exception e) {
            raiseParsingException( e, null );
        }

        long elapsedTime=System.nanoTime() - startTime;
        logger.debug( "Parsed input stream in " + elapsedTime / 1000 / 1000 + " ms" );

        return doc;
    }

    public static Document parseUsingStax2( Reader input ) {
        long startTime=System.nanoTime();

        XMLInputFactory2 xmlif=(XMLInputFactory2) XMLInputFactory2.newInstance();

        Document doc=null;
        try {
            XMLStreamReader2 xmlr=(XMLStreamReader2) xmlif.createXMLStreamReader( input );
            StAXStreamBuilder builder=new StAXStreamBuilder();
            doc=builder.build( xmlr );
        }
        catch (Exception e) {
            raiseParsingException( e, null );
        }

        long elapsedTime=System.nanoTime() - startTime;
        logger.debug( "Parsed input stream in " + elapsedTime / 1000 / 1000 + " ms" );

        return doc;
    }

    public static XMLStreamReader getDefaulStreamReader( InputStream input ) {
        XMLInputFactory2 xmlif=(XMLInputFactory2) XMLInputFactory2.newInstance();
        XMLStreamReader2 xmlr=null;
        try {
            xmlr=(XMLStreamReader2) xmlif.createXMLStreamReader( input );
        }
        catch (Exception e) {
            raiseParsingException( e, null );
        }

        return xmlr;
    }

    public static Document parseUsingStax( XMLStreamReader input ) {
        long startTime=System.nanoTime();

        Document doc=null;
        try {
            StAXStreamBuilder builder=new StAXStreamBuilder();
            doc=builder.build( input );
        }
        catch (Exception e) {
            raiseParsingException( e, null );
        }

        long elapsedTime=System.nanoTime() - startTime;
        logger.debug( "Parsed input stream in " + elapsedTime / 1000 / 1000 + " ms" );

        return doc;
    }

    private static void raiseParsingException( Exception e, File file ) {
        if (file != null) {
            if (e instanceof FileNotFoundException) {
                throw new DPBuddyException( "XML file '%s' does not exist", e, file.getAbsolutePath() );
            }
            if (e instanceof IOException) {
                throw new DPBuddyException( "Errors trying to read from file '%s'", e, file.getAbsoluteFile() );
            }
            else {
                throw new DPBuddyException( "Errors trying to parse xml file '%s'. Is this a valid XML file?", e,
                                file.getAbsoluteFile() );
            }
        }
        else
            throw new DPBuddyException( "Errors trying to parse XML", e );

    }

    public static void persistXML( File file, Document doc ) {
        try {
            FileUtils.writeStringToFile( file, JDomUtils.docToString( doc ), Charset.defaultCharset() );
        }
        catch (IOException ioException) {
            new DPBuddyException( "Error saving the file '%s'", ioException, file.getAbsolutePath() );
        }
    }

}
