package com.myarch.environment;

import org.apache.commons.lang3.StringUtils;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.PropertyHelper;
import org.apache.tools.ant.PropertyHelper.PropertyEvaluator;
import org.apache.tools.ant.Task;

import com.myarch.antutil.AntCommonsLogger;
import com.myarch.antutil.AntUtils;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.ant.BaseDPBuddyTask;
import com.myarch.hocon.HoconAntUtils;
import com.myarch.hocon.HoconUtils;

public class EnvironmentAntTask extends Task {

    private final static String EVALUATOR_REF_NAME="com.myarch.PropertyProvider";
    public final static String DEFAULT_PROVIDER_CLASS_NAME=PrefixedPropertyProvider.class.getName();

    private String providerClass=DEFAULT_PROVIDER_CLASS_NAME;
    private String prefix;

    public static void setEnvironmentPropertyProvider( Project project, String envPrefix ) {
        EnvironmentAntTask envTask=new EnvironmentAntTask();
        envTask.setProject( project );
        envTask.setPrefix( envPrefix );
        try{
            envTask.setUpPropertyProvider();
        }
        catch (Exception e) {
            if (e instanceof BuildException) {
                throw (BuildException) e;
            }
            else {
                // we're wrapping in DPBuddyException to force Ant not to dump the error stack for certain exceptions in non-verbose mode
                throw new BuildException( e.getMessage(), new DPBuddyException(e) );
            }
        }
        
    }

    public static String getCurrentEnvPrefix( Project project ) {
        String evaluatorRefName=EVALUATOR_REF_NAME + "_" + DEFAULT_PROVIDER_CLASS_NAME;

        Object providerObj=project.getReference( evaluatorRefName );
        if (providerObj == null || !( providerObj instanceof PrefixedPropertyProvider ))
            return null;

        PrefixedPropertyProvider provider=(PrefixedPropertyProvider) providerObj;

        return provider.getSelector();
    }

    public EnvironmentAntTask() {
        AntCommonsLogger.init( BaseDPBuddyTask.ROOT_LOGGER );

    }

    public void setPrefix( String envPrefix ) {
        this.prefix=envPrefix;
    }
    
    public void setName( String envName) {
        this.prefix=envName;
    }

    public void setProviderClass( String providerClass ) {
        this.providerClass=providerClass;
    }

    @Override
    public void execute() throws BuildException {
        BaseDPBuddyTask.setUpLogging( getProject() );

        if (prefix == null)
            throw new BuildException( "Environment prefix is required" );
        // This is needed in case if the prefix is a property and it points to a
        // group of prefixes
        String envPrefixVarValue=getProject().getProperty( prefix );
        if (envPrefixVarValue != null && envPrefixVarValue.contains( "," ) && ! isHoconConfig( envPrefixVarValue )) {
            getProject().log( "Environment is pointing to the variable. The environment will be set to the value of this variable, which is "
                            + envPrefixVarValue, Project.MSG_VERBOSE );
            
            setPrefix( envPrefixVarValue );
        }

        setUpPropertyProvider();

        AntCommonsLogger.deregister();
    }

    private boolean isHoconConfig(String configStr){
        return StringUtils.containsAny( configStr, '{', '}');
        
    }

    public static boolean isEnvironmentPrefixValid(Project project, String prefix) {
        if (StringUtils.isBlank( prefix ))
            return true;
        
        boolean isValid=HoconUtils.isPathValidAndExists( HoconAntUtils.getConfig( project ) , prefix);
        if (!isValid) {
            isValid=AntUtils.isPropertyPrefixExist( project, prefix+PrefixedPropertyProvider.DEFAULT_SEPARATOR );
        }
        
        return isValid;
    }
    
    public void setUpPropertyProvider() {
        String cleanedUpPrefix=getFirstPrefixIfMultiEnv(getProject(), prefix);
        if (cleanedUpPrefix!=null)
            cleanedUpPrefix=cleanedUpPrefix.trim();
        
        if (StringUtils.containsAny(prefix, "${", "}") ){
            throw new BuildException(String.format( "Environment prefix '%s' contains unresolved variables", prefix));
        }
/*        
        String cleanedUpPrefix=cleanUpPrefix( getProject(), prefix );
        if (!cleanedUpPrefix.equals( prefix )) {
            getProject().log(
                            String.format( "Environment prefix '%s' contains unresolved variables, changing it to '%s'",
                                            prefix, cleanedUpPrefix ),
                            Project.MSG_INFO );
        }
*/        
        getProject().log( "Setting environment prefix to '" + cleanedUpPrefix + "'",
                        Project.MSG_VERBOSE );

        PropertyEvaluator provider=obtainPropertyProvider();

        if (provider instanceof PrefixedPropertyProvider) {
            ( (PropertyProvider) provider ).setSelector( cleanedUpPrefix );
        }
        
        HoconAntUtils.setEnvironmentPrefix( getProject(), cleanedUpPrefix );

    }

    private PropertyEvaluator obtainPropertyProvider() {

        String evaluatorRefName=EVALUATOR_REF_NAME + "_" + providerClass;

        Object providerObj=getProject().getReference( evaluatorRefName );

        PropertyEvaluator provider=null;
        if (providerObj != null && providerObj instanceof PropertyEvaluator) {
            provider=(PropertyEvaluator) providerObj;
        }
        else {
            getProject().log( "Setting property provider to: " + providerClass, Project.MSG_VERBOSE );
            provider=AntUtils.<PropertyEvaluator> instantiateClass( providerClass );
            getProject().addReference( evaluatorRefName, provider );
        }
        // we have to do it every time in case if the reference got inherited
        // (delegates are not inherited)
        addToPropertyHelper( provider, getProject() );

        return provider;
    }

    private void addToPropertyHelper( PropertyEvaluator provider, Project project ) {
        PropertyHelper propertyHelper=PropertyHelper.getPropertyHelper( project );
        propertyHelper.add( provider );
    }

    /**
     * Ignore unresolved variables in the prefix and get rid of ..
     * 
     * @param prefix
     * @return
     */
    /*
    private static String cleanUpPrefix( Project project, String prefixesToClean ) {
        String prefix=getFirstPrefixIfMultiEnv( project, prefixesToClean );
        String separator=PrefixedPropertyProvider.DEFAULT_SEPARATOR;
        String[] prefixTokens=StringUtils.splitByWholeSeparator( prefix, separator );
        StringBuilder cleanPrefix=new StringBuilder();
        for (String token : prefixTokens) {
            token=token.trim();
            if (!token.startsWith( "${" )) {
                if (cleanPrefix.length() > 0)
                    cleanPrefix.append( separator );
                cleanPrefix.append( token );
            }
        }
        return cleanPrefix.toString();
    }
    */

    private static String getFirstPrefixIfMultiEnv( Project project, String prefix ) {
        // FIXME: should be common with basedptask
        String multiEnvSeparator=",";
        if (prefix.contains( multiEnvSeparator )) {

            String[] envPrefixes=StringUtils.splitByWholeSeparator( prefix, multiEnvSeparator );
            // simply select the first for now
            String selectedPrefix=envPrefixes[0];
            project.log( String.format(
                            "Muliple environment prefixes '%s' were passed to the 'environment' task. The prefix '%s' will be used",
                            prefix, selectedPrefix ) );
            return selectedPrefix;
        }
        else
            return prefix;
    }

}
