package com.myarch.hocon;

import java.util.ArrayDeque;
import java.util.Deque;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.tools.ant.PropertyHelper;
import org.apache.tools.ant.PropertyHelper.PropertyEvaluator;

import com.myarch.util.DecryptionUtils;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import com.typesafe.config.ConfigValue;

public class HoconAntPropertyProvider implements PropertyEvaluator {

    private final static String TRACE_PROP_NAME="dpbuddy.config.trace";
    private final static String ENABLE_AUTO_DECRYPTION="autoDecrypt";

    private final Log logger=LogFactory.getLog( getClass() );
    private Config origConfig=ConfigFactory.empty();
    private String environmentPrefix;

    // private Config currentConfig;
    private Deque<Config> configScopes=new ArrayDeque<>();

    public HoconAntPropertyProvider() {
        setCurrentConfig( origConfig );
    }

    public void setEnvironmentPrefix( String prefix ) {
        environmentPrefix=prefix;
        setCurrentConfig( origConfig );
    }

    public void addConfig( Config config ) {
        origConfig=config.withFallback( origConfig );
        setCurrentConfig( origConfig );
    }

    public void pushConfig( Config config ) {
        configScopes.addFirst( config );
    }

    public void popConfig() {
        configScopes.removeFirst();
    }

    public void setScope( String path ) {
        Config configWithFallback=getCurrentConfig().getConfig( path )
                        .withFallback( getCurrentConfig() );
        setCurrentConfig( configWithFallback );
    }
    
    public Object evaluate( String propertyName, PropertyHelper propertyHelper ) {
//        System.err.println( "User prop:"+propertyName );
        if (isOverrideAvailable( propertyName, propertyHelper )) {
            // kick it down to the next resolver
            return null;
        }
        return getValueFromConfig(propertyName);
    }

    private boolean isOverrideAvailable(String propertyName, PropertyHelper propertyHelper) {
        // allow for overrides using -D (user properties)
        Object propVal=propertyHelper.getUserProperty(propertyName);
        if (propVal==null && StringUtils.isNotBlank( environmentPrefix )) {
            // try with prefix
            propVal=propertyHelper.getUserProperty(environmentPrefix+"."+propertyName);
        }  
        
        return propVal!=null;
       
    }
    
    private Object getValueFromConfig( String path) {
        if (origConfig.isEmpty() || StringUtils.isBlank( path )) {
            return null;
        }

        Object val=null;
        for (Config config : configScopes) {
            if (HoconUtils.isPathValidAndExists( config, path )) {
                val=HoconEnvironmentUtils.getUnwrappedConfigValue( config, path, environmentPrefix );
                break;
            }
        }

        if ( val!=null && val instanceof String && isAutoDecrypt()) {
            val=DecryptionUtils.decryptIfEncrypted( (String)val );
        }
        

        
        return val;
    }

    private boolean isAutoDecrypt() {
        boolean enabled=true;
        Config config=getCurrentConfig();
        if (HoconUtils.isPathValidAndExists( config, ENABLE_AUTO_DECRYPTION )) {
            Object val=HoconEnvironmentUtils.getUnwrappedConfigValue( config, ENABLE_AUTO_DECRYPTION, environmentPrefix );
            enabled=val instanceof Boolean && ( (boolean) val );
        }

        return enabled;
    }

    public final static String ENV_CASE_KEYWORD="_env";
    public final static String DEFAULT_KEYWORD="_default";


    
    
    public ConfigValue getWrappedConfigValue( String path ) {
        return HoconEnvironmentUtils.getWrappedConfigValue( getCurrentConfig(), path, environmentPrefix );
    }
    

    Deque<Config> getScopes() {
        return configScopes;
    }

    private void setCurrentConfig( Config config ) {
        if (isTraceEnabled( config )) {
            logger.info( "Setting config to:\n" + HoconUtils.configToString( config ) );
        }

        config=config.resolve();

        if (isTraceEnabled( config ) && !config.isEmpty()) {
            logger.info( "Configuration after variable substitution:\n"
                + HoconUtils.configToString( config ) );
        }

        if (environmentPrefix != null) {
            config=HoconUtils.buildFallbackHierarchy( config, environmentPrefix );
        }

        configScopes.pollLast();
        configScopes.addLast( config );
    }

    public Config getCurrentConfig() {
        return configScopes.getLast();
    }

    private boolean isTraceEnabled( Config config ) {
        return config.hasPath( TRACE_PROP_NAME ) && config.getBoolean( TRACE_PROP_NAME );
    }

}
