package com.myarch.hocon;

import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

import com.google.gson.JsonObject;
import com.myarch.dpbuddy.DPBuddyException;
import com.myarch.dpbuddy.DPObject;
import com.typesafe.config.*;

public class HoconUtils {

    public final static ConfigRenderOptions DEFAULT_RENDER_OPTIONS=ConfigRenderOptions.defaults()
                    .setOriginComments( false );

    /**
     * Add all key/valus fron hocon to json. Only strings are added
     * 
     * @param json
     * @param hocon
     */
    public static void addStringValuesToJson( ConfigObject hocon, JsonObject json ) {
        Set<Entry<String, ConfigValue>> hoconEntries=hocon.entrySet();
        for (Entry<String, ConfigValue> hoconEntry : hoconEntries) {
            ConfigValue hoconVal=hoconEntry.getValue();
            if (hoconVal.valueType() == ConfigValueType.STRING) {
                json.addProperty( hoconEntry.getKey(), hoconVal.unwrapped().toString() );
            }
        }
    }

    public static String configToString( Config config ) {
        return configValueToString( config.root() );
    }

    public static String configValueToString( ConfigValue config ) {
        String hoconS=config.render( DEFAULT_RENDER_OPTIONS );
        return hoconS;
    }

    public static String get( ConfigObject confObject, String key ) {
        if (!confObject.containsKey( key )) {
            return null;
        }
        return confObject.get( key ).unwrapped().toString();
    }

    public static Config configValueToConfig( ConfigValue configValue ) {
        if (configValue.valueType() != ConfigValueType.OBJECT) {
            throw new DPBuddyException(
                            "Expected JSON/configuration object, the following value was provided instead: %s",
                            configValueToString( configValue ) );
        }

        return ( (ConfigObject) configValue ).toConfig();
    }

    public static Config addToConfig( Config origConfig, String path, Object val ) {
        ConfigValue configVal=ConfigValueFactory.fromAnyRef( val );
        return origConfig.withValue( path, configVal );
    }

    public static boolean isNull( ConfigValue configVal ) {
        return configVal == null || configVal.valueType() == ConfigValueType.NULL;
    }

    public static boolean isPrimitiveType( ConfigValue configVal ) {
        return configVal != null
            && ( configVal.valueType() == ConfigValueType.NUMBER || configVal.valueType() == ConfigValueType.STRING );
    }

    public final static String PATH_SEPARATOR=".";
    /*
     * Hierarchy: l1.l2.l3
     * origConfig->Config(l1)->config(l1.l2)->config(l1.l2.l3)
     */

    public static Config buildFallbackHierarchy( Config origConfig, String path ) {
        if (StringUtils.isBlank( path )) {
            return origConfig;
        }
        Config currentConfig=origConfig;
        String pathKeys[]=StringUtils.splitByWholeSeparator( path, PATH_SEPARATOR );

        for (String pathKey : pathKeys) {
            // terminal keys might be missing, we don't care
            if (!currentConfig.hasPath( pathKey )) {
                break;
            }
            currentConfig=currentConfig.getConfig( pathKey ).withFallback( currentConfig );
        }

        return currentConfig;
    }

    public static boolean isPathValidAndExists( Config config, String path ) {
        boolean hasPath=false;
        try {
            hasPath=config.hasPath( path );
        }
        catch (com.typesafe.config.ConfigException.BadPath badPathEx) {
            // ignore, invalid path
        }

        return hasPath;
    }

    public static List<Pair<String, String>> configToNameStringValueList( Config config ) {
        Set<Entry<String, ConfigValue>> rootEntries=config.root().entrySet();

        List<Pair<String, String>> nameValues=new ArrayList<>();
        for (Entry<String, ConfigValue> entry : rootEntries) {
            ConfigValue configVal=entry.getValue();
            if (configVal != null && ( configVal.valueType() != ConfigValueType.OBJECT
                || configVal.valueType() != ConfigValueType.LIST )) {
                String val=configVal.unwrapped().toString();
                nameValues.add( ImmutablePair.of( entry.getKey(), val ) );
            }
        }

        return nameValues;
    }

    public static List<DPObject> configToDPObjectList( Config config ) {
        List<Pair<String, String>> objectTuples=configToNameStringValueList( config );
        List<DPObject> dpObjects=new ArrayList<>();

        for (Pair<String, String> objectTuple : objectTuples) {
            dpObjects.add( new DPObject( objectTuple ) );
        }

        return dpObjects;
    }

    public static class ConfObjWithKey {

        private String key;
        private ConfigObject confObj;

        public ConfObjWithKey(String key, ConfigObject confObj) {
            super();
            this.key=key;
            this.confObj=confObj;
        }

        public String key() {
            return key;
        }

        public ConfigObject confObj() {
            return confObj;
        }
    }

}
